/*
** Module   : ConfigFieldSetterAspect.java
** Abstract : An aspect which intercepts field assignments in widget config classes.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140926 Created initial version.
** 002 CA  20141207 Transient fields must not be tracked for dirty changes.
** 003 HC  20141207 Implemented synchronization of linked widget coordinate attributes.
*/ 

package com.goldencode.p2j.aspects.ui;

import org.aspectj.lang.*;
import org.aspectj.lang.annotation.*;

import com.goldencode.p2j.ui.*;

/**
 * Aspect which intercepts the assignment of public fields from the 
 * {@code com.goldencode.p2j.ui.*Config} classes.
 * <p>
 * Only the client-side uses this aspect.  All calls from the server-side are ignored.  The field
 * assignments during {@link WidgetConfig} instantiation are also ignored.
 * <p>
 * AspectJ Point Cuts are called during custom setters and only when the field is set outside of 
 * this code:
 * <ol>
 *   <li>{@link WidgetConfig#WidgetConfig constructors}</li>
 *   <li>{@link WidgetConfig#readExternal} and {@link WidgetConfig#writeExternal}</li>
 *   <li>{@link WidgetConfig#applyConfig}</li>
 *   <li>serve-side code</code>
 *   <li><code>com.goldencode.p2j.ui</code> package</li>
 * </ol>
 * While these methods are executed, changes must not be tracked; field changes can be tracked
 * only when the client-side UI runtime changes it explicitly.
 * <p>
 * When collecting the changes, {@link ClientConfigManager#getConfigUpdates} will use the IDs of
 * the collected {@link WidgetConfig} instances, and only the instance with an associated active
 * config will be used.
 */
@Aspect()
public class ConfigFieldSetterAspect
{
   /**
    * Aspect which intercepts the assignment of public fields from the 
    * {@code com.goldencode.p2j.ui.*Config} classes.
    * 
    * @param    cfg
    *           The widget configuration.
    * @param    sp
    *           The details about the static part of this join point.
    */
   @Before("set(public !transient * com.goldencode.p2j.ui.*Config.*) " +
           "&& target(cfg) " +
           "&& (within(com.goldencode.p2j.ui..*)" +
           "   || within(com.goldencode.p2j.aspects.ui.SyncCoordinatesAspect))" +
           "&& !withincode(void readExternal(..)) " +
           "&& !withincode(void applyConfig(..)) " +
           "&& !withincode(com.goldencode.p2j.ui.*Config.new(..)) " +
           "&& !within(com.goldencode.p2j.ui.GenericFrame) " +
           "&& !within(com.goldencode.p2j.ui.*Widget) " +
           "&& !within(com.goldencode.p2j.ui.*Entity) ")
   public void beforeSetField(WidgetConfig cfg, JoinPoint.StaticPart sp)
   throws Throwable
   {
      ConfigManager mgr = ConfigManager.getInstance();
      String field = sp.getSignature().getName();

      mgr.addDirtyConfig(cfg, field);
   }
}
