/*
** Module   : SyncCoordinatesAspect.java
** Abstract : An aspect which intercepts and synchronizes certain coordinate field assignments in widget config classes.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 HC  20141207 Created initial version.
*/
package com.goldencode.p2j.aspects.ui;
   
import org.aspectj.lang.JoinPoint;
import org.aspectj.lang.annotation.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;

/**
 * The aspect implements synchronization of linked coordinate attributes like 
 * WIDTH-CHARS/WIDTH-PXIELS, COLUMN/X, etc. by intercepting the selected public 
 * fields from the {@code com.goldencode.p2j.ui.*Config} classes. 
 * <p>
 * Only the client-side uses this aspect. All calls from the server-side are ignored. 
 * The field assignments during {@link WidgetConfig} instantiation are also ignored as 
 * well as calls to {@link WidgetConfig#applyConfig(WidgetConfig)}.
 */
@Aspect
public class SyncCoordinatesAspect
{
   /**
    * Cached coordinates conversion reference.
    */
   private CoordinatesConversion cc;
   
   /**
    * Cached config manager reference.
    */
   private ConfigManager cm;
   
   /**
    * Default constructor.
    */
   public SyncCoordinatesAspect()
   {
      ThinClient tc = ThinClient.getInstance();
      if (tc != null)
      {
         cc = tc.getOutputManager().coordinates();
         cm = ConfigManager.getInstance();
      }
   }
   
   /**
    * Common pointcut defining the client-side calls.
    */
   @Pointcut("within(com.goldencode.p2j.ui..*)" +
             "&& !withincode(void readExternal(..)) " +
             "&& !withincode(void applyConfig(..)) " +
             "&& !withincode(com.goldencode.p2j.ui.*Config.new(..)) " +
             "&& !within(com.goldencode.p2j.ui.GenericFrame) " +
             "&& !within(com.goldencode.p2j.ui.*Widget) " +
             "&& !within(com.goldencode.p2j.ui.*Entity)")
   public void withinClientPoint()
   {
   }
   
   /**
    * The after-advice for COLUMN attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.column)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public  void afterSetColumn(BaseConfig cfg, JoinPoint.StaticPart sp)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.x = cc.columnToPixels(cfg.column);
      cfg.column = cc.columnFromPixels(cfg.x); 
   }
   
   /**
    * The after-advice for ROW attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.row)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetRow(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.y = cc.rowToPixels(cfg.row);
      cfg.row = cc.rowFromPixels(cfg.y); 
   }
   
   /**
    * The after-advice for X attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.x)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetX(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.column = cc.columnFromPixels(cfg.x);
   }
   
   /**
    * The after-advice for Y attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.y)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetY(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.row = cc.rowFromPixels(cfg.y);
   }
   
   /**
    * The after-advice for WIDTH-CHARS attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.widthChars)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetWidthChars(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.widthPixels = cc.columnToPixels(cfg.widthChars);
      cfg.widthChars = cc.columnFromPixels(cfg.widthPixels); 
   }
   
   /**
    * The after-advice for HEIGHT-CHARS attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.heightChars)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetHeightChars(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.heightPixels = cc.rowToPixels(cfg.heightChars);
      cfg.heightChars = cc.rowFromPixels(cfg.heightPixels); 
   }

   /**
    * The after-advice for WIDTH-PIXELS attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.widthPixels)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetWidthPixels(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.widthChars = cc.columnFromPixels(cfg.widthPixels);
   }

   /**
    * The after-advice for HEIGHT-PIXELS attribute.
    * 
    * @param   cfg
    *          The target config reference being intercepted.
    */
   @After("set(public * com.goldencode.p2j.ui.BaseConfig.heightPixels)" +
          "&& target(cfg) " +
          "&& withinClientPoint()")
   public void afterSetHeightPixels(BaseConfig cfg)
   {
      if (cc == null || cm.isInsideApplyConfig())
      {
         return;
      }
      
      cfg.heightChars = cc.rowFromPixels(cfg.heightPixels);
   }
}
