/*
** Module   : FieldReference.java
** Abstract : Buffer field reference which allows deferred read/write access
**
** Copyright (c) 2004-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 ECF 20051112   @23388 Created initial version. Resolves a buffer
**                           field reference value at runtime. Used as a
**                           query substitution parameter for field
**                           references which can only be resolved as the
**                           query is executing and has partial results.
** 002 ECF 20051216   @23756 Added getValue method. Returns resolved value
**                           but reports a warning to the UI (via the
**                           error manager) if resolution fails.
** 003 ECF 20060104   @23822 Implement Resolvable interface. Allows this
**                           class to be used generically with util APIs.
** 004 ECF 20060202   @24223 Implement Accessor interface. Field access is
**                           now read/write instead of read-only.
** 005 ECF 20060307   @25015 Added new constructor and get method variant.
**                           These permit late binding of a DMO instance.
** 006 ECF 20060323   @25226 Added support for case-insensitive field
**                           references used as query substitution
**                           parameters. New constructor variants accept
**                           an uppercase flag.
** 007 GES 20060907   @29285 Setting an extent field was broken. The index
**                           had to be the first parameter instead of the
**                           second.
** 008 ECF 20060927   @29970 Added special exception handling in set() and
**                           getObject(). ConditionException wrapped in an
**                           InvocationTargetExcpetion is unwrapped and
**                           rethrown.
** 009 ECF 20061027   @30749 Replaced Object with DataModelObject for DMO
**                           arguments to methods/constructors. Required
**                           for compile-time type safety.
** 010 ECF 20070416   @33022 Integrated user interrupt handling.
** 011 ECF 20080528   @38480 Integrated generics.
** 012 ECF 20080606   @38645 Fixed FieldReference(Class<?>, String) c'tor
**                           variant. It was not properly setting the
**                           uppercase instance variable in the case of
**                           references to case-insensitive properties.
**                           Also added support for the reserved 'id'
**                           property in getGetter() method.
** 013 CA  20080815   @39458 Support API change in DBUtils.
** 014 ECF 20090701   @43217 Enabled lazy initialization of backing record
**                           buffer. Database instance variable is extracted
**                           from buffer lazily, rather than at construction.
** 015 SVL 20090723   @43342 Added getPropertyName() function.
** 016 ECF 20090819   @43695 Downgraded some log messages from ERROR to DEBUG.
** 017 GES 20090828   @43799 Don't unconditionally shift into warning mode as
**                           that would disable warning mode on exit from the
**                           getter even if it had already been enabled.
** 018 AIL 20121206          Update from character to TextOps.
** 019 VMN 20130207          Extended from FieldInterface and added support for attributes
**                           BUFFER-VALUE (KW_BUF_VAL)
**                           COLUMN-LABEL (KW_COL_LAB)
**                           EXTENT (KW_EXTENT)
**                           FORMAT (KW_FORMAT)
**                           STRING-VALUE (KW_STR_VAL)
** 020 VMN 20130213          Added support for attributes
**                           BUFFER-NAME (KW_BUF_NAME),
**                           LITERAL-QUESTION (KW_LIT_QSTN)
**                           VALIDATE-MESSAGE (KW_VAL_MSG)
**                           MANDATORY (KW_MAND)
**                           VALIDATE-EXPRESSION (KW_VAL_EXPR)
**                           CURRENT-ITERATION (KW_CUR_ITER)
** 021 CA  20130215          Added dynamic EXTENT and EXTENT parameter support.
** 022 VMN 20130217          Corrected for attribute
**                           CURRENT-ITERATION (KW_CUR_ITER)
**                           WIDTH-CHARS (KW_WIDTH_C)
** 023 VMN 20130219          Added support for attributes
**                           POSITION (KW_POS)
** 024 CA  20130224          Added support for UNIQUE-ID, AMD-DATA, Sizeable and CommonField
**                           attributes. Fixed CHANGE-VALUE. Added Deletable support.
** 025 OM  20130219          Changed return type of value (BUFFER-VALUE)
** 026 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 027 OM  20130304          Added missing overloaded method for value(). 
** 028 CA  20130315          Added missing API for STRING-VALUE.
** 029 ECF 20130417          Moved initial() method from BufferImpl.
** 030 CA  20130603          resourceDelete must return a boolean value.
** 031 SVL 20130822          Implemented extent-related functions and value() function. Added
**                           getParentBuffer.
** 032 HC  20131215          Implemented ADM-DATA and UNIQUE-ID attribute.
** 033 VMN 20140123          Implemented extent(), initial(), stringValue(), getFormat(),
**                           setFormat(), columnLabel(), changeColumnLabel(), getColumnLabel(),
**                           setColumnLabel(), bufferName(), getDbName(), getTable(),
**                           position(), getDataType(), valid(), setValue() for non BaseDataType
**                           values.
** 034 OM  20140122          Fixed value() and value(int) to return unknown value when they are
**                           called for EXTENT, resp. non-EXTENT fields.
**                           Added Long support for changeValue(), minor esthetic changes.
** 035 SVL 20140205          Fixed value(int).
** 036 EVL 20140217          Fixed javadoc comments to be able to compile in Winodws 1252 CP.
** 037 ECF 20140318          Added logic to compensate for instance of unknown being passed to
**                           changeValue(). Replaced Apache commons logging with J2SE logging.
**                           Added error handling for BUFFER-VALUE() to validate index values
**                           passed to value(NumberType) and changeValue(NumberType) (0 is a valid
**                           index for scalar fields). Added widening conversion in changeValue
**                           for polymorphic BaseDataType arguments. Needs additional work to
**                           exactly mimic 4GL errors.
** 038 ECF 20140415          Set NAME attribute to its default value the first time name() is
**                           invoked.
** 039 HC  20140613          Fixed extent fields runtime support. 
** 040 HC  20141207          Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.persist;

import java.lang.reflect.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.*;
import com.goldencode.p2j.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;

/**
 * A readable, writable reference to a buffer field, used in cases where the
 * field needs to be accessed at some arbitrary point in time after it is
 * referenced in application code.  For example, an instance of this class
 * would be used as a substitution parameter in a {@link CompoundQuery}, when
 * an inner loop where clause of a multi-table query refers to a field within
 * a buffer retrieved in an outer loop.
 * <p>
 * This class provides both read and write capability to the backing field,
 * insofar as the DMO has both an accessor (getter) and a mutator (setter)
 * method defined for the specified property.  Both indexed and non-indexed
 * versions of these methods are supported.  It is assumed that the methods
 * have bean-like signatures.  Specifically, the following, public signature
 * patterns are assumed:
 * <p>
 * <dt>Simple Getter:</dt>
 * <dd>
 * <code>&lt;BaseDataType subclass&gt; get&lt;Property&gt;()</code>
 * </dd>
 * <dt>Indexed Getter:</dt>
 * <dd>
 * <code>&lt;BaseDataType subclass&gt; get&lt;Property&gt;(int)</code>
 * </dd>
 * <dt>Simple Setter:</dt>
 * <dd>
 * <code>void set&lt;Property&gt;(&lt;BaseDataType subclass&gt;)</code>
 * </dd>
 * <dt>Indexed Setter:</dt>
 * <dd>
 * <code>void set&lt;Property&gt;(&lt;BaseDataType subclass&gt;, int)</code>
 * </dd>
 * <p>
 */
public final class FieldReference
extends HandleChain
implements Resolvable,
           Accessor,
           FieldInterface,
           Deletable
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(FieldReference.class);
   
   /** Maps DMO interfaces to maps of property names to getter methods */
   private static final Map<Class<?>, Map<String, Method>> getterCache = new HashMap<>();
   
   /** Maps DMO interfaces to maps of property names to setter methods */
   private static final Map<Class<?>, Map<String, Method>> setterCache = new HashMap<>();
   
   /** Maps DMO interfaces to maps of property names to sizer methods */
   private static final Map<Class<?>, Map<String, Method>> sizerCache = new HashMap<>();
   
   /** Property name */
   private final String property;
   
   /** Method used to resolve this field reference at query execution time */
   private final Method getter;
   
   /** Method used to resolve this field reference at query execution time */
   private final Method setter;
   
   /** Method used to resolve this field's size (in case of extent fields). */
   private final Method sizer;
   
   /** Optional index, used with a DMO indexed getter method */
   private final NumberType index;
   
   /** The associated database */
   private Database database = null;
   
   /** DMO or DMO proxy instance;  the referent of the referencing method */
   private Object dmo = null;
   
   /** Flag to force resolved character result to uppercase/right-trim */
   private boolean uppercase = false;
   
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();
   
   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID = new integer(UniqueIdGenerator.getUniqueId(IdKind.BUFFER_FIELD));
   
   /**
    * Create a new field reference given a DMO proxy and property name.  A
    * non-indexed getter method will be used to resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    */
   public FieldReference(DataModelObject dmo, String property)
   {
      this(dmo, property, null);
   }
   
   /**
    * Create a new field reference given a DMO proxy and property name.  A
    * non-indexed getter method will be used to resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value
    *          returned by {@link #getObject} to be uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    */
   public FieldReference(DataModelObject dmo, String property, boolean uppercase)
   {
      this(dmo, property, uppercase, null);
   }
   
   /**
    * Create a new field reference given a DMO interface and property name.
    * A non-indexed getter method will be used to resolve the field.  A DMO
    * instance must be provided at resolution time, using the {@link
    * #get(Persistable)} variant of the <code>get</code> method.
    * <p>
    * Note that the created instance can be used to read the related field only.
    * 
    * @param   database
    *          The associated database.
    * @param   dmoIface
    *          Interface of the target DMO.
    * @param   property
    *          Target property name
    */
   FieldReference(Database database, Class<?> dmoIface, String property)
   {
      this.database = database;
      this.property = property;
      this.getter = getGetter(dmoIface, property, false);
      this.sizer = getSizer(dmoIface, property);
      this.setter = null;
      this.index = null;
      
      if (character.class.equals(getter.getReturnType()))
      {
         String schema = DatabaseManager.getSchemaByInterface(dmoIface);
         if (!DMOIndex.isCaseSensitive(schema, dmoIface, property))
         {
            this.uppercase = true;
         }
      }
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    */
   public FieldReference(DataModelObject dmo, String property, int index)
   {
      this(dmo, property, false, new integer(index));
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value
    *          returned by {@link #getObject} to be uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    */
   public FieldReference(DataModelObject dmo, String property, boolean uppercase, int index)
   {
      this(dmo, property, uppercase, new integer(index));
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property.  An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    *
    * @throws  IllegalArgumentException
    *          if field reference is not recognized.
    */
   public FieldReference(DataModelObject dmo, String property, NumberType index)
   {
      this(dmo, property, false, index);
   }
   
   /**
    * Create a new field reference given a DMO proxy, a property name, and
    * an index for the property. An indexed getter method will be used to
    * resolve the field.
    *
    * @param   dmo
    *          DMO proxy.
    * @param   property
    *          Target property name.
    * @param   uppercase
    *          <code>true</code> to force a <code>character</code> value
    *          returned by {@link #getObject} to be uppercased/right-trimmed;
    *          <code>false</code> to leave the result unchanged.
    * @param   index
    *          Index to pass to property's getter method at resolve time.
    *
    * @throws  IllegalArgumentException
    *          if field reference is not recognized.
    */
   public FieldReference(DataModelObject dmo,
                         String property,
                         boolean uppercase,
                         NumberType index)
   {
      this.dmo = dmo;
      this.property = property;
      RecordBuffer buffer = RecordBuffer.get(dmo);
      Class<?> dmoIface = buffer.getDMOInterface();
      
      // get getter/setter either for an extent component or extent field
      this.getter = getGetter(dmoIface, property, index != null);
      this.setter = getSetter(dmoIface, property, index != null);
      // Initialize sizer for extent fields and do not initialize it for non-extent fields or
      // extent elements.
      this.sizer = (index == null) ? getSizer(dmoIface, property) : null;
      this.index = index;
      this.uppercase = uppercase;
      
      if (getter == null || setter == null)
      {
         throw new IllegalArgumentException(
            "Invalid field reference [" + buffer.getDMOName() + "." +
            property + (index == null ? "" : ":" + index.toStringMessage()) +
            "]");
      }
      
      if (uppercase && !getter.getReturnType().equals(character.class))
      {
         throw new IllegalArgumentException(
            "Only a character-based FieldReference can be set to uppercase");
      }
   }
   
   /**
    * Convenience method to produce a <code>BaseDataType</code> instance
    * initialized to the unknown value.  An attempt is made to return an
    * object of the correct type (i.e., the getter method's return type).
    * If this fails for some reason, a new instance of <code>unknown</code>
    * is returned.
    *
    * @param   getter
    *          Getter method whose return type is used to determine the type
    *          of object to return.
    *
    * @return  An object whose type matches the return type of the getter
    *          method, or an instance of <code>unknown</code>.
    */
   static BaseDataType unknownValue(Method getter)
   {
      BaseDataType value;
      try
      {
         value = (BaseDataType) getter.getReturnType().newInstance();
         value.setUnknown();
      }
      catch (InstantiationException exc)
      {
         value = new unknown();
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException(exc);
      }
      
      return value;
   }
   
   /**
    * Look up the sizer method for the specified property and interface. Introspection is used 
    * the first time this method is invoked for a new DMO type;  thereafter, the sizer is 
    * retrieved from a cache.
    * <p>
    * For the reserved <code>id</code> property, {@link Persistable} interface is substituted for
    * <code>dmoIface</code>.
    *
    * @param   dmoIface
    *          Referent DMO interface.
    * @param   property
    *          Name of target property (i.e., field).
    *
    * @return  Sizer method, or <code>null</code> if <code>property</code> does not match any 
    *          property name in the DMO API or is not an extent property.
    */
   private static Method getSizer(Class<?> dmoIface, String property)
   {
      Map<String, Method> propMap = null;
      Class<?> iface = ("id".equals(property) ? Persistable.class : dmoIface);
      
      synchronized (sizerCache)
      {
         propMap = sizerCache.get(iface);
         if (propMap == null)
         {
            propMap = PropertyHelper.sizersByProperty(iface);
            sizerCache.put(iface, propMap);
         }
      }
      
      return propMap.get(property);
   }

   /**
    * Look up the getter method for the specified property and interface.
    * Introspection is used the first time this method is invoked for a
    * new DMO type;  thereafter, the getter is retrieved from a cache.
    * <p>
    * For the reserved <code>id</code> property, {@link Persistable} interface
    * is substituted for <code>dmoIface</code>.
    * <p>
    * For extent properties the method is capable to return the indexed or
    * the non-indexed getter.
    *
    * @param   dmoIface
    *          Referent DMO interface.
    * @param   property
    *          Name of target property (i.e., field).
    * @param   indexed
    *          The flag indicating whether the indexed or the non-indexed getter
    *          should be returned. The {@code true} value is meaningful only
    *          for extent properties.
    *
    * @return  Getter method, or <code>null</code> if <code>property</code>
    *          does not match any property name in the DMO API.
    */
   private static Method getGetter(Class<?> dmoIface, String property, boolean indexed)
   {
      Map<String, Method> propMap = null;
      Class<?> iface = ("id".equals(property) ? Persistable.class : dmoIface);
      
      synchronized (getterCache)
      {
         propMap = getterCache.get(iface);
         if (propMap == null)
         {
            propMap = PropertyHelper.gettersByProperty(iface);
            propMap = fetchIndexedPropertyMethods(iface, propMap);
            
            getterCache.put(iface, propMap);
         }
      }
      
      property = indexed ? property + "[]" : property;
      return propMap.get(property);
   }
   
   /**
    * Look up the setter method for the specified property and interface.
    * Introspection is used the first time this method is invoked for a
    * new DMO type;  thereafter, the setter is retrieved from a cache.
    * <p>
    * For extent properties the method is capable to return the indexed or
    * the non-indexed setter.
    *
    * @param   dmoIface
    *          Referent DMO interface.
    * @param   property
    *          Name of target property (i.e., field).
    * @param   indexed
    *          The flag indicating whether the indexed or the non-indexed setter
    *          should be returned. The {@code true} value is meaningful only
    *          for extent properties.
    *
    * @return  Setter method, or <code>null</code> if <code>property</code>
    *          does not match any property name in the DMO API.
    */
   private static Method getSetter(Class<?> dmoIface, String property, boolean indexed)
   {
      Map<String, Method> propMap = null;
      synchronized (setterCache)
      {
         propMap = setterCache.get(dmoIface);
         if (propMap == null)
         {
            propMap = PropertyHelper.settersByProperty(dmoIface);
            propMap = fetchIndexedPropertyMethods(dmoIface, propMap);
            setterCache.put(dmoIface, propMap);
         }
      }
      
      property = indexed ? property + "[]" : property;
      return propMap.get(property);
   }
   
   /**
    * Get the <code>name</code> attribute of a BUFFER-FIELD, which is the legacy name of the
    * field referenced by this object.
    * 
    * @return  See above.
    */
   @Override
   public character name()
   {
      if (name.isUnknown())
      {
         // assign directly to the variable, rather than through the API, to avoid a 4052 error
         this.name.assign(TableMapper.getLegacyFieldName((Buffer) dmo, property));
      }
      
      return super.name();
   }
   
   /**
    * Determine the data type that will be returned when this field reference
    * is resolved.
    *
    * @return  Class of the appropriate data type.
    */
   public Class<?> getType()
   {
      return getter.getReturnType();
   }
   
   /**
    * Resolve the field reference's value, bracketed by calls to the error
    * manager to first enable, then disable, warning mode.  This is meant to
    * support the need to continue processing after resolution of a field
    * reference fails, due to an empty record buffer.
    *
    * @return  Current value of the target property in the referent, or the
    *          unknown value if this resolution fails.
    */
   public BaseDataType getValue()
   {
      boolean warning = ErrorManager.isWarningMode();
      
      if (!warning)
      {
         ErrorManager.warningModeEnable();
      }
      
      try
      {
         return get();
      }
      
      finally
      {
         if (!warning)
         {
            ErrorManager.warningModeDisable();
         }
      }
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property.  Note that this may return the unknown
    * value.
    *
    * @return  Current value of the target property in the referent.
    */
   public BaseDataType resolve()
   {
      return get();
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property.  Note that this may return the unknown
    * value.
    *
    * @return  Current value of the target property in the referent.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   public BaseDataType get()
   {
      return (BaseDataType) getObject();
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter method for the 
    * target property. This must be used only in case of extent fields.
    *
    * @return  Current value of the target property in the referent.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there is some other error at
    *          the invocation target.
    */
   public BaseDataType[] getExtent()
   {
      if (sizer == null)
      {
         throw new RuntimeException("Trying to retrieve extent data for a non-extent field.");
      }
      
      return (BaseDataType[]) getObject();
   }
   
   /**
    * Set the current value of the referent field by invoking the setter
    * method for the target property.
    *
    * @param   value
    *          New value to set into the field represented by this object.
    *
    * @throws  ErrorConditionException
    *          if this field reference represents an element in an array, but
    *          the index provided at construction is the unknown value;
    *          if any persistence-level error occurs setting the value into
    *          the DMO, including validation errors.
    * @throws  IllegalStateException
    *          if the backing property is not write accessible (i.e., no
    *          setter method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   public void set(BaseDataType value)
   {
      set(new BaseDataType[] { value });
   }
   
   /**
    * Set the current value of the referent field by invoking the setter method for the target 
    * property.
    * <p>
    * For non-extent fields, the value is placed on the first position in the array. For extent
    * fields, the value is the actual value which needs to be set for the array.
    * 
    * @param   value
    *          New value to set into the field represented by this object.
    *
    * @throws  ErrorConditionException
    *          if this field reference represents an element in an array, but the index provided 
    *          at construction is the unknown value; if any persistence-level error occurs setting
    *          the value into the DMO, including validation errors.
    * @throws  IllegalStateException
    *          if the backing property is not write accessible (i.e., no setter method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there is some other error at
    *          the invocation target.
    */
   public void set(BaseDataType[] value)
   {
      if (setter == null)
      {
         throw new IllegalStateException(
            property + " property is not write-accessible [" +
            dmo.getClass().getName() + "]");
      }
      
      Object[] args;

      // if we are not extent or if we have a subscript, then the value is not an array.
      if (sizer == null || index != null)
      {
         args = new Object[] { value[0] };
      }
      else
      {
         args = new Object[] { value };
      }
      
      // Handle extent fields.
      if (index != null)
      {
         // An unknown value index constitutes an error.
         if (index.isUnknown())
         {
            ErrorManager.recordOrThrowError(
               26,
               "Array subscript 0 is out of range");
            
            return;
         }
         
         // Set up arguments, including index.
         args = new Object[] { new Integer(index.intValue()), args[0] };
      }
      
      // Invoke the setter method on the proxy.
      try
      {
         setter.invoke(dmo, args);
      }
      catch (InvocationTargetException exc)
      {
         handleException(exc);
         
         Throwable cause = exc.getCause();
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Error setting value into field", cause);
         }
         
         if (cause instanceof NumberedException)
         {
            ErrorManager.recordOrThrowError((NumberedException) cause);
         }
         else if (cause instanceof ConditionException)
         {
            throw (ConditionException) cause;
         }
         else
         {
            // Represents programming error.
            throw new RuntimeException(cause);
         }
      }
      catch (IllegalAccessException | IllegalArgumentException exc)
      {
         // Represents programming or JVM configuration error.
         throw new RuntimeException(exc);
      }
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property, using the given DMO as the target.
    * Note that this may return <code>null</code> (or the unknown value for
    * <code>BaseDataTypes</code> return types).
    *
    * @param   dmo
    *          Target object which is used as the method's referent.
    *
    * @return  Current value of the target property in the referent or
    *          <code>null</code> if no value exists.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   Object get(DataModelObject dmo)
   {
      this.dmo = dmo;
      
      return getObject();
   }
   
   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property, using the given DMO as the target.
    * Note that this may return <code>null</code> (or the unknown value for
    * <code>BaseDataTypes</code> return types).
    *
    * @param   dmo
    *          Target object which is used as the method's referent.
    *
    * @return  Current value of the target property in the referent or
    *          <code>null</code> if no value exists.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   Object get(Persistable dmo)
   {
      this.dmo = dmo;
      
      return getObject();
   }

   /**
    * Returns target property name.
    *
    * @return See above.
    */
   String getPropertyName()
   {
      return property;
   }

   /**
    * Returns the record buffer which contains the referent DMO.
    *
    * @return See above
    */
   RecordBuffer getParentBuffer()
   {
      return RecordBuffer.get((DataModelObject) dmo);
   }

   /**
    * Resolve the current value of the referent field by invoking the getter
    * method for the target property. Note that this may return
    * <code>null</code> (or the unknown value for <code>BaseDataTypes</code>
    * return types).
    *
    * @return  Current value of the target property in the referent or
    *          <code>null</code> if no value exists.
    *
    * @throws  ErrorConditionException
    *          if any persistence level error occurs getting the value from
    *          the DMO.
    * @throws  IllegalStateException
    *          if the backing property is not read accessible (i.e., no getter
    *          method was found).
    * @throws  RuntimeException
    *          if the backing, getter method cannot be accessed or if there
    *          is some other error at the invocation target.
    */
   private Object getObject()
   {
      if (getter == null)
      {
         throw new IllegalStateException(
            property + " property is not read-accessible [" +
            dmo.getClass().getName() + "]");
      }
      
      Object[] args;
      
      // Handle extent fields.
      if (index != null)
      {
         // An unknown value index propagates to an unknown value result.
         if (index.isUnknown())
         {
            return unknownValue(getter);
         }
         
         // Set index argument.
         args = new Object[] { new Integer(index.intValue()) };
      }
      else
      {
         args = null;
      }
      
      // Invoke the getter method on the proxy and return the result.
      try
      {
         Object result = getter.invoke(dmo, args);
         
         // If marked for uppercase, then we must uppercase and right trim
         // the result.  This is a special case where the result will be used
         // in Progress-compatible comparisons.
         if (uppercase)
         {
            character c = (character) result;
            result = TextOps.toUpperCase(TextOps.rightTrim(c));
         }
         
         return result;
      }
      catch (InvocationTargetException exc)
      {
         handleException(exc);
         
         Throwable cause = exc.getCause();
         
         if (LOG.isLoggable(Level.FINE))
         {
            LOG.log(Level.FINE, "Error getting value from field", cause);
         }
         
         if (cause instanceof NumberedException)
         {
            ErrorManager.recordOrThrowError((NumberedException) cause);
         }
         else if (cause instanceof ConditionException)
         {
            throw (ConditionException) cause;
         }
         else
         {
            // Represents programming error.
            throw new RuntimeException(cause);
         }
         
         return unknownValue(getter);
      }
      catch (IllegalAccessException exc)
      {
         // Represents programming or JVM configuration error.
         throw new RuntimeException(exc);
      }
   }
   
   /**
    * Handle a thread interruption or database connection error specially,
    * otherwise return normally.
    * 
    * @param   exc
    *          Exception to be inspected.
    * 
    * @see     DBUtils#handleException(Database, Exception)
    */
   private void handleException(Exception exc)
   {
      if (database == null && dmo instanceof DataModelObject)
      {
         RecordBuffer buffer = RecordBuffer.get((DataModelObject) dmo);
         database = buffer.getDatabase();
      }
      
      DBUtils.handleException(database, exc);
   }

   /**
    * Conversion of EXTENT attribute (KW_EXTENT).
    *
    * Getter of EXTENT attribute
    *
    */
   public integer extent()
   {
      if (sizer == null)
      {
         return new integer(0);
      }

      try
      {
         return (integer) sizer.invoke(dmo);
      }
      catch (IllegalAccessException | InvocationTargetException e)
      {
         throw new RuntimeException(e);
      }
   }

   /**
    * Conversion of INITIAL attribute (KW_INIT).
    *
    * Getter of INITIAL attribute.
    *
    * The value of the INITIAL schema field (which is always CHARACTER), formatted with the
    * buffer-field's format. If the INITIAL schema field has the Unknown value (?), the
    * value of the INITIAL attribute is the null string ("").
    *
    * @return  Current value of INITIAL attribute.
    */
   public character initial()
   {
      character initial = character.valueOf(TableMapper.getLegacyFieldInitial(getDMOProxy(),
                                                                              property));
      Class<?> cls = getType();
      BaseDataType baseDataType;

      try
      {
         Class[] parmTypes = { character.class };
         Constructor ctor = cls.getConstructor(parmTypes);
         baseDataType = (BaseDataType) ctor.newInstance(initial);
      }
      catch (Exception exc)
      {
         throw new RuntimeException(exc);
      }

      return character.valueOf(baseDataType.toString(getFormat().getValue()));
   }

   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue()
   {
      if (sizer == null)
      {
         return character.valueOf(value().toString(getFormat().getValue()));
      }
      else
      {
         ErrorManager.recordOrShowError(26, "Array subscript 0 is out of range", false);

         String msg = "Unable to extract BUFFER-VALUE for field %s";
         String err = String.format(msg, property);
         ErrorManager.recordOrShowError(7366, err, false);
         
         return new character();
      }
   }

   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @param   index
    *          An integer expression representing a subscript, for fields that
    *          have extents.
    *
    * @return  current value of STRING-VALUE attribute.
    */
   public character stringValue(int index)
   {
      return stringValue(new integer(index));
   }

   /**
    * Conversion of STRING-VALUE attribute (KW_STR_VAL).
    *
    * Getter of STRING-VALUE attribute.
    *
    * The STRING-VALUE( ) method uses the format attribute to convert the buffer value to
    * a string.
    *
    * @param    index
    *           An integer expression representing a subscript, for fields that
    *           have extents.
    *
    * @return   Current value of STRING-VALUE attribute.
    */
   public character stringValue(NumberType index)
   {
      return character.valueOf(value(index).toString(getFormat().getValue()));
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  Current value of a buffer-field object.
    */
   public BaseDataType value(int index)
   {
      return value(new integer(index));
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    *
    * @return  Current value of a buffer-field object if the object is an extent field or
    *          <code>unknown</code> if it is a scalar one.
    */
   public BaseDataType value(NumberType index)
   {
      if (!validateBufferValueIndex(index))
      {
         return unknownValue(getter);
      }
      
      return (index.intValue() == 0 ? value() : getElementReference(index).value());
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(int index, Object newValue)
   {
      changeValue(new integer(index), newValue);
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   index
    *          An INTEGER expression representing a subscript, for fields
    *          that have extents.
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(NumberType index, Object newValue)
   {
      if (!this.validateBufferValueIndex(index))
      {
         return;
      }
      
      if (index.intValue() == 0)
      {
         changeValue(newValue);
      }
      else
      {
         getElementReference(index).changeValue(newValue);
      }
   }
   
   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Getter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the new value.
    *
    * @return  Current value of a buffer-field object. If the field is an extent field,
    *          <code>unknown</code> value is returned.
    */
   public BaseDataType value()
   {
      return sizer == null ? get() : unknownValue(getter);
   }

   /**
    * Conversion of BUFFER-VALUE attribute (KW_BUF_VAL).
    *
    * Setter of current value of a buffer-field object. If you modify the
    * BUFFER-VALUE attribute, Progress sets the buffer-field object to the
    * new value.
    *
    * @param   newValue
    *          New value of a buffer-field object.
    *
    */
   public void changeValue(Object newValue)
   {
      BaseDataType bdt = null;
      
      if (!(newValue instanceof BaseDataType))
      {
         // try some simple conversions
         if (newValue instanceof Integer)
         {
            bdt = new integer((Integer) newValue);
         }
         else if (newValue instanceof Long)
         {
            bdt = new integer((Long) newValue);
         }
         else if (newValue instanceof Double)
         {
            bdt = new decimal((Double) newValue);
         }
         else if (newValue instanceof Boolean)
         {
            bdt = new logical((Boolean) newValue);
         }
         else if (newValue instanceof String)
         {
            bdt = new character((String) newValue);
         }
         else
         {
            throw new ClassCastException("Can't convert " +
                                         newValue.getClass().getName() +
                                         " to BaseDataType!");
         }
      }
      else if (newValue instanceof unknown)
      {
         bdt = unknownValue(getter);
      }
      else if (!getType().isAssignableFrom(newValue.getClass()))
      {
         try
         {
            // try to invoke a polymorphic constructor of the field's type
            Constructor<?> ctor = getType().getConstructor(BaseDataType.class);
            bdt = (BaseDataType) ctor.newInstance(newValue);
         }
         catch (InvocationTargetException exc)
         {
            // the constructor could not manage to convert the passed value into a meaningful 
            // object of the field's type
            Throwable cause = exc.getCause();
            if (cause instanceof ConditionException)
            {
               throw (ConditionException) cause;
            }
            
            LOG.log(Level.SEVERE, "Error performing runtime type conversion", exc);
            
            return;
         }
         catch (NoSuchMethodException  |
                InstantiationException |
                IllegalAccessException |
                IllegalArgumentException exc)
         {
            ErrorManager.recordOrThrowError(
               5729,
               "Incompatible datatypes found during runtime conversion");
            return;
         }
      }
      else
      {
         // finally...the simple case
         bdt = (BaseDataType) newValue;
      }
      
      set(bdt);
   }
   
   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.
    *
    * @return  current value of COLUMN-LABEL attribute.
    */
   public character columnLabel()
   {
      return character.valueOf(TableMapper.getLegacyFieldColumnLabel(getDMOProxy(), property));
   }

   /**
    * Conversion of COLUMN-LABEL attribute (KW_COL_LAB).
    *
    * Getter of COLUMN-LABEL attribute.

    * @param   columnLabel
    *          New value of a attribute.
    */
   public void changeColumnLabel(character columnLabel)
   {
      setColumnLabel(columnLabel);
   }

   /**
    * Conversion of BUFFER-NAME attribute (KW_BUF_NAME).
    *
    * Gets the name of the buffer object to which the buffer-field object belongs.
    *
    * @return  name of the buffer object to which the buffer-field object belongs.
    */

   public character bufferName()
   {
      return character.valueOf(getParentBuffer().getDMOAlias());
   }

   /**
    * Conversion of LITERAL-QUESTION attribute (KW_LIT_QSTN).
    *
    * Getter of LITERAL-QUESTION attribute.
    *
    * @return  current value of LITERAL-QUESTION attribute.
    */
   public logical literalQuestion()
   {
      return new logical(TableMapper.getLegacyFieldLiteralQuestion(getDMOProxy(), property));
   }

   /**
    * Conversion of LITERAL-QUESTION attribute (KW_LIT_QSTN).
    *
    * Setter of LITERAL-QUESTION attribute.
    *
    * @param   literalQuestion
    *          New value of LITERAL-QUESTION attribute.
    */
   public void changeLiteralQuestion(logical literalQuestion)
   {
      TableMapper.setLegacyFieldLiteralQuestion(getDMOProxy(), property, literalQuestion);
   }

   /**
    * Conversion of VALIDATE-MESSAGE attribute (KW_VAL_MSG).
    *
    * Getter of VALIDATE-MESSAGE attribute.
    *
    * @return  current value of VALIDATE-MESSAGE attribute.
    */
   public character validateMessage()
   {
         return new character(TableMapper.getLegacyFieldValidateMessage(getDMOProxy(), property));
   }

   /**
    * Conversion of VALIDATE-MESSAGE attribute (KW_VAL_MSG).
    *
    * Setter of VALIDATE-MESSAGE attribute.
    *
    * @param   validateMessage
    *          New value of VALIDATE-MESSAGE attribute.
    */
   public void changeValidateMessage(character validateMessage)
   {
      TableMapper.setLegacyFieldValidateMessage(getDMOProxy(),
                                                property,
                                                validateMessage.getValue());
   }

   /**
    * Conversion of MANDATORY attribute (KW_MAND).
    *
    * Getter of KW_MAND attribute.
    *
    * @return current value of KW_MAND attribute.
    */
   public logical mandatory()
   {
      return new logical(TableMapper.isLegacyFieldMandatory(getDMOProxy(), property));
   }

   /**
    * Conversion of VALIDATE-EXPRESSION attribute (KW_VAL_EXPR).
    *
    * Getter of VALIDATE-EXPRESSION attribute.
    *
    * @return current value of VALIDATE-EXPRESSION attribute.
    */
   public character validateExpression()
   {
      return new character(TableMapper.getLegacyFieldValidateExpression(getDMOProxy(),
                                                                        property));
   }

   /**
    * Conversion of VALIDATE-EXPRESSION attribute (KW_VAL_EXPR).
    *
    * Setter of VALIDATE-EXPRESSION attribute.
    *
    * @param  validateExpression
    *         New value of VALIDATE-EXPRESSION attribute.
    */
   public void changeValidateExpression(character validateExpression)
   {
      TableMapper.setLegacyFieldValidateExpression(getDMOProxy(),
                                                   property,
                                                   validateExpression.getValue());
   }

   /**
    * Reports if this object is valid for use.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return getDMOProxy().valid();
   }

   /**
    * Conversion of POSITION attribute (KW_POS).
    *
    * Getter of POSITION attribute.
    *
    * @return current value of POSITION attribute.
    */
   public integer position()
   {
      integer position = TableMapper.getLegacyFieldId(getDMOProxy(), property);
      position.increment();
      return position;
   }

   /**
    * Returns the state of the DBNAME attribute
    *
    * @return    Current value of the DBNAME attribute.
    */
   public character getDbName()
   {
      return new character(getParentBuffer().getDatabase().getName());
   }

   /**
    * Returns the state of the TABLE attribute
    *
    * @return    Current value of the TABLE attribute.
    */
   public character getTable()
   {
      return new character(getParentBuffer().getTable());
   }
   
   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public decimal getHeightChars()
   {
      invalidAttribute("HEIGHT-CHARS");

      return new decimal();
   }

   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(NumberType h)
   {
      readOnlyError("HEIGHT");
   }

   /**
    * HEIGHT-CHARS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    * 
    * @param   h
    *          Not used.
    */
   public void setHeightChars(double h)
   {
      readOnlyError("HEIGHT");
   }
   
   /**
    * Get the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * For buffer-field objects, the WIDTH-CHARS attribute is the number of characters in the
    * STRING-VALUE, which the AVM calculates using the FORMAT attribute.
    *
    * @return  See above.
    */
   public decimal getWidthChars()
   {
      String strValue = stringValue().getValue();
      return new decimal(strValue == null ? 0 : strValue.length());
   }

   /**
    * Set the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * The WIDTH-CHARS attribute of a buffer-field is readable but not writeable.
    *
    * @param   widthChars
    *          The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(NumberType widthChars)
   {
      readOnlyError("WIDTH");
   }

   /**
    * Set the WIDTH-CHARS attribute of this BUFFER-FIELD handle.
    *
    * The WIDTH-CHARS attribute of a buffer-field is readable but not writeable.
    * 
    * @param   widthChars
    *          The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(double widthChars)
   {
      readOnlyError("WIDTH");
   }

   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public integer getWidthPixels()
   {
      invalidAttribute("WIDTH-PIXELS");
      
      return new integer();
   }

   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    widthPixels
    *           Not used.
    */
   public void setWidthPixels(NumberType widthPixels)
   {
      readOnlyError("WIDTH-PIXELS");
   }

   /**
    * WIDTH-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    widthPixels
    *           Not used.
    */
   public void setWidthPixels(int widthPixels)
   {
      readOnlyError("WIDTH-PIXELS");
   }

   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #invalidAttribute} to display appropriate error.
    * 
    * @return  unknown value.
    */
   public integer getHeightPixels()
   {
      invalidAttribute("HEIGHT-PIXELS");
      
      return new integer();
   }

   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    heightPixels
    *           Not used.
    */
   public void setHeightPixels(NumberType heightPixels)
   {
      readOnlyError("HEIGHT-PIXELS");
   }

   /**
    * HEIGHT-PIXELS attribute is not defined for BUFFER-FIELD handle. Calls 
    * {@link #readOnlyError} to display appropriate error.
    *
    * @param    heightPixels
    *           Not used.
    */
   public void setHeightPixels(int heightPixels)
   {
      readOnlyError("HEIGHT-PIXELS");
   }

   /**
    * Getter of FORMAT attribute.
    *
    * @return   current value of FORMAT attribute.
    */
   @Override
   public character getFormat()
   {
      return character.valueOf(TableMapper.getLegacyFieldFormat(getDMOProxy(), property));
   }

   /**
    * Setter of FORMAt attribute.
    *
    * @param   format
    *          The new value.
    */
   @Override
   public void setFormat(String format)
   {
      setFormat(new character(format));
   }

   /**
    * Setter of FORMAt attribute.
    *
    * @param   format
    *          The new value.
    */
   @Override
   public void setFormat(character format)
   {
      TableMapper.setLegacyFieldFormat(getDMOProxy(), property, format.getValue());
   }

   /**
    * Getter of COLUMN-LABEL attribute.
    *
    * @return   current value of COLUMN-LABEL attribute.
    */
   public character getColumnLabel()
   {
      return columnLabel();
   }

   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   public void setColumnLabel(character columnLabel)
   {
      setColumnLabel(columnLabel.getValue());
   }

   /**
    * Setter of COLUMN-LABEL attribute.
    * 
    * @param    columnLabel
    *           New value of a attribute.
    */
   public void setColumnLabel(String columnLabel)
   {
      TableMapper.setLegacyFieldColumnLabel(getDMOProxy(), property, columnLabel);
   }
 
   /**
    * Returns the value of the DATA-TYPE attribute
    *
    * @return   Current value of the DATA-TYPE attribute.
    */
   public character getDataType()
   {
      return character.valueOf(getType().getSimpleName());
   }
 
   /**
    * Set the <code>DATA-TYPE</code> attribute..
    *
    * @param    dataType
    *           The DATA-TYPE to be set.
    */
   public void setDataType(String dataType)
   {
      readOnlyError("DATA-TYPE");
   }

   /**
    * Set the <code>DATA-TYPE</code> attribute..
    *
    * @param    dataType
    *           The DATA-TYPE to be set.
    */
   public void setDataType(character dataType)
   {
      readOnlyError("DATA-TYPE");
   }

   /**
    * Returns the value of the LABEL attribute
    *
    * @return    Current value of the LABEL attribute.
    */
   public character getLabel()
   {
      return character.valueOf(TableMapper.getLegacyFieldLabel(getDMOProxy(), property));
   }

   /**
    * Set the <code>LABEL</code> attribute of handle.
    *
    * @param    label
    *           The LABEL to be set.
    */
   public void setLabel(String label)
   {
      TableMapper.setLegacyFieldLabel(getDMOProxy(), property, label);
   }

   /**
    * Set the <code>LABEL</code> attribute of handle.
    *
    * @param    label
    *           The LABEL to be set.
    */
   public void setLabel(character label)
   {
      setLabel(label.getValue());
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }

   /**
    * Delete the resource.
    *
    * BUFFER-FIELD will be deleted only when the dynamic buffer is deleted. No resource-specific
    * delete is required.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      return false;
   }

   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }

   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }

   /**
    * Get the field reference for the given element of this extent field.
    *
    * @param  index
    *         1-based index of the extent field.
    *
    * @return field reference for the specific extent element.
    */
   private FieldReference getElementReference(NumberType index)
   {
      NumberType idx = MathOps.minus(index, 1);
      return new FieldReference((DataModelObject) dmo, property, uppercase, idx);
   }

   /**
    * Get the DMO proxy which references parent buffer.
    *
    * @return  DMO proxy object.
    */
   private Buffer getDMOProxy()
   {
      return (Buffer) getParentBuffer().getDMOProxy();
   }
   
   /**
    * Validate the index passed to a BUFFER-VALUE method call. For scalar fields, this value must
    * be zero. For extent fields, this value must be a positive number less than or equal to the
    * size of the extent field. Unknown value is invalid in either case.
    * 
    * @param   index
    *          Index value (unadjusted) passed to BUFFER-VALUE method.
    * 
    * @return  <code>true</code> if <code>index</code> is valid, else <code>false</code>.
    * 
    * @throws  ErrorConditionException
    *          if <code>index</code> is invalid and session is not in silent error mode.
    */
   private boolean validateBufferValueIndex(NumberType index)
   {
      int max = 0;
      if (sizer != null)
      {
         try
         {
            integer size = (integer) sizer.invoke(dmo);
            max = size.intValue();
         }
         catch (IllegalAccessException | IllegalArgumentException | InvocationTargetException exc)
         {
            if (LOG.isLoggable(Level.FINE))
            {
               LOG.log(Level.FINE, "Error determining DMO field size", exc);
            }
            
            return false;
         }
      }
      
      int i = (index.isUnknown() ? -1 : index.intValue());
      if (i < 0 || i > max || (max == 0 && i != 0))
      {
         String fieldName = TableMapper.getLegacyFieldName(getDMOProxy(), property);
         String msg = String.format("%s BUFFER-VALUE argument must be valid array index, less " +
                                    "than or equal to field extent %d",
                                    fieldName,
                                    max);
         ErrorManager.recordOrThrowError(7365, msg, false);
         
         return false;
      }
      
      return true;
   }
   
   /**
    * The method is used to overcome the limitation of the methods 
    * {@code PropertyHelper#gettersByProperty(Class)} and 
    * {@code PropertyHelper#settersByProperty(Class)} - they both return only the indexed
    * methods for a property. This fetches the non-indexed counterparts for all the indexed 
    * properties in the map {@code propertyMap}. All the entries of the passed in map together 
    * with the non-indexed methods are returned in a new map instance. To distinguish the indexed 
    * and the non-indexed property methods, the property root name for the indexed methods is 
    * concatenated with the string "[]" in the resulting map.
    * 
    * @param   dmoIface
    *          DMO implementation type. 
    * @param   propertyMap
    *          The map of property root names to their respective methods.
    * 
    * @return  The map of property root names to their respective methods.
    */
   private static Map<String, Method> fetchIndexedPropertyMethods(
                                          Class<?> dmoIface, 
                                          Map<String, Method> propertyMap)
   {
      Map<String, Method> result = new HashMap<>();
      
      for (Entry<String, Method> entry : propertyMap.entrySet())
      {
         String propRootName = entry.getKey();
         Method method = entry.getValue();
         Class<?>[] argTypes = method.getParameterTypes();
         
         // is this entry an indexed method? 
         boolean isIndexed = argTypes.length > 0 && Integer.TYPE.equals(argTypes[0]);
         if (isIndexed)
         {
            Method nonIndexedMethod = null;
            
            // indexed - find the non-indexed counterpart - method of the same name
            // but without the index parameter returning or taking an array type
            for (Method m : method.getDeclaringClass().getDeclaredMethods())
            {
               Class<?>[] paramTypes = m.getParameterTypes();
               
               // check for the setter or getter
               if (m.getName().equals(method.getName()) && 
                   (paramTypes.length == 0 || 
                    paramTypes.length == 1 && paramTypes[0].isArray()))
               {
                  nonIndexedMethod = m;
               }
            }
            
            if (nonIndexedMethod == null)
            {
               throw new RuntimeException(String.format(
                     "Non-indexed method %s is expected for the DMO type %s.",
                     method.getName(), dmoIface.getName()));
            }
            
            // add the non-indexed method to the map
            result.put(propRootName, nonIndexedMethod);
            
            // add the indexed method to the map;
            // concatenate the root name and the string "[]" to form
            // the root property name for the indexed method
            result.put(propRootName + "[]", method);
         }
         else
         {
            // add the non-indexed method to the result
            result.put(propRootName, method);
         }
      }
      
      return result;
   }
}
