/*
** Module   : BaseConfig.java
** Abstract : container for the most common widget state
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 EVL 20050715   @21704 Creation based on general concept. Moving
**                           the widget type constants to eliminate the
**                           unusable class WidgetTransferrable.
** 002 EVL 20050729   @21909 To perform automatic layout usage the row
**                           and column default values must be 0.
** 003 SIY 20050831   @22485 Refactoring (see @@22394-22409).
** 004 SIY 20050907   @22621 Fixed use of @inheritDoc tag in javadoc,
**                           Temporary reverting to 0-based coordinates.
** 005 EVL 20050919   @22752 Excluding the widget's name from transferring
**                           through the network.
** 006 SIY 20050920   @22818 Moved setSensitive() below setWidthChars() to
**                           make sure that it will not cause any negative 
**                           side effects, updated support for SENSITIVE
**                           and SELECTED attributes, using ColorTable for 
**                           the colors.
** 007 SIY 20051129   @23552 Changed type of PARENT attribute to 
**                           GenericWidget, changed type of PRIVATE-DATA
**                           to String.
** 008 SIY 20051130   @23586 Added support for ENTERED/NOT ENTERED
**                           property.
** 009 GES 20060116   @23945 Cleanup of some docs and formatting.
** 010 SIY 20060203   @24237 Support for TO option.
** 011 SIY 20060209   @24486 Major refactoring.
** 012 NVS 20060310   @24966 setModified() method also changes the widget
**                           state to CHANGED.
** 013 SIY 20060412   @25595 Encapsulated 'state' field.
** 014 SIY 20060424   @25679 Added getter/setter for parent ID.
** 015 SIY 20060621   @27534 Removed 'transient' specification for name 
**                           attribute.
** 016 SIY 20060628   @27658 Moved 'title' attribute from ContainerConfig. 
** 017 GES 20060803   @28388 Made externalizable for better performance.
**                           Many code standard cleanups.
** 018 SIY 20061102   @30937 Simplified and generalized for TO/COLON/AT 
**                           options.
** 019 SIY 20061103   @30967 Set coordinate when instance is available
**                           even if alignment is not ALIGN_LEFT.
** 020 NVS 20061204   @31550 "selected" flag is no longer part of the
**                           configuration.
** 021 SIY 20070214   @32129 Added isLeftAligned() method.
** 022 NVS 20070426   @33246 SENSITIVE attribute is no longer maintained
**                           in the configuration.
** 023 SIY 20070627   @34277 Do not set DCOLOR to -1.
** 024 SIY 20070627   @34810 Added accessor for true column value.
** 025 SIY 20080429   @38140 Added saveColors()/restoreColors() methods.
** 026 CA  20080905   @39744 Negative coordinates can be received by the 
**                           client, so INV_COORD was changed to a
**                           smaller value. Fixed COLUMN/ROW getters.
** 027 SIY 20090313   @41595 Added commend about INV_COORD value. 
** 028 SIY 20100619          Switching to own implementation of character UI.
** 029 SIY 20100804          Changes in imports, some refactorings.
** 030 SIY 20100822          Refactored AttributeContainer support.
** 031 SIY 20100915          locateFrame() is moved to UiUtils.
** 032 SIY 20110112          Changes related to new color handling scheme.
** 033 SIY 20111101          Moved from CHUI.
** 034 SIY 20111115          Added default font constant.
** 035 SIY 20111221          Minor cleanups.
** 035 MAG 20140410          Add SIZE-PIXELS, FGCOLOR, BGCOLOR, TOOLTIP options.
** 036 MAG 20140512          Enhanced color support.
** 037 CA  20140711          Enhanced font support, for GUI clients.
** 038 CA  20140805          Changes to isolate ChUI-specific code from common code: added 
**                           generics.
** 039 MAG 20140912          Change FGCOLOR and BGCOLOR to integer types.
** 040 HC  20140911          GUI coordinate support - data types refactoring.
** 041 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 042 MAG 20141006          Set default value to -1 for BGCOLOR and FGCOLOR
** 043 HC  20141021          Implemented window move/position support.
** 044 HC  20141207          Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * The container for the most common widget state variables, except some
 * special cases.
 */
public class BaseConfig
extends WidgetConfig
{
   /** 
    * Invalid coordinate value.   
    */
   // Note that changes of constant value must be properly reflected in
   // ScreenBitmap.INV_COORD
   public static final int INV_COORD = Integer.MIN_VALUE;
   
   /** The integer index of the parent in the array of configs. */ 
   public int parentId = -1;
   
   /** The column position of the widget. */
   public double column = INV_COORD;
   
   /** The row position of the widget. */
   public double row = INV_COORD;
   
   /** The x-axis position of the widget in pixel units. */
   public int x = 0;
   
   /** The y-axis position of the widget in pixel units. */
   public int y = 0;
   
   /** The color is used to render the widget. */
   public Color dcolor = null;

   /** The color of the widget under the focus. */
   public Color pfcolor = Color.INPUT;

   /** The height of the widget. */
   public double heightChars = 0;
   
   /** The width of the widget. */
   public double widthChars = 0;

   /** Rectangle width pixels */
   public int widthPixels;
   
   /** Rectangle height pixels */
   public int heightPixels;
   
   /** Set to true if component is created dynamically by the runtime. */
   public boolean dynamic = false;
   
   /** The tab-stop attribute. */
   public boolean tabStop = true;
   
   /** The name attribute. */
   public String name = null;
   
   /** MODIFIED attribute variable. */
   public boolean modified = false;

   /** The title of the container. */
   public String title = "";

   /** The title font of the container. */
   public int titleFont = FontManager.UNDEFINED_FONT;

   /** TITLE-DCOLOR attribute variable. */
   public Color titleDColor = Color.MESSAGES;

   /** Widget alignment. */
   public int align = ControlEntity.ALIGN_LEFT;
   
   /** Font number. Corresponds to FONT attribute. */
   public int font = FontManager.UNDEFINED_FONT;
      
   /** TOOLTIP option */
   public String tooltip;   
         
   /** FGCOLOR option */
   public int fgcolor = -1;

   /** BGCOLOR option */
   public int bgcolor = -1;
   
   /** default FGCOLOR option */
   public int sysfgcolor = -1;

   /** default BGCOLOR option */
   public int sysbgcolor = -1;
   
   /** SELECTED attribute. TODO: its client-side runtime is not implemented. */
   public boolean selected = false;

   /**
    * Default constructor (only used in deserialization).
    */
   public BaseConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected BaseConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected BaseConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Check if widget is left aligned.
    * 
    * @return  <code>true</code> if widget is left aligned.
    */
   public boolean isLeftAligned()
   {
      return align == ControlEntity.ALIGN_LEFT;
   }
   
   /**
    * Sets column coordinate of the upper left corner of the component
    * relative to the parent container.  This is a 0-based column position
    * and NOT the 1-based position used in Progress.
    * 
    * @param    column
    *           The new column value of upper left component corner.
    */
   public void setColumn(double column)
   {
      column = Coordinate.scale(column);
      
      this.column = column; // direct access
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }
      
      Point location = instance.location();
      location.x = (isLeftAligned()) ? column : BaseConfig.INV_COORD;
      instance.setLocation(location);
   }
   
   /**
    * Sets value of the ROW attribute.  This is a 0-based row position
    * and NOT the 1-based position used in Progress.
    * 
    * @param    row
    *           The new row value for the upper left component corner relative
    *           to it's parent.
    */
   public void setRow(double row)
   {
      row = Coordinate.scale(row);
      
      this.row = row; // direct access
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }

      Point location = instance.location();
      location.y = row;
      instance.setLocation(location);
   }
   
   /**
    * Sets X coordinate of the upper left corner of the component
    * relative to the parent container.
    * 
    * @param    x
    *           0-based pixel value.
    */
   public void setX(int x)
   {
      this.x = x; // direct access
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }

      NativePoint location = instance.physicalLocation();
      location.x = x;
      instance.setPhysicalLocation(location.x, location.y);
   }
   
   /**
    * Sets Y coordinate of the upper left corner of the component
    * relative to the parent container.
    * 
    * @param    y
    *           0-based pixel value.
    */
   public void setY(int y)
   {
      this.y = y; // direct access
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }

      NativePoint location = instance.physicalLocation();
      location.y = y;
      instance.setPhysicalLocation(location.x, location.y);
   }

   /**
    * Sets HEIGHT-CHARS Progress attribute.
    * 
    * @param    heightChars
    *           The integer value for the new HEIGHT-CHARS value.
    */
   public void setHeightChars(double heightChars)
   {
      heightChars = Coordinate.scale(heightChars);
      
      if (heightChars == 0)
         return;

      this.heightChars = heightChars;
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }

      if (instance instanceof Container)
      {
         ((Container<?>) instance).setSize(widthChars, heightChars);
      }
   }
   
   /**
    * Set the WIDTH-CHARS Progress attribute.
    * 
    * @param    widthChars
    *           The integer value for the new WIDTH-CHARS value.
    */
   public void setWidthChars(double widthChars)
   {
      widthChars = Coordinate.scale(widthChars);
      
      this.widthChars = widthChars;
      
      if (id == null || !onClientSide())
      {
         return;
      }
   
      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }
   
      if (instance instanceof Container)
      {
         ((Container<?>) instance).setSize(widthChars, heightChars);
      }
      else if (instance instanceof Label)
      {
         ((Label<?>) instance).setWidth(widthChars);
      }
   }
   
   /**
    * Get COLON attribute value. Note that Progress does not have
    * corresponding attribute accessible to program.
    * 
    * @return   Current attribute value.
    */
   public double getColon()
   {
      if (align == ControlEntity.ALIGN_COLON)
         return column; // direct access
      
      return BaseConfig.INV_COORD;
   }

   /**
    * Set COLON attribute value. Note that Progress does not have
    * corresponding attribute accessible to program.
    * 
    * @param    colon
    *           New value for the attribute.
    */
   public void setColon(double colon)
   {
      colon = Coordinate.scale(colon);
      
      if (colon == BaseConfig.INV_COORD)
         return;
      
      this.column = colon; // direct access
      this.align = ControlEntity.ALIGN_COLON;
   }

   /**
    * Sets the rightmost column of the widget.
    * 
    * @param    pos
    *           The 0-based column number.
    */
   public void setTo(double pos)
   {
      pos = Coordinate.scale(pos);
      
      if (pos == BaseConfig.INV_COORD)
         return;
      
      this.column = pos; // direct access
      this.align = ControlEntity.ALIGN_RIGHT;
   }
      
   /**
    * Get the rightmost column of the widget.
    *
    * @return   Current value of the attribute. 
    */
   public double getTo()
   {
      if (align == ControlEntity.ALIGN_RIGHT)
         return column; // direct access
      
      return BaseConfig.INV_COORD;
   }

   /**
    * Set new value to the MODIFIED attribute.
    * 
    * @param    modified
    *           New value for the attribute.
    */
   public void setModified(boolean modified)
   {
      this.modified = modified;
      this.state = ScreenBuffer.CHANGED;
   }
   
   /**
    * Synchronize the configuration with the widget state.
    */
   @Override
   public void syncWithWidget()
   {
      super.syncWithWidget();
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);
      
      if (instance == null)
      {
         return;
      }

      if (name != null)
      {
         instance.setName(name);
      }
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      BaseConfig cfg = (BaseConfig) config;
      
      parentId = cfg.parentId;
      setColumn(cfg.column);
      setRow(cfg.row);
      x = cfg.x;
      y = cfg.y;
      dcolor = cfg.dcolor;
      pfcolor = cfg.pfcolor;
      setHeightChars(cfg.heightChars);
      setWidthChars(cfg.widthChars);
      widthPixels = cfg.widthPixels;
      heightPixels = cfg.heightPixels;
      dynamic = cfg.dynamic;
      tabStop = cfg.tabStop;
      name = cfg.name;
      setModified(cfg.modified);
      title = cfg.title;
      titleFont = cfg.titleFont;
      titleDColor = cfg.titleDColor;
      align = cfg.align;
      font = cfg.font;
      tooltip = cfg.tooltip;
      fgcolor = cfg.fgcolor;
      bgcolor = cfg.bgcolor;
      sysfgcolor = cfg.sysfgcolor;
      sysbgcolor = cfg.sysbgcolor;
      selected = cfg.selected;
   }
   
   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      parentId     = in.readInt();
      column       = in.readDouble();
      row          = in.readDouble();
      x            = in.readInt();
      y            = in.readInt();
      font         = in.readInt();
      dcolor       = (Color) in.readObject();
      pfcolor      = (Color) in.readObject();
      fgcolor      = in.readInt();
      bgcolor      = in.readInt();
      heightChars  = in.readDouble();
      widthChars   = in.readDouble();
      widthPixels  = in.readInt();
      heightPixels = in.readInt();
      align        = in.readInt();
      dynamic      = in.readBoolean();
      tabStop      = in.readBoolean();
      name         = (String) in.readObject();
      title        = (String) in.readObject();
      tooltip      = (String) in.readObject();
      titleFont    = in.readInt();
      titleDColor  = (Color) in.readObject();
      modified     = in.readBoolean();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeInt(parentId);
      out.writeDouble(column);
      out.writeDouble(row);
      out.writeInt(x);
      out.writeInt(y);;
      out.writeInt(font);
      out.writeObject(dcolor);
      out.writeObject(pfcolor);
      out.writeInt(fgcolor);
      out.writeInt(bgcolor);
      out.writeDouble(heightChars);
      out.writeDouble(widthChars);
      out.writeInt(widthPixels);
      out.writeInt(heightPixels);
      out.writeInt(align);
      out.writeBoolean(dynamic);
      out.writeBoolean(tabStop);
      out.writeObject(name);
      out.writeObject(title);
      out.writeObject(tooltip);
      out.writeInt(titleFont);
      out.writeObject(titleDColor);
      out.writeBoolean(modified);
   }
}
