/*
** Module   : BaseEntity.java
** Abstract : server side base class for most widgets
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23479 Created initial version.
** 002 GES 20060116   @23947 Implemented conversion of 1-based to 0-based positions.
** 003 SIY 20060209   @24432 Added missing method for width setting, other minor changes.
** 004 SIY 20060424   @25680 Changed setParent() signature.
** 005 GES 20060803   @28363 Major cleanup and refactoring.
** 006 NVS 20061204   @31549 SELECTED attribute is server side only from now on.
** 007 NVS 20070426   @33241 SENSITIVE attribute is no longer maintained separately, but is backed
**                           up with ENABLED flag of CHARVA on the client.
** 008 NVS 20080307   @37409 setRow() and setColumn() methods now support dynamic settings that is
**                           required for ROW and COL widget attributes support.
** 009 GES 20080319   @37579 Added NEXT-CHILD and PREV-CHILD support.
** 010 CA  20080925   @39960 Support for dynamic browse title - base field is now protected.
** 011 NVS 20081211   @40872 getParent() method takes into account field group the widget is in.
** 012 GES 20081230   @41001 Return a NullWidget instance instead of null for safety.
** 013 SIY 20101004          Added necessary method stubs.
** 014 SIY 20101005          Generalized setX/setY.
** 015 SVL 20101006          Added FGCOLOR, BGCOLOR attributes.
** 016 SIY 20101007          Use NumberType instead of decimal.
** 017 SIY 20110112          Changes related to new color handling scheme.
** 018 SIY 20111101          Updated imports.
** 019 CA  20130112          getNextSibling and getPrevSibling must return a handle instance.
** 020 CA  20130222          Added NPE protection for cases when a widget is used before is 
**                           attached to a frame (after a CREATE WIDGET statement is used).
** 021 CA  20130323          The parent must be a CommonWidget, not Object.
** 022 OM  20130527          Added int64 support for colors.
** 023 CA  20130927          Fixed getNext/PrevSibling - NullWidget is not needed.
** 024 CA  20131115          Changes related to properly instantiate dynamic resources.
** 025 HC  20131215          Removed PRIVATE-DATA attribute, it is already implemented 
**                           in the base class.
** 026 MAG 20140410          Implements FGCOLOR, BGCOLOR, TOOLTIP and SIZE-PIXELS attributes.           
** 027 VIG 20140314          Changed constructor signature, adding new "before" semantic in
**                           the logic of inserting new BaseEntity to HandleChain.
**                           Added possibility to use HandleChain siblings methods.
** 028 MAG 20140508          Set FGCOLOR, BGCOLOR using integer values.           
** 029 MAG 20140512          Enhanced color support.           
** 030 CA  20140711          Enhanced font support, for GUI clients.
** 031 MAG 20140912          Change FGCOLOR and BGCOLOR to integer types.
** 032 HC  20140911          GUI coordinate support - data types refactoring.
** 033 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 034 CA  20141104          Removed setHeight/Width: all widgets need to use the explicit char or
**                           pixel unit APIs.
** 035 HC  20141021          Implemented window move/position support.
** 036 EVL 20141111          Adding getter methods for DCOLOR, PFCOLOR attributes.
** 037 HC  20141207          Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * The server side base class for most widgets. 
 */
public abstract class BaseEntity<T extends BaseConfig>
extends GenericWidget<T>
{
   /** PARENT attribute */
   private CommonWidget parent = null;
   
   /** Group container for the widget and its siblings. */
   protected Group group = null;

   /** Saved DCOLOR attribute. */
   private Color saveDColor = null;

   /** Saved PFCLOR attribute. */
   private Color savePfColor = null;

   /** Flag which is set during color save and reset during color restore. */
   private boolean colorsSaved = false;

   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    base
    *           The container for all state to be transferred to the client.
    */
   public BaseEntity(boolean dynamic, T base)
   {
      super(dynamic, base);
   }
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    base
    *           The container for all state to be transferred to the client.
    * @param    before
    *           The right neighbor of the BaseEntity been created or null if new BaseEntity
    *           will be the last in the chain. If <b>before</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    * @param    firstAsLast
    *           The flag indicating that this new entity must be inserted as last element 
    *           in the chain in the case when <b>before</b> - is the first element of the chain.
    *           When <b>before</b> is the first element of the chain, default behavior (expected
    *           from the name of this parameter is to insert new element as the first. 
    *           This flag allows to change the expected behavior of <b>before</b> parameter when 
    *           it needs and makes sense only if <b>before</b> is the first element.
    */
   public BaseEntity(boolean dynamic, T base, BaseEntity<T> before, boolean firstAsLast)
   {
      super(dynamic, base, before, firstAsLast);
   }
   
   /**
    * Restore configured colors if there was previously saved ones.
    */
   @Override
   public void restoreColors()
   {
      if (colorsSaved)
      {
         config.dcolor  = saveDColor;
         config.pfcolor = savePfColor;
         colorsSaved = false;
      }
   }

   /**
    * Save configured colors.
    */
   @Override
   public void saveColors()
   {
      saveDColor  = config.dcolor;
      savePfColor = config.pfcolor;
      colorsSaved = true;
   }

   /**
    * Gets the COLUMN writable attribute.
    *
    * @return   The 1-based column number at which to start displaying the 
    *           widget.  This is converted from a 0-based column number for
    *           the internal runtime.
    */
   @Override
   public decimal getColumn()
   {
      double col = config.column < 0 ? config.column - 1 : config.column + 1;
      return new decimal(col);
   }

   /**
    * Sets the column position for this widget.
    *
    * @param    c
    *           The 1-based row number at which to start displaying the 
    *           widget.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(NumberType c)
   {
      boolean unknown = c.isUnknown();
      setColumnOrRow(unknown, unknown ? 0 : c.doubleValue(), false);
   }

   /**
    * Sets the column position for this widget.
    *
    * @param    c
    *           The 1-based row number at which to start displaying the 
    *           widget.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(double c)
   {
      setColumnOrRow(false, c, false);
   }
   
   /**
    * Gets the display color writable attribute.
    *
    * @return   The current value of the display color attribute.
    */
   @Override
   public integer getDColor()
   {
      int color = config.dcolor.color();
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the display color.
    *
    * @param    param
    *           The display color.
    */
   public void setDColor(NumberType param)
   {
      config.dcolor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the display color.
    * 
    * @param    displayColor
    *           The display color.
    */
   public void setDColor(Color displayColor)
   {
      config.dcolor = displayColor;
      pushScreenDefinition();
   }

   /**
    * Set widget DCOLOR
    * 
    * @param   color
    *          Integer color value.
    */
   public void setDColor(int color)
   {
      setDColor(new ColorSpec(color).convert());
   }
   
   /**
    * Gets the FGCOLOR writable attribute.
    *
    * @return   The current value of the FGCOLOR attribute.
    */
   public integer getFgColor()
   {
      int color = config.fgcolor;
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   @Override
   public void setFgColor(integer fgcolor)
   {
      int color = (fgcolor == null || fgcolor.isUnknown()) ? -1 : fgcolor.intValue();
      setFgColor(color);
   }   

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   public void setFgColor(int64 fgcolor)
   {
      int color = (fgcolor == null || fgcolor.isUnknown()) ? -1 : fgcolor.intValue(); 
      setFgColor(color);
   }

   /**
    * Sets the FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the FGCOLOR attribute.
    */
   public void setFgColor(int fgcolor)
   {
      config.fgcolor = fgcolor;
      pushScreenDefinition();            
   }

   /**
    * Gets the BGCOLOR writable attribute.
    *
    * @return   The current value of the BGCOLOR attribute.
    */
   public integer getBgColor()
   {
      int color = config.bgcolor;
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   @Override
   public void setBgColor(integer bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue();
      setBgColor(color);
   }

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   public void setBgColor(int64 bgcolor)
   {
      int color = (bgcolor == null || bgcolor.isUnknown()) ? -1 : bgcolor.intValue(); 
      setBgColor(color);
   }

   /**
    * Sets the BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the BGCOLOR attribute.
    */
   public void setBgColor(int bgcolor)
   {
      config.bgcolor = bgcolor;
      pushScreenDefinition();            
   }
   
   /**
    * Gets the HEIGHT-CHARS writable attribute.
    *
    * @return   The current value of the HEIGHT-CHARS attribute.
    */
   @Override
   public decimal getHeightChars()
   {
      return new decimal(config.heightChars);
   }

   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   public void setHeightChars(NumberType height)
   {
      boolean unknown = height.isUnknown();
      setWidthOrHeight(unknown, unknown ? 0 : height.doubleValue(), false);
   }
   
   /**
    * Sets the HEIGHT-CHARS writable attribute.
    *
    * @param    height
    *           The new value for the HEIGHT-CHARS attribute.
    */
   public void setHeightChars(double height)
   {
      setWidthOrHeight(false, height, false);
   }
   
   /**
    * Gets the WIDTH-CHARS writable attribute.
    *
    * @return   The current value of the WIDTH-CHARS attribute.
    */
   public decimal getWidthChars()
   {
      return new decimal(config.widthChars);
   }

   /**
    * Sets the WIDTH-CHARS writable attribute.
    *
    * @param    width
    *           The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(NumberType width)
   {
      boolean unknown = width.isUnknown();
      setWidthOrHeight(unknown, unknown ? 0 : width.doubleValue(), true);
   }
   
   /**
    * Sets the WIDTH-CHARS writable attribute.
    *
    * @param    width
    *           The new value for the WIDTH-CHARS attribute.
    */
   public void setWidthChars(double width)
   {
      setWidthOrHeight(false, width, true);
   }
   
   /**
    * Get rectangle height pixels.
    * SIZE-PIXELS option.
    * 
    * @return  The height of the widget in pixels.
    */
   public integer getHeightPixels()
   {
      return new integer(config.heightPixels);
   }   

   /**
    * Get rectangle width pixels.
    * SIZE-PIXELS option.
    * 
    * @return  The width of the widget in pixels.
    */
   public integer getWidthPixels()
   {
      return new integer(config.widthPixels);
   }   
   
   /**
    * Gets the PARENT writable attribute.
    * If the parent attribute was set explicitly, it is returned.
    * Otherwise, the field group of this widget is returned as parent.
    * For the regular widgets, it always has a value. It's the default field
    * group. Non-regular widgets override this method to return a different
    * result.  
    *
    * @return   the parent widget for this widget
    */
   @Override
   public CommonWidget getParent()
   {
      // explicitly set parent attribute
      if (parent != null)
      {
         return parent;
      }

      // normal case
      return group;
   }

   /**
    * Sets the PARENT writable attribute.
    *
    * @param    widget
    *           The new value for the PARENT attribute.
    */
   public void setParent(GenericWidget<?> widget)
   {
      this.parent = widget;
      config.parentId = widget.getId();
      pushScreenDefinition();
   }

   /**
    * Establishes a relationship between the container and the elements in
    * a given widget group.
    *
    * @param    group
    *           The container of the widget.
    */
   @Override
   public void setGroup(Group group)
   {
      this.group = group;
   }
   
   /**
    * Obtains the next sibling widget in the current field group.
    *
    * @return   The next sibling or an unknown handle if this is the only widget or if this is the 
    *           last widget.
    */
   @Override
   public handle getNextSibling()
   {
      if (useHandleChainSiblings())
      {
         return super.getNextSibling();
      }
      else
      {
         GenericWidget<?> widget = (group == null) ? null : group.getNextChild(this);
         return new handle(widget);
      }
   }
   
   /**
    * Obtains the previous sibling widget in the current field group.
    *
    * @return   The previous sibling or an unknown handle if this only widget or if this is the 
    *           first widget.
    */
   @Override
   public handle getPrevSibling()
   {
      if (useHandleChainSiblings())
      {
         return super.getPrevSibling();
      }
      else
      {
         GenericWidget<?> widget = (group == null) ? null : group.getPrevChild(this);
         return new handle(widget);
      }
   }

   /**
    * Gets the input (prompt/edit mode) color writable attribute.
    *
    * @return   The current value of the input (prompt/edit mode) color attribute.
    */
   @Override
   public integer getPfColor()
   {
      int color = config.pfcolor.color();
      return (color == -1) ? new integer() : new integer(color);
   }
   
   /**
    * Sets the input (prompt/edit mode) color.
    *
    * @param    param
    *           The color to use.
    */
   public void setPfColor(NumberType param)
   {
      config.pfcolor = new ColorSpec(param.intValue()).convert();
      pushScreenDefinition();
   }

   /**
    * Sets the input (prompt/edit mode) color.
    * 
    * @param    inputColor
    *           The color to use for when in "input" mode.
    */
   public void setPfColor(Color inputColor)
   {
      config.pfcolor = inputColor;
      pushScreenDefinition();
   }

   /**
    * Set widget PFCOLOR
    * 
    * @param   color
    *          Integer color value.
    */
   public void setPfColor(int color)
   {
      setPfColor(new ColorSpec(color).convert());
   }
   
   /**
    * Gets the ROW writable attribute.
    *
    * @return   The 1-based row number at which to start displaying the 
    *           widget.  This is converted from a 0-based row number for the
    *           internal runtime.
    */
   @Override
   public decimal getRow()
   {
      double row = config.row < 0 ? config.row - 1 : config.row + 1;
      return new decimal(row);
   }

   /**
    * Sets the row position for this widget.
    *
    * @param    r
    *           The 1-based row number at which to start displaying the 
    *           widget.  This is converted to a 0-based row number for the
    *           internal runtime.
    */
   @Override
   public void setRow(NumberType r)
   {
      boolean unknown = r.isUnknown();
      setColumnOrRow(unknown, unknown ? 0 : r.doubleValue(), true);
   }

   /**
    * Sets the row position for this widget.
    *
    * @param    r
    *           The 1-based row number at which to start displaying the 
    *           widget.  This is converted to a 0-based row number for the
    *           internal runtime.
    */
   @Override
   public void setRow(double r)
   {
      setColumnOrRow(false, r, true);   
   }

   /**
    * Gets the SELECTED writable attribute.
    *
    * @return    The current value of the SELECTED attribute.
    */
   @Override
   public logical isSelected()
   {
      return new logical(config.selected);
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(logical sel)
   {
      // TODO: validate this

      config.selected = sel.booleanValue();
   }

   /**
    * Sets the SELECTED writable attribute.
    *
    * @param    sel
    *           The new value for the SELECTED attribute.
    */
   @Override
   public void setSelected(boolean sel)
   {
      config.selected = sel;
   }
   
   /**
    * Gets the TITLE writable attribute.
    *
    * @return   The current value of the TITLE attribute.
    */
   @Override
   public character getTitle()
   {
      return new character(config.title);
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   @Override
   public void setTitle(character title)
   {
      setTitle(title.isUnknown() ? "" : title.toStringMessage());
   }

   /**
    * Sets the TITLE writable attribute.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   @Override
   public void setTitle(String title)
   {
      config.title = title;
      pushScreenDefinition();
   }

   /**
    * Get TOOLTIP attribute.
    *
    * @return   The current value of the TOOLTIP attribute.
    */
   @Override
   public character getTooltip()
   {
      return new character(config.tooltip);
   }

   /**
    * Gets the X writable attribute.
    *
    * @return   The current value of the X attribute.
    */
   @Override
   public integer getX()
   {
      return new integer(config.x);
   }

   /**
    * Sets the X writable attribute.
    *
    * @param    x
    *           The new value for the X attribute.
    */
   public void setX(NumberType x)
   {
      boolean unknown = x.isUnknown();
      setXorY(unknown, unknown ? 0 : x.intValue(), true);
   }
   
   /**
    * Sets the X writable attribute.
    *
    * @param    x
    *           The new value for the X attribute.
    */
   @Override
   public void setX(int64 x)
   {
      setX((NumberType) x);
   }

   /**
    * Sets the X writable attribute.
    *
    * @param    x
    *           The new value for the X attribute.
    */
   public void setX(int x)
   {
      setXorY(false, x, true);
   }
   
   /**
    * Gets the Y writable attribute.
    *
    * @return    The current value of the Y attribute.
    */
   @Override
   public integer getY()
   {
      return new integer(config.y);
   }

   /**
    * Sets the Y writable attribute.
    *
    * @param    y
    *           The new value for the Y attribute.
    */
   public void setY(NumberType y)
   {
      boolean unknown = y.isUnknown();
      setXorY(unknown, unknown ? 0 : y.intValue(), false);
   }
   
   /**
    * Sets the Y writable attribute.
    *
    * @param    y
    *           The new value for the Y attribute.
    */
   @Override
   public void setY(int64 y)
   {
      setY((NumberType) y);
   }

   /**
    * Sets the Y writable attribute.
    *
    * @param    y
    *           The new value for the Y attribute.
    */
   public void setY(int y)
   {
      setXorY(false, y, false);
   }

   /**
    * Get height.
    * 
    * @return  widget height. 
    */
   @Override
   public decimal getHeight()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Get width.
    * 
    * @return  widget width. 
    */
   @Override
   public decimal getWidth()
   {
      throw new UnsupportedOperationException("Not implemented");
   }

   /**
    * Set new value of TOOLTIP attribute.
    *
    * @param tooltip
    *        new value of TOOLTIP attribute
    */
   public void setTooltip(String tooltip)
   {
      config.tooltip = tooltip;
      pushScreenDefinition();      
   }   
   
   /**
    * Sets the WIDTH-PIXELS writable attribute.
    *
    * @param    widthPixels
    *           The new value for the WIDTH-PIXELS attribute.
    */
   public void setWidthPixels(NumberType widthPixels)
   {
      boolean unknown = widthPixels.isUnknown();
      setWidthOrHeightPixels(unknown, unknown ? 0 : widthPixels.intValue(), true);
   }
   
   /**
    * Sets the WIDTH-PIXELS writable attribute.
    *
    * @param    widthPixels
    *           The new value for the WIDTH-PIXELS attribute.
    */
   public void setWidthPixels(int widthPixels)
   {
      setWidthOrHeightPixels(false, widthPixels, true);
   }
   
   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   public void setHeightPixels(NumberType heightPixels)
   {
      boolean unknown = heightPixels.isUnknown();
      setWidthOrHeightPixels(unknown, unknown ? 0 : heightPixels.intValue(), false);
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param heightPixels
    *        new value of HEIGHT-PIXELS attribute
    */
   public void setHeightPixels(int heightPixels)
   {
      setWidthOrHeightPixels(false, heightPixels, false);
   }

   /**
    * Set the FONT attribute of this widget.
    * 
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.font = font;
      pushScreenDefinition();
   }
   
   /**
    * Get the FONT attribute of this widget.
    * 
    * @return   A value representing an entry in the user's font-table.
    */
   @Override
   public integer getFont()
   {
      int font = config.font;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Indicates an implementation of siblings methods. Child classes can override
    * this method to specify the usage of HandleChain siblings methods.
    * 
    * @return   TRUE  - if the logic of parent class {@link HandleChain} is used,
    *           FALSE - if own siblings implementation logic is used. 
    */
   protected boolean useHandleChainSiblings()
   {
      return false;
   }
   
   /**
    * Sets the value for the attribute WIDTH-CHARS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setWidthCharsWorker(double width)
   {
      config.setWidthChars(width);
      pushScreenDefinition();
   }

   /**
    * Sets the value for the attribute HEIGHT-CHARS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setHeightCharsWorker(double height)
   {
      config.setHeightChars(height);
      pushScreenDefinition();
   }

   /**
    * Sets the value for the attribute WIDTH-PIXELS in the associated config instance.
    * Only valid width values must be passed in the width argument.
    * 
    * @param    width
    *           Valid width value.
    */
   protected void setWidthPixelsWorker(int width)
   {
      config.widthPixels = width;
      pushScreenDefinition();
   }

   /**
    * Sets the value for the attribute HEIGHT-PIXELS in the associated config instance.
    * Only valid height values must be passed in the height argument.
    * 
    * @param    height
    *           Valid height value.
    */
   protected void setHeightPixelsWorker(int height)
   {
      config.heightPixels = height;
      pushScreenDefinition();
   }   
   
   /**
    * Sets the row or column position for this widget.
    *
    * @param    isUnknown
    *           If <code>true<code> the attribute value being set is an unknown value.
    * @param    value
    *           The 1-based row or column number.
    * @param    row
    *           If <code>true</code> <code>value</code> represents a row number, otherwise 
    *           a column number.
    */
   private void setColumnOrRow(boolean isUnknown, double value, boolean row)
   {
      // TO-DO:
      // value > screen-width - frame-width should produce PROGRESS error 4054
      // "**Unable to set COL.  FRAME f does not fit in WINDOW widget. (4054)"
      
      boolean isChui = LogicalTerminal.isChui();
      
      if (isChui)
      {
         value = Coordinate.chuiRound(value);
         
         if (isUnknown || value < Coordinate.MIN_CHUI_CH)
         {
            String msg = String.format(
                  "**Attribute %s must be greater than or equal to 1 on %s widget", 
                  row ? "ROW" : "COL",
                  type());
            ErrorManager.recordOrShowError(4081, msg, false, false, false);
            return;
         }
         
         if (value > Coordinate.MAX_CHUI_CH)
         {
            String msg = String.format("Invalid character unit value %d changed to %d", 
                                       Math.abs((int) value), Coordinate.MAX_CHUI_CH);
            ErrorManager.recordOrShowError(13734, msg, false, true, false);
         }
         
         value = Coordinate.adjustMinMaxChuiChar(value);
      }
      else
      {
         value = Coordinate.guiRound(value);
         
         if (isUnknown || (value > -1 && value < 1))
         {
            String msg = String.format(
                  "**Attribute %s must be greater than or equal to 1 on %s widget", 
                  row ? "ROW" : "COL",
                  type());
            ErrorManager.recordOrShowError(4081, msg, false, false, false);
            return;
         }
         
         if (value < Coordinate.MIN_GUI_CH || value > Coordinate.MAX_GUI_CH)
         {
            double origValue = value;
            value = Coordinate.adjustMinMaxGuiChar(value); 
            String msg;
            if (origValue == (long) origValue)
            {
               msg = String.format("Invalid character unit value %d changed to %d", 
                     Math.abs((long) origValue), Math.abs((long) value));
            }
            else
            {
               msg = String.format("Invalid character unit value %s changed to %s", 
                     Math.abs(origValue), Math.abs(value));
            }
            ErrorManager.recordOrShowError(13734, msg, false, true, false);
         }
      }
      
      // make the value 0-based, it is treated as such by the rest of the framework
      value = value < 0 ? value + 1 : value - 1;
      
      CoordinatesConversion coordinates = LogicalTerminal.coordinates();
      
      // setting the row or column attribute?
      if (row)
      {
         int y = coordinates.rowToPixels(value);
         
         if (_isRealized() && !isChui)
         {
            y = Coordinate.adjustMinMaxGuiPixel(y);
         }
         
         // the physical location must be recalculated back to the character
         // so that the corresponding exact character location is determined
         value = coordinates.rowFromPixels(y);
   
         // bring back to 1-based, to do more validation
         value = value < 0 ? value - 1 : value + 1;
         
         // strangely, the value -1 has a special behavior, where it is silently changed to 1 when
         // assigned to ROW/COLUMN
         if (value == -1)
         {
            value = 1;
         }
         
         if (isChui)
         {
            value = Coordinate.adjustMinMaxChuiChar(value);
         }
         else
         {
            value = Coordinate.adjustMinMaxGuiChar(value);
         }
         
         // config stores 0-based value
         value = value < 0 ? value + 1 : value - 1;
         
         config.setRow(value);
         config.setY(y);
      }
      else
      {
         int x = coordinates.columnToPixels(value);
         
         if (_isRealized() && !isChui)
         {
            x = Coordinate.adjustMinMaxGuiPixel(x);
         }
         
         // the physical location must be recalculated back to the character
         // so that the corresponding exact character location is determined
         value = coordinates.columnFromPixels(x);
   
         // bring back to 1-based, to do more validation
         value = value < 0 ? value - 1 : value + 1;
         
         // strangely, the value -1 has a special behavior, where it is silently changed to 1 when
         // assigned to ROW/COLUMN
         if (value == -1)
         {
            value = 1;
         }
         
         if (isChui)
         {
            value = Coordinate.adjustMinMaxChuiChar(value);
         }
         else
         {
            value = Coordinate.adjustMinMaxGuiChar(value);
         }
         
         // config stores 0-based value
         value = value < 0 ? value + 1 : value - 1;
         
         config.setColumn(value);
         config.setX(x);
      }
      
      if (frame != null && !frame.isBatch())
      {
         LogicalTerminal.setLocation(frame, getId(), config.row, config.column);
      }
   
      pushScreenDefinition();
   }

   /**
    * Sets the X or Y position for this widget.
    *
    * @param    isUnknown
    *           If <code>true<code> the attribute value being set is an unknown value.
    * @param    value
    *           The X or Y position value.
    * @param    x
    *           If <code>true</code> <code>value</code> represents an X position, otherwise 
    *           a Y position.
    */
   private void setXorY(boolean isUnknown, int value, boolean x)
   {
      if (isUnknown)
      {
         String msg = String.format(
               "**Unable to assign UNKNOWN value to attribute %s on %s widget", 
               x ? "X" : "Y", 
               type());
         
         ErrorManager.recordOrShowError(4083, msg, false, false, false);
         return;
      }
      
      if (LogicalTerminal.isChui())
      {
         String msg = String.format(
               "**Unable to set attribute %s on %s widget in the current environment", 
               x ? "X" : "Y", 
               type());
         
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return;
      }
      else
      {
         if (_isRealized())
         {
            value = Coordinate.adjustMinMaxGuiPixel(value);
         }
      }
      
      CoordinatesConversion coordinates = LogicalTerminal.coordinates();
      
      // setting the X or Y attribute?
      if (x)
      {
         config.setX(value);
         config.setColumn(Coordinate.adjustMinMaxGuiChar(coordinates.columnFromPixels(value)));
      }
      else
      {
         config.setY(value);
         config.setRow(Coordinate.adjustMinMaxGuiChar(coordinates.rowFromPixels(value)));
      }
      
      if (frame != null && !frame.isBatch())
      {
         LogicalTerminal.setLocation(frame, getId(), config.row, config.column);
      }
      
      pushScreenDefinition();
   }

   /**
    * Sets the WIDTH or HEIGHT attribute for this widget.
    *
    * @param    isUnknown
    *           If <code>true<code> the attribute value being set is an unknown value.
    * @param    value
    *           The attribute value.
    * @param    width
    *           If <code>true</code> <code>value</code> represents WIDTH, otherwise 
    *           a HEIGHT.
    */
   protected void setWidthOrHeight(boolean isUnknown, double value, boolean width)
   {
      if (isUnknown || value <= 0)
      {
         String msg = String.format("**Attribute %s must be greater than zero on %s widget",
               width ? "WIDTH" : "HEIGHT", type());
         ErrorManager.recordOrShowError(4084, msg, false, false, false);
         return;
      }

      boolean isChui = LogicalTerminal.isChui();
      
      value = isChui ? Coordinate.chuiRound(value) : Coordinate.guiRound(value);
      
      if (value > Coordinate.MAX_GUI_CH)
      {
         String msg;
         if (value == (long) value)
         {
            msg = String.format("Invalid character unit value %d changed to %d", 
                  (long) value, (long) Coordinate.MAX_GUI_CH);
         }
         else
         {
            msg = String.format("Invalid character unit value %s changed to %d", 
                  value, (long) Coordinate.MAX_GUI_CH);
         }
         
         ErrorManager.recordOrShowError(13734, msg, false, true, false);
         
         value = Coordinate.MAX_GUI_CH;
      }

      CoordinatesConversion cc = LogicalTerminal.coordinates();

      // setting the WIDTH or HEIGHT attribute?
      if (width)
      {
         int widthPixels = cc.columnToPixels(value);         
         
         // adjust the input value to the widget minimum/maximum width
         if (_isRealized() && !isChui && widthPixels > Coordinate.MAX_GUI_WIDTH_PIXELS)
         {
            widthPixels = Coordinate.MAX_GUI_WIDTH_PIXELS;
         }

         // the physical location must be recalculated back to the character
         // so that the corresponding exact character value is determined
         value = cc.columnFromPixels(widthPixels);

         setWidthCharsWorker(value);
         setWidthPixelsWorker(widthPixels);
      }
      else
      {
         int heightPixels = cc.rowToPixels(value);

         if (_isRealized() && !isChui && heightPixels > Coordinate.MAX_GUI_HEIGHT_PIXELS)
         {
            heightPixels = Coordinate.MAX_GUI_HEIGHT_PIXELS;
         }

         // the physical location must be recalculated back to the character
         // so that the corresponding exact character value is determined
         value = cc.rowFromPixels(heightPixels);

         setHeightCharsWorker(value);
         setHeightPixelsWorker(heightPixels);
      }

      pushScreenDefinition();
   }
   
   /**
    * Sets the WIDTH-PIXELS or HEIGHT-PIXELS attribute for this widget.
    *
    * @param    isUnknown
    *           If <code>true<code> the attribute value being set is an unknown value.
    * @param    value
    *           The attribute value.
    * @param    width
    *           If <code>true</code> <code>value</code> represents WIDTH-PIXELS, otherwise 
    *           a HEIGHT-PIXELS.
    */
   protected void setWidthOrHeightPixels(boolean isUnknown, int value, boolean width)
   {
      if (isUnknown || value <= 0)
      {
         String msg = String.format("**Attribute %s must be greater than zero on %s widget",
               width ? "WIDTH-PIXELS" : "HEIGHT-PIXELS", type());
         ErrorManager.recordOrShowError(4084, msg, false, false, false);
         return;
      }
      
      CoordinatesConversion cc = LogicalTerminal.coordinates();
      double charValue = 0;

      if (_isRealized() && !LogicalTerminal.isChui())
      {
         // pixel values exceeding the pixel limits for width and height are
         // silently adjusted to the max values
         if (width && value > Coordinate.MAX_GUI_WIDTH_PIXELS)
         {
            value = Coordinate.MAX_GUI_WIDTH_PIXELS;
         }
         
         if (!width && value > Coordinate.MAX_GUI_HEIGHT_PIXELS)
         {
            value = Coordinate.MAX_GUI_HEIGHT_PIXELS;
         }
         
         charValue = width ? cc.columnFromPixels(value) : cc.rowFromPixels(value);
      }
      else
      {
         charValue = width ? cc.columnFromPixels(value) : cc.rowFromPixels(value); 
         if (charValue > Coordinate.MAX_GUI_CH)
         {
            String msg;
            if (charValue == (long) charValue)
            {
               msg = String.format("Invalid character unit value %d changed to %d", 
                     (long) charValue, (long) Coordinate.MAX_GUI_CH);
            }
            else
            {
               msg = String.format("Invalid character unit value %s changed to %d", 
                     charValue, (long) Coordinate.MAX_GUI_CH);
            }
            ErrorManager.recordOrShowError(13734, msg, false, true, false);
            
            charValue = Coordinate.MAX_GUI_CH;
            
            // pixel value for HEIGHT is not adjusted back (at least for window widget)
            if (width)
            {
               value = cc.columnToPixels(charValue);
            }
         }
      }
      
      // setting the WIDTH-PIXELS or HEIGHT-PIXELS attribute?
      if (width)
      {
         setWidthPixelsWorker(value);
         setWidthCharsWorker(charValue);
      }
      else
      {
         setHeightPixelsWorker(value);
         setHeightCharsWorker(charValue);
      }
      
      pushScreenDefinition();
   }
}
