/*
** Module   : ClientConfigManager.java
** Abstract : Management of widget configurations, on client-side.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140929 Created initial version.
** 002 CA  20141014 Separated data which needs to survive the appserver agent's context reset into 
**                  a different container.
** 003 HC  20141106 Javadocs fixes.
** 004 CA  20141125 Added API to deactivate tracking of the changed configs.
** 005 HC  20141207 Added Widget.syncWithConfig(), so far limited to window widget and only some
**                  cases.
*/ 

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Management of widget configurations, on client-side.
*/
public class ClientConfigManager
extends ConfigManager
implements WidgetStateListener
{
   /** This list contains all config classes which implement {@link WidgetConfig}.*/
   private static final List<Class<? extends WidgetConfig>> CONFIG_CLASSES;
   
   /**
    * Flag indicating when the client-side can start pushing updates.  The notification to 
    * start pushing updates will come from the server-side, via 
    * {@link ClientExports#activateConfigUpdates()}.
    */
   private boolean canPushUpdates = false;
   
   /** Cache of the {@link ThinClient} instance. */
   private final ThinClient tc;
   
   static
   {
      // add to this list all config classes which implement WidgetConfig - the entire hierarchy 
      // of classes must be added in this list
      List<Class<? extends WidgetConfig>> configs = new ArrayList<>();
      configs.add(BaseConfig.class);
      configs.add(BrowseConfig.class);
      configs.add(BrowseColumnConfig.class);
      configs.add(ButtonConfig.class);
      configs.add(ComboBoxConfig.class);
      configs.add(ControlConfig.class);
      configs.add(ControlSetConfig.class);
      configs.add(ControlTextConfig.class);
      configs.add(EditorConfig.class);
      configs.add(FieldGroupConfig.class);
      configs.add(FillInConfig.class);
      configs.add(FrameConfig.class);
      configs.add(ImageConfig.class);
      configs.add(LabelConfig.class);
      configs.add(PaneConfig.class);
      configs.add(RadioSetConfig.class);
      configs.add(RectangleConfig.class);
      configs.add(SelectionListConfig.class);
      configs.add(ScrollPaneConfig.class);
      configs.add(SkipConfig.class);
      configs.add(TextConfig.class);
      configs.add(ToggleBoxConfig.class);
      configs.add(WidgetConfig.class);
      configs.add(WindowConfig.class);
      
      CONFIG_CLASSES = Collections.unmodifiableList(configs);
   }

   /**
    * Initialize the client-side by collecting all widget configuration classes (registered in 
    * the {@link ClientConfigManager#CONFIG_CLASSES}) and store them in 
    * the {@link ConfigManager.WorkArea#cfgDefs} map.
    */
   public ClientConfigManager()
   {
      WorkArea wa = survivals.get();
      
      // walk the widget config classes and resolve their configuration
      for (Class<? extends WidgetConfig> cls : CONFIG_CLASSES)
      {
         WidgetConfigDef cfgDef = new WidgetConfigDef(cls);
         
         wa.cfgDefs.put(cls, cfgDef);
      }
      
      this.tc = ThinClient.getInstance();
   }

   /**
    * On client-side, send the widget configurations to the specified server.
    * 
    * @param    server
    *           The server where to  send the configs.
    *           
    * @see      ServerExports#initWidgetConfigurations(Map)
    */
   public void pushConfigDefs(ServerExports server)
   {
      WorkArea wa = survivals.get();
      
      // this is a one-time setup, so we can afford using a map. 
      // key: the config class name
      // value: the list of instance field names
      Map<String, List<String>> configs = new HashMap<>();
      for (WidgetConfigDef cfg : wa.cfgDefs.values())
      {
         String clsName = cfg.cfgCls.getName();
         
         configs.put(clsName, cfg.fieldNames);
      }
      
      server.initWidgetConfigurations(configs);
   }

   /**
    * Inform the client-side that it can no longer send config updates.
    */
   public void deactivateConfigUpdates()
   {
      canPushUpdates = false;
   }
   
   /**
    * Inform the client-side that it can start sending the config updates.
    */
   public void activateConfigUpdates()
   {
      canPushUpdates = true;
   }

   /**
    * Get the widget configuration updates, by comparing the active with the backup configuration.
    * 
    * @return    A list of widget configuration updates to be sent to the server side.
    */
   public WidgetConfigUpdates[] getConfigUpdates()
   {
      // do not send config updates until the server-side is fully initialized.
      if (!canPushUpdates)
         return null;

      return super.getConfigUpdates();
   }

   /**
    * Notify the given widget has been destroyed.  This includes removing its ID from the
    * maps.
    * 
    * @param    w
    *           The deleted widget.
    */
   @Override
   public void notifyWidgetDestroyed(Widget<? extends OutputManager<?>> w)
   {
      WidgetConfig cfg = w.config();
      if (cfg == null)
      {
         // not all widgets have a config
         return;
      }

      WidgetId wid = w.config().id;
      
      notifyWidgetDestroyed(wid);
   }
   
   /**
    * Check if the given ID is associated with a registered widget.
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   <code>true</code> if a widget exists in the registry, with the given ID.
    */
   protected boolean isRegisteredWidget(WidgetId id)
   {
      return tc.getWidget(id.getId()) != null;
   }

   /**
    * Post all the widget configuration from the source to the destination.
    * 
    * @param    dest
    *           The destination config, where the source will be copied.
    * @param    source
    *           The source config.
    */
   @Override
   public void replaceConfig(WidgetConfig dest, WidgetConfig source)
   {
      super.replaceConfig(dest, source);
      
      Widget<?> w = tc.getWidget(dest.id);
      if (w != null)
      {
         w.syncWithConfig();
      }
   }
}
