/*
** Module   : CommonWindow.java
** Abstract : Define all window-related APIs available via a window handle instance.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 CA  20130112 Created initial version.
** 002 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 003 EVL 20131008 Added legacy attributes and methods annotation.
** 004 MAG 20140512 Added new attributes.
** 005 CA  20140701 Added support for font-related attributes:  MESSAGE-AREA-FONT, 
**                  STATUS-AREA-FONT.
** 006 EVL 20141204 Added LOAD-SMALL-ICON setter.
** 007 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;
 
/**
 * Provide access to all window-related APIs which can be accessed via a 
 * handle instance.
 */
@LegacyResource(resource = LegacyResource.WINDOW)
public interface CommonWindow
{
   /**
    * Get icon file name previously loaded by {@link #loadIcon(character)}.
    * 
    * @return  icon file name.
    */
   @LegacyAttribute(name = "ICON")
   public character getIcon();

   /**
    * Gets the FULL-HEIGHT-CHARS writable attribute.
    *
    * @return current value of FULL-HEIGHT-CHARS attribute
    */
   @LegacyAttribute(name = "FULL-HEIGHT-CHARS")
   public decimal getFullHeightChars();
   
   /**
    * Gets the FULL-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of FULL-HEIGHT-PIXELS attribute
    */
   @LegacyAttribute(name = "FULL-HEIGHT-PIXELS")
   public decimal getFullHeightPixels();
   
   /**
    * Gets the FULL-WIDTH-CHARS writable attribute.
    *
    * @return current value of FULL-WIDTH-CHARS attribute
    */
   @LegacyAttribute(name = "FULL-WIDTH-CHARS")
   public decimal getFullWidthChars();
   
   /**
    * Gets the FULL-WIDTH-PIXELS writable attribute.
    *
    * @return current value of FULL-WIDTH-PIXELS attribute
    */
   @LegacyAttribute(name = "FULL-WIDTH-PIXELS")
   public decimal getFullWidthPixels();
   
   /**
    * Get window maximal height.
    * 
    * @return  maximal height.
    */
   @LegacyAttribute(name = "MAX-HEIGHT")
   public decimal getMaxHeight();

   /**
    * Gets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MAX-HEIGHT-CHARS attribute
    */
   @LegacyAttribute(name = "MAX-HEIGHT-CHARS")
   public decimal getMaxHeightChars();

   /**
    * Gets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of MAX-HEIGHT-PIXELS attribute
    */
   @LegacyAttribute(name = "MAX-HEIGHT-PIXELS")
   public integer getMaxHeightPixels();

   /**
    * Get window maximal width.
    * 
    * @return  maximal width.
    */
   @LegacyAttribute(name = "MAX-WIDTH")
   public decimal getMaxWidth();

   /**
    * Gets the MAX-WIDTH-CHARS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-CHARS")
   public decimal getMaxWidthChars();

   /**
    * Gets the MAX-WIDTH-PIXELS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-PIXELS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-PIXELS")
   public integer getMaxWidthPixels();

   /**
    * Gets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @return   The current value of the KEEP-FRAME-Z-ORDER attribute.
    */
   @LegacyAttribute(name = "KEEP-FRAME-Z-ORDER")
   public logical isKeepFrameZOrder();
   
   /**
    * Check if window has KEEP-Z-ORDER option set.
    * 
    * @return <code>true</code> if window has KEEP-Z-ORDER option set.
    */
   @LegacyAttribute(name = "KEEP-Z-ORDER")
   public logical isKeepZOrder();

   /**
    * Check if window has message area.
    * 
    * @return <code>true</code> if window has message area.
    */
   @LegacyAttribute(name = "MESSAGE-AREA")
   public logical isMessageArea();

   /**
    * Check if window size can be changed.
    * 
    * @return  <code>true</code> if window size can be changed.
    */
   @LegacyAttribute(name = "RESIZE")
   public logical isResize();

   /**
    * Check if window has scrollbars.
    * 
    * @return  <code>true</code> if window has scrollbars.
    */
   @LegacyAttribute(name = "SCROLL-BARS")
   public logical isScrollBars();

   /**
    * Check if window has message area.
    * 
    * @return  <code>true</code> if window has message area.
    */
   @LegacyAttribute(name = "STATUS-AREA")
   public logical isStatusArea();

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-ICON")
   public logical loadIcon(character searchPath);

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-ICON")
   public logical loadIcon(String searchPath);

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-SMALL-ICON")
   public logical loadSmallIcon(character searchPath);

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @LegacyMethod(name = "LOAD-SMALL-ICON")
   public logical loadSmallIcon(String searchPath);
   
   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @LegacyAttribute(name = "KEEP-FRAME-Z-ORDER", setter = true)
   public void setKeepFrameZOrder(boolean keep);

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @LegacyAttribute(name = "KEEP-FRAME-Z-ORDER", setter = true)
   public void setKeepFrameZOrder(logical keep);

   /**
    * Enable/disable KEEP-Z-ORDER option.
    * 
    * @param   logical
    *          New flag value.
    */
   @LegacyAttribute(name = "KEEP-Z-ORDER", setter = true)
   public void setKeepZOrder(logical logical);

   /**
    * Set maximal height.
    * 
    * @param   decimal
    *          New option value.
    */
   @LegacyAttribute(name = "MAX-HEIGHT", setter = true)
   public void setMaxHeight(NumberType decimal);

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-HEIGHT-CHARS", setter = true)
   public void setMaxHeightChars(NumberType max);

   /**
    * Sets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-PIXELS attribute.
    */
   @LegacyAttribute(name = "MAX-HEIGHT-PIXELS", setter = true)
   public void setMaxHeightPixels(integer max);

   /**
    * Set maximal width.
    * 
    * @param   decimal
    *          New option value.
    */
   @LegacyAttribute(name = "MAX-WIDTH", setter = true)
   public void setMaxWidth(NumberType decimal);

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-CHARS", setter = true)
   public void setMaxWidthChars(NumberType max);

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   @LegacyAttribute(name = "MAX-WIDTH-PIXELS", setter = true)
   public void setMaxWidthPixels(integer max);

   /**
    * Enable/disable displaying of message area.
    * 
    * @param   logical
    *          New flag value.
    */
   @LegacyAttribute(name = "MESSAGE-AREA", setter = true)
   public void setMessageArea(logical logical);

   /**
    * Enable/disable window resizing.
    * 
    * @param   logical
    *          New option value.
    */
   @LegacyAttribute(name = "RESIZE", setter = true)
   public void setResize(logical logical);

   /**
    * Enable/disable displaying of scrollbars.
    * 
    * @param   logical
    *          New option value.
    */
   @LegacyAttribute(name = "SCROLL-BARS", setter = true)
   public void setScrollBars(logical logical);

   /**
    * Enable/disable displaying of status area.
    * 
    * @param   logical
    *          New option value.
    */
   @LegacyAttribute(name = "STATUS-AREA", setter = true)
   public void setStatusArea(logical logical);

   /**
    * Get the message-area's font.
    * 
    * @return   The message-area's font.
    */
   @LegacyAttribute(name = "MESSAGE-AREA-FONT")
   public int64 getMessageAreaFont();

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "MESSAGE-AREA-FONT", setter = true)
   public void setMessageAreaFont(long fontNum);

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "MESSAGE-AREA-FONT", setter = true)
   public void setMessageAreaFont(int64 fontNum);
   
   /**
    * Get the status-area's font.
    * 
    * @return   The status-area's font.
    */
   @LegacyAttribute(name = "STATUS-AREA-FONT")
   public int64 getStatusAreaFont();

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "STATUS-AREA-FONT", setter = true)
   public void setStatusAreaFont(long fontNum);

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @LegacyAttribute(name = "STATUS-AREA-FONT", setter = true)
   public void setStatusAreaFont(int64 fontNum);
   
   /**
    * Gets the SCREEN-LINES readable attribute.
    *
    * @return   current value of SCREEN-LINES attribute
    */
   @LegacyAttribute(name = "SCREEN-LINES")
   public decimal getScreenLines();
}
