/*
** Module   : ConfigManager.java
** Abstract : Management of widget configurations.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140924 Created initial version.
** 002 CA  20141014 Separated data which needs to survive the appserver agent's context reset into 
**                  a different container.
** 003 CA  20141207 Added an explicit message if field which was not supposed to be tracked has
**                  been collecting dirty changes.
** 004 HC  20141207 Improved config state client-server synchronization.
*/ 

package com.goldencode.p2j.ui;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.client.widget.Widget;
import com.goldencode.util.*;

/**
 * Provides management of widget configurations, as the widget configuration must be shared by all
 * widgets with the same widget ID (be it on server or client-side).
 * <p>
 * On server-side, this class allows update of the widget configurations with the received 
 * {@link WidgetConfigUpdates client-side changes}.
 * <p>
 * On client-side, this class allows the collection of all changes in widget configurations and 
 * package them to be sent to the remote side.  This will only be done after the server-side is 
 * fully initialized and has informed the client-side that it can start sending the updates.
 * Also, the client-side is responsible of pushing the widget config config definitions to the
 * server-side, so that the field IDs are synced on both sides.
 * <p>
 * When new widget configuration classes are added, always add them to the 
 * {@link ClientConfigManager#CONFIG_CLASSES} list.
 * 
 * @see   WidgetConfigDef
 * @see   WidgetConfig
 * @see   WidgetConfigUpdates
 */
public abstract class ConfigManager
{
   /** Context-local state which needs to survive the context reset. */
   protected static ContextLocal<WorkArea> survivals = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      };
      
      @Override
      protected boolean isResetAllowed()
      {
         return false;
      }
   };

   /** Context-local state holding the live widget configurations. */
   private static ContextLocal<ConfigManager> local = new ContextLocal<ConfigManager>()
   {
      protected ConfigManager initialValue()
      {
         return SessionManager.get().isLeaf() ? new ClientConfigManager() 
                                              : new ServerConfigManager();
      };
   };
   
   /** This set is being set to true around the call to 
    * {@link WidgetConfig#applyConfig(WidgetConfig)}.
    */
   private boolean insideApplyConfig = false;
   
   /** Map of widget ID to its {@link WidgetConfigDef definition} . */
   protected final Map<WidgetId, WidgetConfigDef> widgetCfg = new HashMap<>();

   /** Map of active, live, widget configurations. */
   protected final Map<WidgetId, WidgetConfig> activeConfigs = new HashMap<>();

   /** 
    * Map of backup configuration (as when the control was last received by this side). 
    */
   protected final Map<WidgetId, WidgetConfig> backupConfigs = new HashMap<>();

   /** A set of dirty widget configuration. */
   protected final Map<WidgetId, Set<String>> dirtyConfigs = new HashMap<>();

   /**
    * Get the config manager instance for this context. 
    * 
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public static <T extends ConfigManager> T getInstance()
   {
      return (T) local.get();
   }
   
   /**
    * Check if the given ID is associated with a registered widget.
    * 
    * @param    id
    *           The widget ID.
    *           
    * @return   <code>true</code> if a widget exists in the registry, with the given ID.
    */
   protected abstract boolean isRegisteredWidget(WidgetId id);

   /**
    * Determine if the specified config instance can be tracked.
    * <p>
    * The config can be tracked if tracking of the config fields is active at this moment and
    * if the widget is not a {@link WidgetId#virtualWidget(WidgetId) virtual widget}.
    * 
    * @param    cfg
    *           The {@link WidgetConfig} instace which needs to be tracked.
    *           
    * @return   See above.
    */
   public boolean trackConfig(WidgetConfig cfg)
   {
      return !WidgetId.virtualWidget(cfg.id);
   }

   /**
    * Register the specified field in the configuration for the specified widget ID as being 
    * dirty.
    * <p>
    * This is will exclude any negative widget ID and also any widget IDs for which there is no
    * {@link WidgetConfigDef} instance associated (these are new widgets).
    * 
    * @param    cfg
    *           The widget config.
    * @param    field
    *           The changed field.
    */
   public void addDirtyConfig(WidgetConfig cfg, String field)
   {
      if (!trackConfig(cfg))
      {
         return;
      }
      
      Set<String> fields = dirtyConfigs.get(cfg.id);
      if (fields == null)
      {
         dirtyConfigs.put(cfg.id, fields = new HashSet<>());
      }

      if (field != null)
      {
         fields.add(field);
      }
   }

   /**
    * Determine the active configuration for the given ID.  If one does not yet exist, create a
    * new instance based on the given class and set it as the active configuration.
    * 
    * @param    id
    *           The widget ID.
    * @param    cls
    *           The widget's configuration class.
    *           
    * @return   The config instances for this widget.
    */
   public <T extends WidgetConfig> T resolveWidgetConfig(WidgetId id, Class<T> cls)
   {
      T activeConfig = getActiveConfig(id);
      
      if (activeConfig != null)
      {
         return activeConfig;
      }
      
      // create a new, empty config and initialize it.
      try
      {
         WidgetConfigDef cfg = getConfigDef(cls);
         
         activeConfig = cfg.newInstance();
         activeConfig.id = id;
   
         addWidgetConfig(activeConfig);
         
         return activeConfig;
      }
      catch (Throwable e)
      {
         throw new RuntimeException(e);
      }
   }
   
   /**
    * Determine the active configuration for the given ID.  If one does not yet exist, duplicate
    * the received configuration instance.
    * <p>
    * This does not register the config as the active config.  It is the caller's responsibility
    * to decide if the config needs to be registered or not. 
    * 
    * @param    id
    *           The widget ID.
    * @param    cfgInstance
    *           The widget's configuration class.
    *           
    * @return   The config instances for this widget.
    */
   public <T extends WidgetConfig> T resolveWidgetConfig(WidgetId id, T cfgInstance)
   {
      T activeConfig = getActiveConfig(id);
      
      if (activeConfig == null)
      {
         activeConfig = duplicate(cfgInstance);
         activeConfig.id = id;

         // this does not register the config as the active config: this is mainly caused by the
         // fact that the frame's down body doesn't have distinct widgets for each row, on the
         // server-side
      }

      return activeConfig;
   }

   /**
    * Get the widget configuration updates, using the captured dirty configs.
    * 
    * @return    A list of widget configuration updates to be sent to the server side.
    */
   public WidgetConfigUpdates[] getConfigUpdates()
   {
      if (dirtyConfigs.isEmpty())
      {
         return null;
      }

      List<WidgetConfigUpdates> res = new ArrayList<>();

      for (WidgetId wid : dirtyConfigs.keySet())
      {
         if (!activeConfigs.containsKey(wid))
         {
            // skip configs which have no ID or are not part of the active configurations
            continue;
         }

         if (!isRegisteredWidget(wid))
         {
            // skip widgets which are not part of the official registry
            continue;
         }
         
         WidgetConfigDef wdef   = widgetCfg.get(wid);
         WidgetConfig    active = activeConfigs.get(wid);
         WidgetConfig    backup = backupConfigs.get(wid);

         boolean[] changedIds = new boolean[wdef.fieldCount];
         Arrays.fill(changedIds, false);
                  
         for (String field : dirtyConfigs.get(wid))
         {
            Integer fid = wdef.fieldIds.get(field);
            if (fid != null)
            {
               changedIds[fid] = true;
            }
            else
            {
               throw new NullPointerException("Field " + field + " for class " + wdef.cfgCls + 
                                              " has dirty changes, but must not be tracked!");
            }
         }
         
         BitSet changedAttributes = new BitSet(wdef.fieldCount);
         
         List<Object> changes = new ArrayList<>();
         List<NativeTypeSerializer> workers = new ArrayList<>();
         for (int fid = 0; fid < wdef.fieldCount; fid++)
         {
            if (!changedIds[fid] || wdef.ignoredFields[fid])
            {
               changedAttributes.set(fid, false);
               continue;
            }

            try
            {
               Object activeVal = wdef.fields[fid].get(active);
               Object backupVal = wdef.fields[fid].get(backup);

               boolean set = !((activeVal == null && backupVal == null) ||
                               (activeVal != null && activeVal.equals(backupVal)));

               if (set)
               {
                  changes.add(activeVal);
                  workers.add(wdef.workers[fid]);
                  
                  // update the backup too, because we need to track only future changes which
                  // were not sent to the server-side
                  wdef.fields[fid].set(backup, activeVal);
               }
               
               changedAttributes.set(fid, set);
            }
            catch (Throwable t)
            {
               final String msg =
                  "Could not resolve the active/backup values for widget config %s in class %s.";

               String fname = wdef.fieldNames.get(fid);
               throw new RuntimeException(String.format(msg, fname, active.getClass().getName()), 
                                                        t);
            }
         }
         
         if (!changes.isEmpty())
         {
            res.add(new WidgetConfigUpdates(wid.getId(),
                                            changedAttributes, 
                                            changes.toArray(),
                                            workers.toArray(new NativeTypeSerializer[0])));
         }
      }
      
      // clear the dirty configs
      dirtyConfigs.clear();

      return res.isEmpty() ? null : res.toArray(new WidgetConfigUpdates[res.size()]);
   }
   
   /**
    * Post all the widget configuration from the source to the destination.
    * 
    * @param    dest
    *           The destination config, where the source will be copied.
    * @param    source
    *           The source config.
    */
   public void replaceConfig(WidgetConfig dest, WidgetConfig source)
   {
      boolean inApply = insideApplyConfig;
      try 
      {
         insideApplyConfig = true;
         dest.applyConfig(source); // no changes are tracked
      }
      finally
      {
         insideApplyConfig = inApply;
      }
      
      // TODO: replace syncWithWidget with Widget.syncWithConfig
      dest.syncWithWidget();
   }
   
   /**
    * Remove the widget config associated with the given ID, unless there are shared instances
    * of this widget.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void removeWidgetConfig(WidgetId wid)
   {
      removeWidgetConfig(wid, false);
   }
   
   /**
    * Remove the widget config associated with the given ID, only if there aren't shared instances
    * of this widget or the removal is forced.
    * 
    * @param    wid
    *           The widget ID.
    * @param    force
    *           Flag indicating that removal needs to be forced.
    */
   public void removeWidgetConfig(WidgetId wid, boolean force)
   {
      if (wid == null)
      {
         return;
      }
      
      notifyWidgetDestroyed(wid);
   }

   /**
    * Add the given config (associated with a new widget) to the active and backup registries.  
    * Also, the ID is marked as being associated with a new widget (thus ensuring the full 
    * configuration is transmitted to the remote side).
    * 
    * @param    config
    *           The widget configuration associated with a new widget.
    */
   public void addWidgetConfig(WidgetConfig config)
   {
      addWidgetConfig(config, null);
   }
   
   /**
    * Add the given config (associated with a new widget) to the active and backup registries.  
    * Also, the ID is marked as being associated with a new widget (thus ensuring the full 
    * configuration is transmitted to the remote side).
    * <p>
    * This method can be used in case when a widget is initialized and its config state
    * is changed during the init process. The initial config passed to the created widget
    * is in this case interpreted as the backup confing and the changed config as the active
    * config. The difference ensures that the changes are properly synchronized with the other 
    * side (server/client).
    * 
    * @param    activeConfig
    *           The active widget configuration associated with a new widget.
    * @param    backupConfig
    *           The widget configuration that will be used as a backup config.
    *           If null, the activeConfig will be used as backup as well.  
    */
   public void addWidgetConfig(WidgetConfig activeConfig, WidgetConfig backupConfig)
   {
      WidgetId id = activeConfig.id;
      
      if (getActiveConfig(id) != null)
      {
         throw new RuntimeException("A configuration is already associated for " + 
                                    activeConfig.toString());
      }

      backupConfig = backupConfig != null ? duplicate(backupConfig) : duplicate(activeConfig);
      
      // the backup doesn't have an ID, because we don't want to access the widget
      backupConfig.id = null;

      WidgetConfigDef wdef = getConfigDef(activeConfig.getClass());
      
      widgetCfg.put(id, wdef);
      activeConfigs.put(id, activeConfig);
      backupConfigs.put(id, backupConfig);
   }

   /**
    * Get the active config associated with the given widget ID.
    * 
    * @param    widgetId
    *           The widget ID.
    *           
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   public <T extends WidgetConfig> T getActiveConfig(WidgetId widgetId)
   {
      return (T) activeConfigs.get(widgetId);
   }
   
   /**
    * The method returns true when {@link WidgetConfig#applyConfig(WidgetConfig)}
    * is called by this instance.
    * 
    * @return   See the description.
    */
   public boolean isInsideApplyConfig()
   {
      return insideApplyConfig;
   }

   /**
    * Update the local active and backup configurations with the received changes.
    * 
    * @param    cfgUpdate
    *           The container with the config updates for a given widget ID.
    */
   protected void applyConfigUpdates(WidgetConfigUpdates cfgUpdate)
   {
      // there is a "chicken-and-the-egg" problem here.  if the widget has just been created, the
      // state sync is applied BEFORE the pushScreenDefinition API is executed.  this means that
      // we don't know the widget's config class (and other info), as we are too early.  for this
      // reason, all calls for which there is no widget config available are dropped.  it is 
      // expected for the server-side to setup the widget properly (including the active/backup 
      // config instances) via pushScreenDefinition calls.
      
      // the server-side holds all client-side state in a single config...
      WidgetId wid = new WidgetId(cfgUpdate.widgetId);
      
      WidgetConfigDef wdef = widgetCfg.get(wid);
      if (wdef == null)
      {
         return;
      }
   
      // determine the mapping of field indices to their update slot
      Map<Integer, Object> fid2upd = new HashMap<>();
      int fidx = 0;
      for (int fid = 0; fid < wdef.fieldCount; fid++)
      {
         if (cfgUpdate.changedAttributes.get(fid))
         {
            Object val = cfgUpdate.updates[fidx];
            fidx = fidx + 1;
            
            fid2upd.put(fid, val);
         }
      }
      
      WidgetConfig active = activeConfigs.get(wid);
      WidgetConfig backup = backupConfigs.get(wid);
   
      // use the cfg classes to walk the fields; this ensures the fields are updated in the proper
      // order, by custom/explicit setter and by correct class hierarchy
      for (Class<? extends WidgetConfig> cls : wdef.fieldsByClass.keySet())
      {
         Set<Integer> fields = wdef.fieldsByClass.get(cls);
         
         for (int fid : fields)
         {
            if (fid2upd.containsKey(fid))
            {
               Object val = fid2upd.get(fid);
               
               setWidgetConfig(active, val, wdef, fid);
            }
         }
      }

      boolean inApply = insideApplyConfig;
      try 
      {
         insideApplyConfig = true;
         // update the backup configuration; no changes are tracked
         backup.applyConfig(active);
      }
      finally
      {
         insideApplyConfig = inApply;
      }

      // TODO: replace syncWithWidget with Widget.syncWithConfig
      // post-processing, only for active
      active.syncWithWidget();
   }

   /**
    * Get the {@link WidgetConfigDef} instance associated with the given {@link WidgetConfig} 
    * implementation.
    * 
    * @param    cls
    *           The widget config class.
    *           
    * @return   See above.
    */
   WidgetConfigDef getConfigDef(Class<? extends WidgetConfig> cls)
   {
      WorkArea wa = survivals.get();
      
      return wa.cfgDefs.get(cls);
   }

   /**
    * Get the {@link WidgetConfigDef} instance associated with the given widget.
    * 
    * @param    wid
    *           The widget ID.
    *           
    * @return   See above.
    */
   WidgetConfigDef getConfigDef(int wid)
   {
      WorkArea wa = survivals.get();
      
      WidgetConfig cfg = activeConfigs.get(new WidgetId(wid));
      
      return wa.cfgDefs.get(cfg.getClass());
   }

   /**
    * Dynamically set a specific widget configuration field.  If the widget config class (or one  
    * of its parent classes) does not define a field with the specified name, then this will be a 
    * no-op.
    *  
    * @param    cfg
    *           The config instance which needs to be updated.
    * @param    fname
    *           The field name.
    * @param    ftype
    *           The field type.
    * @param    value
    *           The new field value.
    */
   <T> void setDynamicConfig(WidgetConfig cfg, String fname, Class<T> ftype, T value)
   {
      WidgetConfigDef wdef = getConfigDef(cfg.getClass());
      Integer fid = wdef.fieldIds.get(fname);
      
      if (fid == null)
      {
         // field not found
         return;
      }
   
      // this must be done before the setter call
      addDirtyConfig(cfg, fname);

      setWidgetConfig(cfg, value, wdef, fid);
   }

   /**
    * Dynamically get a specific widget configuration field.  If the widget config class (or one  
    * of its parent classes) does not define a field with the specified name, then this will be 
    * return the default value specified as parameter.
    * 
    * @param    cfg
    *           The config instance from where the field is retrieved. 
    * @param    fname
    *           The field name.
    * @param    ftype
    *           The field type.
    * @param    defVal
    *           The default value.
    *           
    * @return   See above.
    */
   @SuppressWarnings("unchecked")
   <T> T getDynamicConfig(WidgetConfig cfg, String fname, Class<T> ftype, T defVal)
   {
      T res = defVal;
      
      WidgetConfigDef wdef = getConfigDef(cfg.getClass());
      Integer fid = wdef.fieldIds.get(fname);
      
      if (fid == null)
      {
         // field not found
         return defVal;
      }
   
      try
      {
         res = (T) wdef.fields[fid].get(cfg);
      }
      catch (Throwable t)
      {
         final String msg = "Could not get the config attr %s for widget %d with class %s.";
         
         throw new RuntimeException(String.format(msg, fname, cfg.id, wdef.cfgCls.getName()), 
                                    t);
      }
      
      return res;
   }

   /**
    * Duplicate the given configuration.
    * <p>
    * Changes are not tracked during duplication.
    * 
    * @param    config
    *           The configuration to be duplicated.
    *           
    * @return   A duplicate with the state matching the specified config.
    * 
    * @throws   NullPointerException
    *           If the received config is <code>null</code> or the associated 
    *           {@link WidgetConfigDef} instance could not be found. 
    */
   <T extends WidgetConfig> T duplicate(T config)
   {
      if (config == null)
      {
         throw new NullPointerException("Received a null configuration source!");
      }

      WidgetConfigDef wdef = getConfigDef(config.getClass());
      
      if (wdef == null)
      {
         throw new NullPointerException("Could not resolve the configuration def for " + config);
      }
      
      try
      {
         // clone the widget config

         T clone = wdef.newInstance();
         
         boolean inApply = insideApplyConfig;
         try 
         {
            insideApplyConfig = true;
            clone.applyConfig(config);
         }
         finally
         {
            insideApplyConfig = inApply;
         }

         return clone;
      }
      catch (Throwable t)
      {
         final String msg = "Could not duplicate config class %s for widget %s.";
         
         throw new RuntimeException(String.format(msg, wdef.cfgCls.getName(), config.id), t);
      }
   }
   
   /**
    * Set a widget configuration field to the specified value.
    * <p>
    * This ensures that the proper {@link WidgetConfigDef#customFieldSetters custom setter} is
    * called, if one exists.  Otherwise, the field is set directly.
    * 
    * @param    cfg
    *           The widget configuration to be updated.
    * @param    value
    *           The new field value.
    * @param    wdef
    *           The definition of the widget configuration.
    * @param    fid
    *           The ID of the changed field.
    */
   <T extends WidgetConfig> void setWidgetConfig(T               cfg, 
                                                 Object          value,
                                                 WidgetConfigDef wdef,
                                                 int             fid)
   {
      // check for special processing when a value has changed
      Method customSetter = wdef.customFieldSetters[fid];
      if (customSetter != null)
      {
         try
         {
            customSetter.invoke(cfg, value);
         }
         catch (Throwable t)
         {
            final String msg = "Could not process custom setter for config attr %s for " +
                               "widget %s with class %s.";
            String fname = wdef.fieldNames.get(fid);
            
            throw new RuntimeException(String.format(msg, fname, cfg.id, wdef.cfgCls.getName()),
                                       t);
         }
      }
      else
      {
         try
         {
            Field f = wdef.fields[fid];
            f.set(cfg, value);
         }
         catch (Throwable t)
         {
            final String msg = "Could not set the config attr %s for widget %d with class %s.";
            
            String fname = wdef.fieldNames.get(fid);
            
            throw new RuntimeException(String.format(msg, fname, cfg.id, wdef.cfgCls.getName()), 
                                       t);
         }
      }
   }

   /**
    * Remove this widget IDs from all the maps.
    * 
    * @param    wid
    *           The widget ID to be removed.
    */
   void notifyWidgetDestroyed(WidgetId wid)
   {
      widgetCfg.remove(wid);
      activeConfigs.remove(wid);
      backupConfigs.remove(wid);
      dirtyConfigs.remove(wid);
   }

   /**
    * Widget-specific data which needs to survive the context-reset.
    */
   static class WorkArea
   {
      /** Map of widget configuration definitions by their config class. */
      protected final Map<Class<? extends WidgetConfig>, WidgetConfigDef> cfgDefs = new HashMap<>();
   }
}
