/*
** Module   : ControlConfig.java
** Abstract : stores common non-container widget attributes
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 EVL 20050714   @21692 Creation based on general concept. Necessary
**                           modification to get compiled.
** 002 EVL 20050728   @21909 Moving screenValue variable to ControlConfig
**                           from ControlSetConfig because not only set
**                           like widgets must use it, but text related
**                           too. For example - FillIn.
** 003 EVL 20050822   @22372 Adding the event list the server will wait
**                           for.
** 004 SIY 20050831   @22495 Refactoring (see @@22394-22409) 
** 005 SIY 20050907   @22627 Fixed use of @inheritDoc tag in javadoc
** 006 SIY 20050909   @22662 Fixed setting of the label and screen value.
** 007 NVS 20051129   @23559 Changed class access to public.
** 008 SIY 20060203   @24248 Added force-label flag.
** 009 SIY 20060209   @24477 Major refactoring.
** 010 SIY 20060331   @25291 Added setLabels() method.
** 011 SIY 20060515   @26148 Encapsulated access to label field.
** 012 SIY 20060515   @26149 Moved COLUMN-LABEL attribute from 
**                           ComponentConfig.
** 013 GES 20060803   @28400 Made externalizable for better performance.
**                           Many code standard cleanups.
** 014 NVS 20060918   @29644 Added setAttrSpace(), isAttrSpace() methods
**                           that support per field level widget ATTR-SPACE
**                           attribute.
** 015 SIY 20080513   @38272 Do not allow reducing of label length.
** 016 SIY 20080504   @38547 Updated comments to reflect current findings.
** 017 GES 20090424   @41961 Import change.
** 018 SIY 20100619          Switching to own implementation of character UI.
** 019 SIY 20100804          Changes in imports, some refactorings.
** 020 SIY 20101122          Changes related to Button hierarchy cleanup.
** 021 SIY 20111101          Moved from CHUI package. Minor 
**                           cleanups/refactorings in preparation to GUI client.
** 022 SIY 20111202          Updated imports.
** 023 SIY 20111221          Updated imports.
** 024 HC  20140911          GUI coordinate support - data types refactoring.
** 025 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 026 HC  20141021          Implemented window move/position support.
** 027 HC  20141207          Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import java.io.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.util.*;

/**
 * The class holds the non-container widget generic attributes.
 */
public class ControlConfig
extends BaseConfig
{
   /** The Progress AUTO-RESIZE attribute. */
   public boolean autoResize = true;
   
   /** The NO-LABELS flag. */
   public boolean labels = true;

   /** The ATTR-SPACE flag. */
   public boolean attrSpace = false;

   /** The Progress DATA-TYPE attribute. */
   public String dataType = "character";

   /** Force setting of the label. */
   public boolean forceLabel = false;
   
   /** Help string used when widget becomes focused. */
   public String help = null;
   
   /** The LABEL attribute. */
   public String label = null;

   /** The COLUMN-LABEL attribute. */
   public String clabel = null;

   /** Value of the INDEX attribute. */
   public int index = 0;
   
   /** The previous label value. */
   private transient String prevLabel = null;

   /**
    * Default constructor (only used in de-serialization).
    */
   public ControlConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected ControlConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected ControlConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new value to WIDTH-CHARS attribute.
    * 
    * @param    widthChars
    *           New value for the attribute.
    */
   @Override
   public void setWidthChars(double widthChars)
   {
      widthChars = Coordinate.scale(widthChars);
      
      if (widthChars == 0)
         return;
      
      super.setWidthChars(widthChars);
      
      if (id == null || !onClientSide())
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);

      if (instance == null)
      {
         return;
      }
      
      instance.repaint();
   }
   
   /**
    * Synchronize the configuration with the widget state.
    */
   @Override
   public void syncWithWidget()
   {
      super.syncWithWidget();
      
      if (!onClientSide())
      {
         if (this.label == null)
         {
            this.label = this.prevLabel;
         }
         else
         {
            String label = this.label;

            // 4GL does no allow applying of very long labels and just rejects them.
            // If new label is shorter than existing one, new label is aligned right.
            if (prevLabel != null && 
                label != null     &&
                label.length() < prevLabel.length())
            {
               int len = prevLabel.length() - label.length();
               label = StringHelper.repeatChar(' ', len) + label;
            }
            
            this.prevLabel = this.label;
            this.label = label;
         }

         return;
      }

      if (id == null)
      {
         return;
      }

      Widget<?> instance = tc.getWidget(id);
      
      if (instance == null)
      {
         return;
      }

      if (instance instanceof Text)
         ((Text<?>) instance).setDataType(dataType);
      else if (instance instanceof FillIn)
         ((FillIn<?>) instance).setDataType(dataType);
      else if (instance instanceof Button)
         ((Button<?>) instance).setTextLabel(label);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      ControlConfig cfg = (ControlConfig) config;
      
      autoResize = cfg.autoResize;
      labels = cfg.labels;
      attrSpace = cfg.attrSpace;
      dataType = cfg.dataType;
      forceLabel = cfg.forceLabel;
      help = cfg.help;
      label = cfg.label;
      clabel = cfg.clabel;
      index = cfg.index;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      autoResize         = in.readBoolean();
      labels             = in.readBoolean();
      attrSpace          = in.readBoolean();
      forceLabel         = in.readBoolean();
      dataType           = (String) in.readObject();
      help               = (String) in.readObject();
      label              = (String) in.readObject();
      clabel             = (String) in.readObject();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeBoolean(autoResize);
      out.writeBoolean(labels);
      out.writeBoolean(attrSpace);
      out.writeBoolean(forceLabel);
      out.writeObject(dataType);
      out.writeObject(help);
      out.writeObject(label);
      out.writeObject(clabel);
  }
}
