/*
** Module   : Coordinate.java
** Abstract : Provides elementary operations for the UI coordinate data types.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20140911 Created initial version.
** 002 MAG 20140926 Added double to integer conversion helper.
** 003 HC  20141021 Implemented window move/position support.
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import java.math.BigDecimal;

import com.goldencode.p2j.ui.client.*;

/**
 * The class provides elementary operations for the UI coordinate data types.   
 */
public class Coordinate
{
   /**
    * Minimum character coordinate value for ChUI environments.
    */
   public static final int MIN_CHUI_CH =    1;

   /**
    * Maximum character coordinate value for ChUI environments.
    */
   public static final int MAX_CHUI_CH =    131072;

   /**
    * Minimum character coordinate value for GUI environments.
    */
   public static final double MIN_GUI_CH = -131072;

   /**
    * Maximum character coordinate value for GUI environments.
    */
   public static final double MAX_GUI_CH =  131072;
   
   /**
    * Minimum pixel coordinate value for a realized widget.
    */
   public static final int MIN_GUI_PIXEL = -32768;
   
   /**
    * Maximum pixel coordinate value for a realized widget.
    */
   public static final int MAX_GUI_PIXEL = 32767;
   
   /** 
    * Maximum pixel width for a realized widget in GUI mode.
    */
   public static final int MAX_GUI_WIDTH_PIXELS = 65527; 

   /** 
    * Maximum pixel height for a realized widget in GUI mode.
    */
   public static final int MAX_GUI_HEIGHT_PIXELS = 65456;
   
   /**
    * Performs half-up rounding of a character coordinate value.
    * The result is meaningful only to ChUI environments.
    * 
    * @param   ch
    *          The coordinate value in character units.
    * 
    * @return  The rounded coordinate value.
    */
   public static double chuiRound(double ch)
   {
      // take a shortcut in case the input is already a whole number 
      int ich = (int) ch;
      if (ich == ch)
      {
         return ich;
      }
      
      return BigDecimal.valueOf(ch).setScale(0, BigDecimal.ROUND_HALF_UP).doubleValue();
   }
   
   /**
    * Performs half-up rounding of a character coordinate value.
    * The result is meaningful only to GUI environments.
    * 
    * @param   ch
    *          The coordinate value in character units.
    * 
    * @return  The rounded coordinate value.
    */
   public static double guiRound(double ch)
   {
      // take a shortcut in case the input is already a whole number 
      int ich = (int) ch;
      if (ich == ch)
      {
         return ich;
      }
      
      return BigDecimal.valueOf(ch).setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   ch
    *          The coordinate value in character units.
    * 
    * @return  The scaled coordinate value.
    */
   public static double scale(double ch)
   {
      validate(ch);
      
      // take a shortcut in case the input is already a whole number 
      int ich = (int) ch;
      if (ich == ch)
      {
         return ich;
      }
      
      return BigDecimal.valueOf(ch).setScale(2, BigDecimal.ROUND_HALF_UP).doubleValue();
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   insets
    *          An {@link Insets} instance. 
    * 
    * @return  The scaled {@link Insets} instance.
    */
   public static Insets scale(Insets insets)
   {
      return new Insets(scale(insets.top), scale(insets.left), 
                        scale(insets.bottom), scale(insets.right));
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   point
    *          A {@link Point} instance. 
    * 
    * @return  The scaled {@link Point} instance.
    */
   public static Point scale(Point point)
   {
      return new Point(scale(point.x), scale(point.y));
   }
   
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   rect
    *          An {@link Rectangle} instance. 
    * 
    * @return  The scaled {@link Rectangle} instance.
    */
   public static Rectangle scale(Rectangle rect)
   {
      return new Rectangle(scale(rect.top()), scale(rect.left()), 
                           scale(rect.bottom()), scale(rect.right()));
   }
   
   /**
   /**
    * Sets the input value to the default scale (of two decimal places) for the character 
    * coordinate type. Use this method whenever the value holds an unknown precision 
    * (for example a result of an arithmetic operation) and the default scale is required
    * (for example when comparing two character values).
    * <p>
    * Note that the result of this method - the default scale for a character coordinate value -
    * doesn't depend on the UI mode. It is the same for both GUI and ChUI.
    * 
    * @param   dim
    *          An {@link Dimension} instance. 
    * 
    * @return  The scaled {@link Dimension} instance.
    */
   public static Dimension scale(Dimension dim)
   {
      return new Dimension(scale(dim.width), scale(dim.height));
   }
   
   /**
    * Validates the input value. The method throws {@link RuntimeException}
    * if the input value is:
    * <ul>
    * <li>equal to <code>Double.NaN,</code></li>
    * <li>infinitely large.</code></li>
    * </ul> 
    * 
    * @param   ch
    *          Coordinate value in character units.
    *          
    * @throws  RuntimeException
    *          When the <code>ch</code> is not a valid coordinate value.
    */
   private static void validate(double ch)
   {
      if (Double.isInfinite(ch) || Double.isNaN(ch))
      {
         throw new RuntimeException("Invalid character coordinate value.");
      }
   }

   /**
    * Convert a double value to integer. Round integer value.
    * 
    * @param   val
    *          Double value.
    *          
    * @return  Integer rounded value.
    */
   public static int toInt(double val)
   {
      return Double.valueOf(val + 0.5).intValue();
   }
   
   /**
    * The input value is adjusted according to minimum and maximum possible
    * character coordinate values for ChUI environments.
    * <p>
    * If the input value is less than {@linkplain #MIN_CHUI_CH} then {@linkplain #MIN_CHUI_CH}
    * is returned. If the input value is greater than {@linkplain #MAX_CHUI_CH} then 
    * {@linkplain #MAX_CHUI_CH} is returned. Otherwise the input value is returned.
    * 
    * @param   value
    *          Character coordinate value being adjusted.
    *          
    * @return  The adjusted character coordinate value.
    */
   public static double adjustMinMaxChuiChar(double value)
   {
      if (value < MIN_CHUI_CH)
      {
         return MIN_CHUI_CH;
      }
      else if (value > MAX_CHUI_CH)
      {
         return MAX_CHUI_CH;
      }
      
      return value;
   }
   
   /**
    * The input value is adjusted according to minimum and maximum possible
    * character coordinate values for ChUI environments.
    * <p>
    * If the input value is less than {@linkplain #MIN_GUI_CH} then {@linkplain #MIN_GUI_CH}
    * is returned. If the input value is greater than {@linkplain #MAX_GUI_CH} then 
    * {@linkplain #MAX_GUI_CH} is returned. Otherwise the input value is returned.
    * 
    * @param   value
    *          Character coordinate value being adjusted.
    *          
    * @return  The adjusted character coordinate value.
    */
   public static double adjustMinMaxGuiChar(double value)
   {
      if (value < MIN_GUI_CH)
      {
         return MIN_GUI_CH;
      }
      else if (value > MAX_GUI_CH)
      {
         return MAX_GUI_CH;
      }
      
      return value;
   }
   
   /**
    * The input value is adjusted according to minimum and maximum possible
    * character coordinate values for ChUI environments.
    * <p>
    * If the input value is less than {@linkplain #MIN_GUI_PIXEL} then 
    * {@linkplain #MIN_GUI_PIXEL} is returned. If the input value is greater than 
    * {@linkplain #MAX_GUI_PIXEL} then {@linkplain #MAX_GUI_PIXEL} is returned. 
    * Otherwise the input value is returned.
    * 
    * @param   value
    *          Character coordinate value being adjusted.
    *          
    * @return  The adjusted character coordinate value.
    */
   public static int adjustMinMaxGuiPixel(int value)
   {
      if (value < MIN_GUI_PIXEL)
      {
         return MIN_GUI_PIXEL;
      }
      else if (value > MAX_GUI_PIXEL)
      {
         return MAX_GUI_PIXEL;
      }
      
      return value;
   }
}
