/*
** Module   : LogicalTerminalConversionParams.java
** Abstract : Server-side resolution of coordinate conversion parameters.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 HC  20141021 Created initial version.
** 002 HC  20141207 Implemented runtime support for several SESSION attributes.
*/
package com.goldencode.p2j.ui;

import com.goldencode.p2j.security.*;

/**
 * The class resolves coordinate conversion parameters based on the static methods
 * of {@link LogicalTerminal}.
 */
class LogicalTerminalConversionParams
implements CoordinatesConversionParams
{
   /** Context-local instance of the cached parameters */
   private static final ContextLocal<Params> params =
   new ContextLocal<Params>()
   {
      @Override
      protected Params initialValue()
      {
         Params result = new Params();
         ClientExports client = LogicalTerminal.getClient();
         result.isChui = client.isChui();
         result.pixelsPerColumn = client.getPixelsPerColumn();
         result.pixelsPerRow = client.getPixelsPerRow();
         return result;
      }
   };

   /**
    * Flag indicating the ChUI or GUI mode.
    * 
    * @return  Boolean value indicating the ChUI or GUI mode.
    */
   @Override
   public boolean isChui()
   {
      return params.get().isChui;
   }

   /**
    * Returns the number of pixels in each column of the display. This usually corresponds
    * to the 4GL attribute SESSION:PIXELS-PER-COLUMN.
    * 
    * @return  Number of pixels in each column.
    */
   @Override
   public int getPixelsPerColumn()
   {
      return params.get().pixelsPerColumn;
   }

   /**
    * Returns the number of pixels in each row of the display. This usually corresponds
    * to the 4GL attribute SESSION:PIXELS-PER-ROW.
    * 
    * @return  Number of pixels in each row.
    */
   @Override
   public int getPixelsPerRow()
   {
      return params.get().pixelsPerRow;
   }

   /**
    * Get the width of the given font.
    * 
    * @param    windowId
    *           The ID of the window from which the font-table is used.
    * @param    font
    *           Font index.
    *          
    * @return   Width of the font.
    */
   @Override
   public int getFontWidth(WidgetId windowId, int font)
   {
      throw new RuntimeException("Not implemented!");
   }

   /**
    * Get the height of the given font.
    * 
    * @param    windowId
    *           The ID of the window from which the font-table is used.
    * @param    font
    *           Font index.
    *          
    * @return   Height of the font.
    */
   @Override
   public int getFontHeight(WidgetId windowId, int font)
   {
      throw new RuntimeException("Not implemented!");
   }
   
   /**
    * Container for storing the conversion parameters. 
    */
   private static class Params
   {
      /**
       * Flag indicating the ChUI or GUI mode. 
       */
      public boolean isChui;
      
      /**
       * Pixels per column.
       */
      public int pixelsPerColumn;
      
      /**
       * Pixels per row.
       */
      public int pixelsPerRow;
   }
}