/*
** Module   : PaneConfig.java
** Abstract : container for the DIALOG, FRAME, WINDOS widget state
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140514 Created initial version.
** 002 HC  20140911 GUI coordinate support - data types refactoring.
** 003 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import java.io.*;

/**
 * The container for common widget state variables shared by 
 * DIALOG, FRAME, WINDOW widgets.
 */
public abstract class PaneConfig
extends BaseConfig
{
   /** THREE-D options */
   public boolean threeD = true;
   
   /** VIRTUAL-WIDTH-CHARS option */
   public double virtualWidthChars; 

   /** VIRTUAL-HEIGHT-CHARS option */
   public double virtualHeightChars; 
   
   /** VIRTUAL-WIDTH-PIXELS option */
   public double virtualWidthPixels; 

   /** VIRTUAL-HEIGHT-PIXELS option */
   public double virtualHeightPixels; 
   
   /**
    * Default constructor (only used in deserialization).
    */
   public PaneConfig()
   {
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected PaneConfig(int id)
   {
      super(id);
   }

   /**
    * Create a new config and associate it with the given widget.
    * 
    * @param    id
    *           The ID of the widget to which this configuration belongs.
    */
   protected PaneConfig(WidgetId id)
   {
      super(id);
   }

   /**
    * Set new values from the provided configuration instance.
    * 
    * @param    config
    *           The instance from which to take values.
    */
   @Override
   public void applyConfig(WidgetConfig config)
   {
      super.applyConfig(config);
      
      PaneConfig cfg = (PaneConfig) config;
      
      threeD = cfg.threeD;
      virtualWidthChars = cfg.virtualWidthChars;
      virtualHeightChars = cfg.virtualHeightChars;
      virtualWidthPixels = cfg.virtualWidthPixels;
      virtualHeightPixels = cfg.virtualHeightPixels;
   }

   /**
    * Replacement for the default object reading method. The latest
    * state is read from the input source.  The parent's state is
    * read first.
    * 
    * @param    in
    *           The input source from which fields will be restored.
    *
    * @throws   IOException
    *           In case of I/O errors.
    * @throws   ClassNotFoundException
    *           If payload can't be instantiated.
    */
   @Override
   public void readExternal(ObjectInput in)
   throws IOException,
          ClassNotFoundException
   {
      super.readExternal(in);
      threeD             = in.readBoolean();
      virtualHeightChars = in.readDouble();
      virtualWidthChars  = in.readDouble();
      virtualHeightPixels = in.readDouble();
      virtualWidthPixels  = in.readDouble();
   }

   /**
    * Replacement for the default object writing method. The latest
    * state is written to the output destination.  The parent's state is
    * written first.
    * 
    * @param    out
    *           The output destination to which fields will be saved.
    *
    * @throws   IOException
    *           In case of I/O errors.
    */
   @Override
   public void writeExternal(ObjectOutput out)
   throws IOException
   {
      super.writeExternal(out);
      out.writeBoolean(threeD);
      out.writeDouble(virtualHeightChars);
      out.writeDouble(virtualWidthChars);      
      out.writeDouble(virtualHeightPixels);
      out.writeDouble(virtualWidthPixels);      
   }
}
