/*
** Module   : PaneEntity.java
** Abstract : server side base class for DIALOG, FRAME, WINDOW widgets
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140514 Created initial version.      
** 002 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 003 HC  20141021 Implemented window move/position support.
** 004 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.util.*;

/**
 * The server side base class for DIALOG, FRAME, WINDOW widgets. 
 */
public class PaneEntity<T extends PaneConfig>
extends BaseEntity<T>
{
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    paneConfig
    *           The container for all state to be transferred to the client.
    */
   public PaneEntity(boolean dynamic, T paneConfig)
   {
      super(dynamic, paneConfig);
   }
   
   /**
    * Constructor.
    *
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    paneConfig
    *           The container for all state to be transferred to the client.
    * @param    before
    *           The right neighbor of the BaseEntity been created or null if new BaseEntity
    *           will be the last in the chain. If <b>before</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    * @param    firstAsLast
    *           The flag indicating that this new entity must be inserted as last element 
    *           in the chain in the case when <b>before</b> - is the first element of the chain.
    *           When <b>before</b> is the first element of the chain, default behavior (expected
    *           from the name of this parameter is to insert new element as the first. 
    *           This flag allows to change the expected behavior of <b>before</b> parameter when 
    *           it needs and makes sense only if <b>before</b> is the first element.
    */
   public PaneEntity(boolean dynamic, T paneConfig, PaneEntity<T> before, boolean firstAsLast)
   {
      super(dynamic, paneConfig, before, firstAsLast);
   }
   
   /**
    * Check if window has THREE-D option set.
    * 
    * @return <code>true</code> if window has THREE-D option set.
    */
   @Override
   public logical isThreeD()
   {
      return new logical(config.threeD);
   }

   /**
    * Enable/disable 3D look.
    * 
    * @param   threeD
    *          New flag value.
    */
   @Override
   public void setThreeD(logical threeD)
   {
      config.threeD = threeD.booleanValue();
      pushScreenDefinition();
   }

   /**
    * Get VIRTUAL-HEIGHT-CHARS attribute. 
    */
   @Override
   public decimal getVirtHeight()
   {
      return new decimal(config.virtualHeightChars);
   }

   /**
    * Set VIRTUAL-HEIGHT-CHARS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-CHARS value
    */
   @Override
   public void setVirtHeight(NumberType height)
   {
      config.virtualHeightChars = Coordinate.scale(height.doubleValue());
      pushScreenDefinition();
   }

   /**
    * Get VIRTUAL-WIDTH-CHARS attribute. 
    */
   @Override
   public decimal getVirtWidth()
   {
      return new decimal(config.virtualWidthChars);
   }

   /**
    * Set VIRTUAL-WIDTH-CHARS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-CHAR value
    */
   @Override
   public void setVirtWidth(NumberType width)
   {
      config.virtualWidthChars = Coordinate.scale(width.doubleValue());
      pushScreenDefinition();
   }
   
   /**
    * Get VIRTUAL-HEIGHT-PIXELS attribute. 
    */
   @Override
   public decimal getVirtHeightPixels()
   {
      return new decimal(config.virtualHeightPixels);
   }

   /**
    * Set VIRTUAL-HEIGHT-PIXELS attribute. 
    *
    * @param   height
    *          VIRTUAL-HEIGHT-PIXELS value
    */
   @Override
   public void setVirtHeightPixels(NumberType height)
   {
      config.virtualHeightPixels = height.intValue();
      pushScreenDefinition();
   }

   /**
    * Get VIRTUAL-WIDTH-PIXELS attribute. 
    */
   @Override
   public decimal getVirtWidthPixels()
   {
      return new decimal(config.virtualWidthPixels);
   }

   /**
    * Set VIRTUAL-WIDTH-PIXELS attribute. 
    *
    * @param   width
    *          VIRTUAL-WIDTH-PIXELS value
    */
   @Override
   public void setVirtWidthPixels(NumberType width)
   {
      config.virtualWidthPixels = width.intValue();
      pushScreenDefinition();
   }
}
