/*
** Module   : RectangleWidget.java
** Abstract : RECTANGLE widget server side implementation.
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20101002 Created initial version
** 002 SIY 20111101 Updated imports.
** 003 CA  20130222 Added NPE protection for cases when a widget is used before is 
**                  attached to a frame (after a CREATE WIDGET statement is used).
** 004 CA  20130927 Added LegacyResource annotation, to determine resource type.
** 005 CA  20131115 Changes related to properly instantiate dynamic resources.
** 006 MAG 20140410 Add RECTANGLE specific options.
** 007 HC  20140911 GUI coordinate support - data types refactoring.
** 008 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 009 HC  20141021 Implemented window move/position support.
** 010 EVL 20141105 Implementing stubs for setter/getter for rectangle specific attributes:
**                  EDGE-CHARS, EDGE-PIXELS, FILLED, GRAPHIC-EDGE, GROUP-BOX, ROUNDED.
** 011 EVL 20141121 Changed return values from boolean to logical.
** 012 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;

/**
 * Rectangle widget.
 * 
 * @author 	SIY
 * @version 1.0
 */
@LegacyResource(resource = LegacyResource.RECTANGLE)
public class RectangleWidget
extends ControlEntity<RectangleConfig>
{
   /**
    * Default constructor.
    */
   public RectangleWidget()
   {
      this(false);
   }

   /**
    * Default constructor.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   public RectangleWidget(boolean dynamic)
   {
      super(dynamic, new RectangleConfig());
   }

   /**
    * Get EDGE-CHARS attribute.
    *          
    * @return  The value of the EDGE-CHARS attribute.
    */
   public decimal getEdgeChars()
   {
      return new decimal(config.edgeChars);
   }

   /**
    * Get EDGE-PIXELS attribute.
    * 
    * @return  The value of the EDGE-PIXELS attribute.
    */
   public integer getEdgePixels()
   {
      return new integer(config.edgePixels);
   }

   /**
    * Get FILLED attribute.
    * 
    * @return  The value of the FILLED attribute.
    */
   public logical getFilled()
   {
      return new logical(config.filled);
   }
   
   /**
    * Get GRAPHIC-EDGE option.
    * 
    * @return  The value of the GRAPHIC-EDGE attribute.
    */
   public logical getGraphicEdge()
   {
      return new logical(config.graphicEdge);
   }

   /**
    * Get GROUP-BOX attribute.
    * 
    * @return  The value of the GROUP-BOX attribute.
    */
   public logical getGroupBox()
   {
      return new logical(config.groupBox);
   }
   
   /**
    * Get ROUNDED attribute.
    * 
    * @return  The value of the ROUNDED attribute.
    */
   public logical getRounded()
   {
      return new logical(config.rounded);
   }
   
   /**
    * Set FILLED attribute or NO-FILL option.
    * 
    * @param   flag
    *          New value for the FILLED attribute.
    */
   public void setFilled(boolean flag)
   {
      config.filled = flag;
      pushScreenDefinition();
   }

   /**
    * Set FILLED attribute NO-FILL option.
    * 
    * @param   flag
    *          New value for the FILLED attribute.
    */
   @Override
   public void setFilled(logical flag)
   {
      setFilled(flag.booleanValue());
   }

   /**
    * Set EDGE-PIXELS option.
    * 
    * @param   pixels
    *          Option value.
    */
   public void setEdgePixels(int pixels)
   {
      config.edgePixels = pixels;
      pushScreenDefinition();
   }

   /**
    * Set EDGE-PIXELS option.
    * 
    * @param   pixels
    *          Option value.
    */
   @Override
   public void setEdgePixels(integer pixels)
   {
      setEdgePixels(pixels.intValue());
   }

   /**
    * Set EDGE-CHARS option.
    *
    * @param   chars
    *          Option value.
    */
   public void setEdgeChars(double chars)
   {
      config.edgeChars = chars;
      pushScreenDefinition();
   }

   /**
    * Set EDGE-CHARS option.
    * 
    * @param   chars
    *          Option value.
    */
   @Override
   public void setEdgeChars(NumberType chars)
   {
      setEdgeChars(chars.doubleValue());
   }

   /**
    * Set EDGE-CHARS option.
    * 
    * @param   chars
    *          Option value.
    */
   public void setEdgeChars(int chars)
   {
      setEdgeChars((double)chars);
   }

   /**
    * Set GRAPHIC-EDGE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   public void setGraphicEdge(boolean flag)
   {
      config.graphicEdge = flag;
      pushScreenDefinition();
   }

   /**
    * Set GRAPHIC-EDGE option.
    * 
    * @param   flag
    *          New value for the option.
    */
   @Override
   public void setGraphicEdge(logical flag)
   {
      setGraphicEdge(flag.booleanValue());
   }

   /**
    * Set GROUP-BOX attribute.
    * 
    * @param   flag
    *          New value for the GROUP-BOX attribute.
    */
   public void setGroupBox(boolean flag)
   {
      config.groupBox = flag;
      pushScreenDefinition();
   }

   /**
    * Set GROUP-BOX attribute.
    * 
    * @param   flag
    *          New value for the GROUP-BOX attribute.
    */
   @Override
   public void setGroupBox(logical flag)
   {
      setGroupBox(flag.booleanValue());
   }

   /**
    * Set ROUNDED attribute.
    * 
    * @param   flag
    *          New value for the ROUNDED attribute.
    */
   public void setRounded(boolean flag)
   {
      config.rounded = flag;
      pushScreenDefinition();
   }

   /**
    * Set ROUNDED attribute.
    * 
    * @param   flag
    *          New value for the ROUNDED attribute.
    */
   @Override
   public void setRounded(logical flag)
   {
      setRounded(flag.booleanValue());
   }
}
