/*
** Module   : WindowWidget.java
** Abstract : server side window widget implementation
**
** Copyright (c) 2005-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 NVS 20051026   @23498 Created initial version.
** 002 GES 20060803   @28385 Major cleanup and refactoring.
** 003 SIY 20061201   @31528 Ignore attempts to change window visibility
**                           for CHUI.
** 004 GES 20080320   @37595 Added TYPE attribute.
** 005 SVL 20081202   @40751 setVisible(false) should have the same effect
**                           as LogicalTerminal.hideAll(true).
** 006 SIY 20081224   @40990 Replaced hideAll() in setVisible() with call 
**                           to dedicated client API.
** 007 SIY 20101004          Added necessary method stubs.
** 008 SIY 20101005          Icon-related method stubs.
** 009 SIY 20101007          Added asWidget() method.
** 010 SIY 20111101          Updated imports.
** 011 CA  20130112          Extracted into CommonWindow interface all window
**                           APIs.
** 012 CA  20130221          Refactored usage of TYPE attribute.
** 013 CA  20130223          Removed assign, as it window assignment is done using
**                           LogicalTerminal APIs.
** 014 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 015 CA  20130323          The PARENT attribute must be a CommonWidget, not Object.
** 016 CA  20130927          Resource type is determined from LegacyResource annotation.
** 017 CA  20131115          Changes related to properly instantiate dynamic resources.
** 018 MAG 20140512          Implemented new attributes and support for CREATE WINDOW.
** 019 CA  20140701          Added stubbed support for font-related attributes:  MESSAGE-AREA-FONT, 
**                           STATUS-AREA-FONT.
** 020 CA  20140711          Enhanced font support, for GUI clients.
** 021 CA  20140805          Push the window config to the client, if the title gets set.
**                           resourceDelete needs to call LT.deregisterWindow, to cleanup both
**                           server and client-side.
** 022 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 023 HC  20141021          Implemented window move/position support.
** 024 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 025 CA  20141123          Added first pass at dynamic window dimension/location.
** 026 EVL 20141204          Added LOAD-SMALL-ICON implementation.
** 027 HC  20141207          Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.util.*;
 
/**
 * This class represents the only window widget that exists in a CHUI
 * environment.
 * <p>
 * The window widget differs from all the other widgets also in this: it is 
 * the only widget that is not contained in a frame, but rather contains
 * frames (and other top-level components) itself. 
 */
public class WindowWidget
extends PaneEntity<WindowConfig>
implements CommonWindow
{
   /**
    * Default constructor.  It is package private because an instance must
    * only be created in a controllable way.
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    */
   WindowWidget(boolean dynamic)
   {
      super(dynamic, new WindowConfig());
      
      // Register window
      if (dynamic)
      {
         config.visible = false;

         // Register window
         LogicalTerminal.registerWindow(config, this);
      }
   }

   /**
    * Reports if this widget object is valid for use.
    *
    * @return   Always <code>true</code>.
    */
   @Override
   public boolean valid()
   {
      return true;
   }

   /**
    * Gets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @return   The current value of the KEEP-FRAME-Z-ORDER attribute.
    */
   @Override
   public logical isKeepFrameZOrder()
   {
      return new logical(config.keepFrameZOrder);
   }

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @Override
   public void setKeepFrameZOrder(logical keep)
   {
      setKeepFrameZOrder(keep.booleanValue()); 
   }

   /**
    * Sets the KEEP-FRAME-Z-ORDER writable attribute.
    *
    * @param    keep
    *           The new value for the KEEP-FRAME-Z-ORDER attribute.
    */
   @Override
   public void setKeepFrameZOrder(boolean keep)
   {
      config.keepFrameZOrder = keep;
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Gets the FULL-HEIGHT-CHARS writable attribute.
    *
    * @return current value of FULL-HEIGHT-CHARS attribute
    */
   @Override
   public decimal getFullHeightChars()
   {
      throw new RuntimeException("Not implemented");
   }
   
   /**
    * Gets the FULL-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of FULL-HEIGHT-PIXELS attribute
    */
   @Override
   public decimal getFullHeightPixels()
   {
      throw new RuntimeException("Not implemented");
   }
   
   /**
    * Gets the FULL-WIDTH-CHARS writable attribute.
    *
    * @return current value of FULL-WIDTH-CHARS attribute
    */
   @Override
   public decimal getFullWidthChars()
   {
      throw new RuntimeException("Not implemented");
   }
   
   /**
    * Gets the FULL-WIDTH-PIXELS writable attribute.
    *
    * @return current value of FULL-WIDTH-PIXELS attribute
    */
   @Override
   public decimal getFullWidthPixels()
   {
      throw new RuntimeException("Not implemented");
   }
   
   /**
    * Gets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @return current value of MAX-HEIGHT-CHARS attribute
    */
   @Override
   public decimal getMaxHeightChars()
   {
      return new decimal(config.maxHeightChars);
   }

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   @Override
   public void setMaxHeightChars(NumberType max)
   {
      setMaxHeightChars(max.doubleValue());
   }

   /**
    * Sets the MAX-HEIGHT-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-CHARS attribute.
    */
   public void setMaxHeightChars(double max)
   {
      config.maxHeightChars = Coordinate.scale(max);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Gets the MAX-WIDTH-CHARS writable attribute.
    *
    * @return   The current value of MAX-WIDTH-CHARS attribute.
    */
   @Override
   public decimal getMaxWidthChars()
   {
      return new decimal(config.maxWidthChars);
   }

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   @Override
   public void setMaxWidthChars(NumberType max)
   {
      setMaxWidthChars(max.doubleValue());
   }

   /**
    * Sets the MAX-WIDTH-CHARS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-CHARS attribute.
    */
   public void setMaxWidthChars(double max)
   {
      config.maxWidthChars = Coordinate.scale(max);
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Set maximal height.
    * 
    * @param   maxHeight
    *          New option value.
    */
   @Override
   public void setMaxHeight(NumberType maxHeight)
   {
      this.setMaxHeightChars(maxHeight);
   }

   /**
    * Set maximal width.
    * 
    * @param   maxWidth
    *          New option value.
    */
   @Override
   public void setMaxWidth(NumberType maxWidth)
   {
      this.setMaxWidth(maxWidth);
   }

   /**
    * Enable/disable window resizing.
    * 
    * @param   resize
    *          New option value.
    */
   @Override
   public void setResize(logical resize)
   {
      config.resize = resize.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable displaying of scrollbars.
    * 
    * @param   scrollBars
    *          New option value.
    */
   @Override
   public void setScrollBars(logical scrollBars)
   {
      config.scrollBars = scrollBars.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable displaying of status area.
    * 
    * @param   statusArea
    *          New option value.
    */
   @Override
   public void setStatusArea(logical statusArea)
   {
      config.statusArea = statusArea.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets the TITLE writable attribute and pushes the config to the client-side.
    *
    * @param    title
    *           The new value for the TITLE attribute.
    */
   @Override
   public void setTitle(String title)
   {
      config.title = title;
      LogicalTerminal.pushWindow(config);
   }
   
   /**
    * Set (default) background color. 
    */
   public void setBgColor()
   {
      // FIXME what's the default color?
      super.setBgColor(new integer(0));
   }
   
   /**
    * Set (default) foreground color.
    */
   public void setFgColor()
   {
      // FIXME what's the default color?
      super.setFgColor(new integer(0));
   }

   /**
    * Enable/disable KEEP-Z-ORDER option.
    * 
    * @param   keepFrameZOrder
    *          New flag value.
    */
   @Override
   public void setKeepZOrder(logical keepFrameZOrder)
   {
      config.keepFrameZOrder = keepFrameZOrder.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Enable/disable displaying of message area.
    * 
    * @param   messageArea
    *          New flag value.
    */
   @Override
   public void setMessageArea(logical messageArea)
   {
      config.messageArea = messageArea.booleanValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Get window maximal height.
    * 
    * @return  maximal height.
    */
   @Override
   public decimal getMaxHeight()
   {
      return this.getMaxHeightChars();
   }

   /**
    * Get window maximal width.
    * 
    * @return  maximal width.
    */
   @Override
   public decimal getMaxWidth()
   {
      return this.getMaxWidthChars();
   }

   /**
    * Check if window size can be changed.
    * 
    * @return  <code>true</code> if window size can be changed.
    */
   @Override
   public logical isResize()
   {
      return new logical(config.resize);
   }

   /**
    * Check if window has scrollbars.
    * 
    * @return  <code>true</code> if window has scrollbars.
    */
   @Override
   public logical isScrollBars()
   {
      return new logical(config.scrollBars);
   }

   /**
    * Check if window has status area.
    * 
    * @return  <code>true</code> if window has status area.
    */
   @Override
   public logical isStatusArea()
   {
      return new logical(config.statusArea);
   }

   /**
    * Check if window has KEEP-Z-ORDER option set.
    * 
    * @return <code>true</code> if window has KEEP-Z-ORDER option set.
    */
   @Override
   public logical isKeepZOrder()
   {
      return new logical(config.keepFrameZOrder);
   }

   /**
    * Check if window has message area.
    * 
    * @return <code>true</code> if window has message area.
    */
   @Override
   public logical isMessageArea()
   {
      return new logical(config.messageArea);
   }

   /**
    * Get icon file name previously loaded by {@link #loadIcon(character)}.
    * 
    * @return  icon file name.
    */
   @Override
   public character getIcon()
   {
      return new character(config.iconFileName);
   }

   /**
    * Load icon for window.
    * 
    * @param   iconPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadIcon(character iconPath)
   {
      config.iconFileName = iconPath.toStringMessage();
      LogicalTerminal.pushWindow(config);
      return new logical(true);
   }

   /**
    * Load icon for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadIcon(String searchPath)
   {
      return loadIcon(new character(searchPath));
   }

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   iconPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadSmallIcon(character iconPath)
   {
      // TODO: differentiate from 32x32 loader 
      return loadIcon(iconPath);
   }

   /**
    * Load small icon(16x16) for window.
    * 
    * @param   searchPath
    *          Icon file name.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public logical loadSmallIcon(String searchPath)
   {
      return loadSmallIcon(new character(searchPath));
   }

   /**
    * Return reference to {@link WindowWidget} as reference to 
    * {@link GenericWidget}.
    * 
    * @return  this.
    */
   public GenericWidget<?> asWidget()
   {
      return this;
   }

   /**
    * Sets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-HEIGHT-PIXELS attribute.
    */
   @Override
   public void setMaxHeightPixels(integer max)
   {
      config.maxHeightPixels = max.intValue();
      LogicalTerminal.pushWindow(config);
   }

   /**
    * Sets the MAX-WIDTH-PIXELS writable attribute.
    *
    * @param    max
    *           The new value for the MAX-WIDTH-PIXELS attribute.
    */
   @Override
   public void setMaxWidthPixels(integer max)
   {
      config.maxWidthPixels = max.intValue();
      LogicalTerminal.pushWindow(config);
   }
   
   /**
    * Gets the MAX-HEIGHT-PIXELS writable attribute.
    *
    * @return current value of MAX-HEIGHT-PIXELS attribute
    */
   @Override
   public integer getMaxHeightPixels()
   {
      return new integer(config.maxHeightPixels);
   }   

   /**
    * Gets the MAX-WIDTH-PIXELS writable attribute.
    *
    * @return current value of MAX-WIDT-PIXELS attribute
    */
   @Override
   public integer getMaxWidthPixels()
   {
      return new integer(config.maxWidthPixels);
   }
   
   /**
    * Get the message-area's font.
    * 
    * @return   The message-area's font.
    */
   @Override
   public int64 getMessageAreaFont()
   {
      int font = config.messageAreaFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setMessageAreaFont(long fontNum)
   {
      setMessageAreaFont(new int64(fontNum));
   }

   /**
    * Set the message-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setMessageAreaFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.messageAreaFont = font;
      pushScreenDefinition();
   }
   
   /**
    * Get the status-area's font.
    * 
    * @return   The status-area's font.
    */
   @Override
   public int64 getStatusAreaFont()
   {
      int font = config.statusAreaFont;
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setStatusAreaFont(long fontNum)
   {
      setStatusAreaFont(new int64(fontNum));
   }

   /**
    * Set the status-area's font.
    * 
    * @param    fontNum
    *           The font number, an entry in the font-table.
    */
   @Override
   public void setStatusAreaFont(int64 fontNum)
   {
      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());

      config.statusAreaFont = font;
      pushScreenDefinition();
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      boolean delete = super.resourceDelete();

      if (delete)
      {
         LogicalTerminal.deregisterWindow(this);
      }
      
      return delete;
   }

   /**
    * Sets the WIDTH or HEIGHT attribute for this widget.
    *
    * @param    isUnknown
    *           If <code>true<code> the attribute value being set is an unknown value.
    * @param    value
    *           The attribute value.
    * @param    width
    *           If <code>true</code> <code>value</code> represents WIDTH, otherwise 
    *           a HEIGHT.
    */
   @Override
   protected void setWidthOrHeight(boolean isUnknown, double value, boolean width)
   {
      if (LogicalTerminal.isChui())
      {
         String msg = String.format(
               "**Unable to set attribute %s on %s in the current environment.", 
               width ? "WIDTH" : "HEIGHT", type());
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return;
      }
      
      super.setWidthOrHeight(isUnknown, value, width);
   }

   /**
    * Sets the WIDTH-PIXELS or HEIGHT-PIXELS attribute for this widget.
    *
    * @param    isUnknown
    *           If <code>true<code> the attribute value being set is an unknown value.
    * @param    value
    *           The attribute value.
    * @param    width
    *           If <code>true</code> <code>value</code> represents WIDTH-PIXELS, otherwise 
    *           a HEIGHT-PIXELS.
    */
   @Override
   protected void setWidthOrHeightPixels(boolean isUnknown, int value, boolean width)
   {
      if (LogicalTerminal.isChui())
      {
         String msg = String.format(
               "**Unable to set attribute %s on %s in the current environment.", 
               width ? "WIDTH-PIXELS" : "HEIGHT-PIXELS", type());
         ErrorManager.recordOrShowError(4079, msg, false, false, false);
         return;
      }
      
      super.setWidthOrHeightPixels(isUnknown, value, width);
   }

   /**
    * Gets the SCREEN-LINES readable attribute.
    *
    * @return   current value of SCREEN-LINES attribute
    */
   @Override
   public decimal getScreenLines()
   {
      return new decimal(config.screenLines);
   }
}
