/*
** Module   : Editor.java
** Abstract : client side implementation of the Editor component
**
** Copyright (c) 2006-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------------------Description-----------------------------
** 001 SIY 20060428   @26083 First stage of the implementation of the 
**                           EDITOR widget. Initial version is capable 
**                           to properly calculate size and display 
**                           itself taking into account assigned text, 
**                           size (assigned via SIZE phrase and via 
**                           INNER-CHARS/INNER-LINES), box flag and 
**                           scroll-bars flags. No real editing 
**                           operations are implemented.
** 002 SIY 20060515   @26151 Added getColumnLabelText().
** 003 NVS 20060703   @27801 getParent.nextFocus() replaced with
**                           ThinClient.nextFocus() call to maintain
**                           Progress semantics. Similar replacement has
**                           been done to getParent.previousFocus().
** 004 SIY 20060810   @28557 Fixed displaying of enabled but not assigned
**                           widget.
** 005 SIY 20060816   @28644 Second stage of implementation. All editing 
**                           functions are implemented.
** 006 SIY 20060824   @28849 Removed getP2JType() method.
** 007 GES 20061110   @31089 Fixed drawing during editing when the widget
**                           starts in the uninitialized state. There
**                           needed to be a sync request at the end of
**                           the key processing (when we re-draw) too
**                           because wait-for loops won't force a sync
**                           otherwise.
** 008 NVS 20061204   @31556 Switched to using a private widgetFocused
**                           flag instead of widgetFocused
** 009 SIY 20061204   @31545 Fixed handling of backspace.
** 010 NVS 20070426   @33251 Enabled state is checked instead of SENSITIVE
**                           attribute.
** 011 SIY 20070807   @34811 Added isColumnFixed() method.
** 012 SVL 20071221   @36530 Label is shown/hidden together with the
**                           widget.
** 013 EVL 20071224   @36559 Synchronization of the insert mode indicator
**                           with the state of the bottom-right insert mode
**                           area painting when the editor is gaining the
**                           input focus.
** 014 SVL 20080125   @36904 An empty value is set inside the constructor
**                           to prevent NPE upon initialisation.
** 015 EVL 20080303   @37231 Adding code to repaint insert mode indicator
**                           and to turn the insert mode off when the focus
**                           is leaving the editor.
** 016 NVS 20080320   @37616 Fixed drawing of editors in focus when
**                           they are disabled.
** 017 NVS 20080401   @37774 Added isFocusable() method implementation.
** 018 GES 20080421   @38060 I18N charset support for key events.
** 019 SIY 20080711   @39100 Fixed activation behavior to match 4GL one.
** 020 NVS 20080812   @39368 Added fullHeight() method implementation. It
**                           calculates the number of lines of text the
**                           editor is holding.
** 021 CA  20090423   @41899 Added drawCaretUnconditionally(), which draws the
**                           editor caret even if the editor is not in focus.
**                           This is needed to fix special read-only editor
**                           behavior, related to caret.
** 022 SVL 20090424   @41923 Fixed insert indicator state for read-only
**                           mode.
** 023 GES 20090424   @41962 Import change.
** 024 SIY 20090428   @42026 Fixed word wrapping edge case, when word ends 
**                           exactly on the widget boundary.
** 025 SIY 20100619          Switching to own implementation of character UI.
** 026 SIY 20100804          Changes in imports, some refactorings.
** 027 SIY 20100822          Refactored AttributeContainer support.
** 028 SIY 20100915          Renamed base class.
** 029 SIY 20101122          Switching from KeyEvent to KeyInput.
** 030 SIY 20110112          Changes related to new color handling scheme.
** 031 SIY 20111101          Updated imports, minor cleanups/refactorings.
** 032 SIY 20111202          Refactored and moved from CHUI package.
** 033 GES 20140627          Changed color attribute method names to match naming standards.
** 034 CA  20140805          Added generics to force the abstract widget classes to not use
**                           implementation-specific code: the abstract widgets will have access
**                           only to the base APIs provided by the OutputManager.
** 035 HC  20140911          GUI coordinate support - data types refactoring.
** 036 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 037 HC  20141021          Implemented window move/position support.
** 038 HC  20141207          Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;
import com.goldencode.util.*;

/**
 * Implementation of Progress EDITOR component.
 * Besides what was known from Progress manuals, the following facts describe
 * some peculiarities in the redirected output found experimentally:
 * <ul>
 *   <li>If the number of lines the editor holds is less than or equal to the
 *       height of the widget, the output will consist of the text and some  
 *       empty lines so that the height of the widget equals the number of
 *       output lines.
 *   <li>If the number of lines the editor holds is greater than the height of  
 *       the widget, all lines will be output, not just the visible portion.
 *   <li>No matter how many lines of text the editor holds, the visible lines
 *       is what is counted when placing the frame on the current page.
 *   <li>Unlike other widgets, editors can split the output into multiple
 *       pages. The page overflow condition is checked after every line of
 *       the editor's output. If the current page spills, the form feed 
 *       character is inserted as well as frame headers, if any. Page headers
 *       appear on the page as soon as the editor's output terminates, but no
 *       form feed character is printed with the page headers in this case.
 *   <li>If the editor is long enough, multiple page breaks and frame headers
 *       will be created.
 *   <li>Iterations of the down frames containing editors may have editors of
 *       various sizes.
 *   <li>Every iteration of the down frames containing editors can cause page
 *       break as described.
 * </ul>
 */
public abstract class Editor<O extends OutputManager<?>>
extends LabeledDataContainer<O>
{
   /** Editor widget TAB size in spaces. */
   private static final int EDITOR_TAB_SIZE = 4;

   /** Component attribute container. */
   protected EditorConfig config = null;

   /** Editor content. */
   protected String content = new String("");

   /** Editor content parsed into separate string. */
   private Line[] parsedText = null;
   
   /** Cursor X position relative to left edge of the inner rectangle of widget. */
   protected int cursorX = 0;
   
   /** Cursor Y position relative to to edge of the inner rectangle of widget. */
   protected int cursorY = 0;
   
   /** Horizontal scrolling of the text. */
   private int scrollX = 0;
   
   /** Vertical scrolling of the text. */
   protected int scrollY = 0;

   /** Insert mode. */
   protected boolean insert = true;

   /**
    * If true then the editor is into read-only mode and INSERT key hasn't
    * been pressed. In this case we should display OFF state while actual
    * insert state may be either ON or OFF. 
    **/
   private boolean fakeInsert = false;

   /** Widget in focus flag. */
   protected boolean widgetFocused = false;

   /**
    * Special constructor used to restore component using given configuration.
    * <p>
    * Call {@link #initialize} to finish the widget initialization.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public Editor(WidgetId id, EditorConfig cfg)
   {
   }
   
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig cfg)
   {
      config = (EditorConfig) ConfigManager.getInstance().resolveWidgetConfig(id, cfg);
      ConfigManager.getInstance().replaceConfig(config, cfg);
      
      setValue(new character(content));
      addFocusListener(new LocalFocusListener());
      setEnabled(false);
   }
   
   /**
    * Getting the BaseDataType object representing the current value.
    *
    * @return  The value representing the current widget value.
    */
   @Override
   public BaseDataType getValue()
   {
      return new character(content);
   }
   
   /**
    * Requesting the focus to the widget.
    */
   @Override
   public void requestFocus()
   {
      if (!isEnabled())
         return;
      
      super.requestFocus();
   }
   
   /**
    * Provide access to the attributes stored in the widget.
    * 
    * @return  Reference to attribute container.
    */
   @SuppressWarnings("unchecked")
   @Override
   public EditorConfig config()
   {
      return config;
   }
   
   /**
    * Assign new instance of <code>Label</code> to the fill-in widget.
    * 
    * @param   label
    *          Reference to new <code>Label</code> instance.
    */
   @Override
   public void setLabelInstance(Label<O> label)
   {
      super.setLabelInstance(label);
      
      if (label != null)
      {
         label.config().dcolor = config().dcolor;
         label.config().pfcolor = config().pfcolor;
      }
   }

   /**
    * Sets position of terminal's cursor.
    */
   @Override
   public void drawCaret()
   {
      // this check is required since the method can be called from multiple
      // drivers
      if (!hasFocus() || !isEnabled())
      {
         return;
      }

      drawCaretUnconditionally();
   }

   /**
    * Sets position of terminal's cursor unconditionally (even if this editor
    * is not in focus).
    */
   public abstract void drawCaretUnconditionally();

   /**
    * Draws text. <code>DisplayFormat.Presentation</code> instance is
    * rendered into screen value. Screen value is left justified or padded to
    * the right bound with spaces. Non first lines are filled with spaces. If
    * widget is focused then after rendering terminal's cursor is set to
    * appropriate position in the widget.
    */
   @Override
   public void draw()
   {
      Color dc = UiUtils.getDColor(this);
      Color pf = config.pfcolor;

      Color color = (widgetFocused && isEnabled()) ? pf:dc;

      if (focusTraversable() /*&& useUnderline*/)
         color = color.withUnderline();

      draw(color);
   }

   /**
    * Actual drawing routine.
    *
    * @param   color
    *          Color for drawing.
    */
   public abstract void draw(Color color);

   /**
    * Prepare text for for displaying.
    * 
    * @param   uninit
    *          
    * @param   width
    *          Width of the displayed area.
    * @param   space
    *          Prepared string filled with spaces.
    * @param   line
    *          Index in the array of parsed strings.
    *
    * @return  Text string ready for displaying.
    */
   protected String prepareTextRow(boolean uninit, int width, String space, int line)
   {
      String text;
      if (line < parsedText.length)
      {
         if (parsedText[line].length() < scrollX || uninit)
            text = space;
         else
         {
            text = parsedText[line].getText() + space;
            text = text.substring(scrollX, scrollX + width);
         }
      }
      else
         text = space;
      return text;
   }

   /**
    * Get screen representation of the data.
    * 
    * @return  Text representation of the variable value according to all
    *          current settings, flags, format, etc.
    */
   @Override
   public String getText()
   {
      return content;
   }

   /**
    * Clear input variable.
    */
   public void clear()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      config.setModified(true);
      cursorHome();
      parseContent("", -1, false);
   }

   /**
    * This component will receive an input focus when Tab or Shift-Tab is
    * pressed.
    * 
    * @return  <code>true</code> if this component is enabled and visible.
    */
   @Override
   public boolean focusTraversable()
   {
      return isEnabled() && isVisible();
   }

   /**
    * Returns this component directly because it is not an usual container and
    * have no child components.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#currentFocus()
    */
   @Override
   public Widget<O> currentFocus()
   {
      return this;
   }

   /**
    * Returns this component directly because it is not an usual container and
    * have no child components.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Container#defaultFocus()
    */
   @Override
   public Widget<O> defaultFocus()
   {
      return this;
   }

   /**
    * Sets the new variable value in the fill-in.
    * 
    * @param   value
    *          New value for the screen buffer.
    */
   @Override
   public void setValue(BaseDataType value)
   {
      if (value == null)
         return;
      
      cursorHome();
      parseContent(value.toStringMessage(), -1, false);
   }

   /**
    * Process KeyEvents for this widget.
    * 
    * @param   ke
    *          Key event structure.
    */
   @Override
   public void processKeyEvent(KeyInput ke)
   {
      // Call all KeyListener objects that may have been registered
      // for this component.
      super.processKeyEvent(ke);
      boolean forceRedraw = false;

      // check if any of the KeyListeners consumed the KeyEvent
      if (ke.isConsumed() || !isEnabled())
      {
         return;
      }

      int key = ke.actionCode();
      
      // handle all printable characters and '\n'
      if ((key >= 0x20 && key <= 0xff) || 
          key == Keyboard.KA_RETURN    ||
          key == Keyboard.KA_NEW_LINE)
      {
         if (config.readOnly)
         {
            return;
         }
         
         config.setModified(true);

         char ch = Utils.toChar(key);

         if (key == Keyboard.KA_RETURN ||
             key == Keyboard.KA_NEW_LINE)
         {
            ch = '\n';
            
            if (key == Keyboard.KA_NEW_LINE)
               cursorEndOfLine();
         }
         
         placeChar(ch);
         
         forceRedraw = true;
      }
      else
      {
         if (!hasFocus())
            return;
      }
      
      // key actions
      
      if (key == Keyboard.KA_INSERT_MODE)
      {
         fakeInsert = false;
         insert = !getInsertMode();
         setInsertMode(insert);
         drawCaret();
         requestSync();
         return;
      }
      else if (key == Keyboard.KA_TAB)
      {
         // the following call should be conditional
         ThinClient.nextFocus(this);
         return;
      }
      else if (key == Keyboard.KA_BACK_TAB)
      {
         // the following call should be conditional
         ThinClient.previousFocus(this);
         return;
      }
      else if (key == Keyboard.KA_CURSOR_DOWN)
      {
         cursorDown();
      }
      else if (key == Keyboard.KA_CURSOR_UP)
      {
         cursorUp();
      }
      else if (key == Keyboard.KA_CURSOR_LEFT)
      {
         cursorLeft();
      }
      else if (key == Keyboard.KA_CURSOR_RIGHT)
      {
         cursorRight();
      }
      else if (key == Keyboard.KA_HOME)
      {
         cursorHome();
      }
      else if (key == Keyboard.KA_LEFT_END)
      {
         cursorBeginOfLine();
      }
      else if (key == Keyboard.KA_RIGHT_END)
      {
         cursorEndOfLine();
      }
      else if (key == Keyboard.KA_BACKSPACE)
      {
         backspace();
      }
      else if (key == Keyboard.KA_DELETE_CHARACTER)
      {
         deleteCharacter();
      }
      else if (key == Keyboard.KA_CLEAR)              
      {
         clear();
      }
      else if (key == Keyboard.KA_EDITOR_TAB)
      {
         editorTab();
      }
      else if (key == Keyboard.KA_EDITOR_BACKTAB)
      {
         editorBackTab();
      }
      else if (key == Keyboard.KA_APPEND_LINE)
      {
         appendLine();
      }
      else if (key == Keyboard.KA_DELETE_END_LINE)    
      {
         if (absoluteX() < getLine().length())
            deleteToEndOfLine();   
         else
            bell();
      }
      else if (key == Keyboard.KA_DELETE_LINE)        
      {
         cursorBeginOfLine();
         deleteToEndOfLine();
         deleteCharacter();
      }
      else if (key == Keyboard.KA_END)
      {
         while (absoluteY() < lastRow())
            cursorDown();
         cursorEndOfLine();
      }
      else if (key == Keyboard.KA_DELETE_WORD)
      {
         // note different handling of word separators in comparison
         // to word-left word-right operations
         deleteWord();
      }
      else if (key == Keyboard.KA_PREV_WORD)
      {
         wordLeft();
      }
      else if (key == Keyboard.KA_NEXT_WORD)
      {
         wordRight();
      }
      else if (key == Keyboard.KA_OPEN_LINE_ABOVE)
      {
         openLineAbove();
      }
      else if (key == Keyboard.KA_PAGE_DOWN)
      {
         cursorPageDown();
      }
      else if (key == Keyboard.KA_PAGE_UP)
      {
         cursorPageUp();
      }
      else if (key == Keyboard.KA_PAGE_LEFT)
      {
         if (absoluteX() < config.innerChars)
            cursorBeginOfLine();
         else
            setCursor(absoluteX() - config.innerChars + 1);
      }
      else if (key == Keyboard.KA_PAGE_RIGHT)
      {
         if (absoluteX() + config.innerChars >= config.bufferChars - 1)
            cursorEndOfLine();
         else
            setCursor(absoluteX() + config.innerChars - 1);
      }
      else if (key == Keyboard.KA_BELL)
      {
         bell();
      }
      else if (!forceRedraw)
      {
         // ignore all unknown actions
         return; 
      }

      setInsertMode(insert);
      
      draw();
      requestSync();
   }

   /**
    * Handle APPEND-LINE event.
    */
   private void appendLine()
   {
      if (absoluteY() < lastRow() && 
          getLine().length() + getNextWordSize() < config.bufferChars)
      {
         cursorEndOfLine();
         placeChar(' ');
         deleteCharacter();
      }
   }

   /**
    * Get length of the first word in the next row.
    * 
    * @return  The length of the first word in the row immediately under the
    *          current.
    */
   private int getNextWordSize()
   {
      if (absoluteY() >= lastRow())
         return config.bufferChars;
      
      int sz = 0;
      
      char[] text = parsedText[absoluteY() + 1].getText().toCharArray();
      
      while (sz < text.length && text[sz] == ' ')
         sz++;
      
      while (sz < text.length && text[sz] != ' ')
         sz++;
      
      return sz;
   }

   /**
    * Scroll one page up.
    */
   private void cursorPageUp()
   {
      if (absoluteY() >= config.innerLines - 1)
      {
         scrollY -= config.innerLines - 1;
         
         if (scrollY < 0)
            scrollY = 0;
      }
   }

   /**
    * Scroll one page down.
    */
   private void cursorPageDown()
   {
      if ((scrollY + config.innerLines - 1) < lastRow())
      {
         scrollY += config.innerLines - 1;
         
         if (scrollY + config.innerLines - 1 > lastRow())
            scrollY = lastRow() - config.innerLines + 1;
      }
   }

   /**
    * Handle EDITOR-BACKTAB event. 
    */
   private void editorBackTab()
   {
      if (absoluteX() == 0)
      {
         bell();
         return;
      }
      
      if (absoluteX() < EDITOR_TAB_SIZE)
      {
         cursorBeginOfLine();
         return;
      }
      
      int pos = absoluteX()/EDITOR_TAB_SIZE - 1;
      pos *= EDITOR_TAB_SIZE;
      
      setCursor(pos);
   }

   /**
    * Handle EDITOR-TAB event.
    */
   private void editorTab()
   {
      if (insert)
      {
         // insert spaces
         if (getLine().length() + EDITOR_TAB_SIZE <= 
             config.bufferChars)
         {
            if (config.readOnly)
            {
               bell();
               return;
            }
            
            config.setModified(true);
            
            int pos = (absoluteX()/EDITOR_TAB_SIZE) + 1;
            pos *= EDITOR_TAB_SIZE;
            
            StringBuffer buff = new StringBuffer(content);
            
            int len = pos - absoluteX();
            int inspos = absoluteX();
            
            if (absoluteX() > buff.length())
            {
               len += absoluteX() - getLine().length();
               inspos = getLine().length();
            }
            
            buff.insert(inspos, StringHelper.repeatChar(' ', len));

            parseContent(buff.toString(), absoluteY(), false);
            
            setCursor(pos);
         }
         else
            bell();
      }
      else
      {
         // move cursor
         if (absoluteX() + EDITOR_TAB_SIZE <= 
             config.bufferChars)
         {
            int pos = (absoluteX()/EDITOR_TAB_SIZE) + 1;
            pos *= EDITOR_TAB_SIZE;

            setCursor(pos);
         }
         else
            bell();
      }
   }

   /**
    * Set cursor position. If position is not near the edge, then cursor is
    * positioned about at the middle of screen.
    * 
    * @param   pos
    *          New cursor position.
    */
   private void setCursor(int pos)
   {
      if (pos < scrollX || pos >= scrollX + config.innerChars)
      {
         // scrolling is necessary
         cursorX = (config.innerChars - 1)/2 - 1;
         scrollX = pos - cursorX;
         
         if (scrollX < 0 || 
            (scrollX + config.innerChars) >= config.bufferChars)
         {
            if (scrollX < 0)
            {
               cursorX -= scrollX;
               scrollX = 0;
            }
            else
            {
               cursorX = config.innerChars - 
                         config.bufferChars + pos - 1;
               scrollX = pos - cursorX;
            }
         }
      }
      else
      {
         cursorX = pos - scrollX;
      }
   }

   /**
    * Delete word. 
    */
   private void deleteWord()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      config.setModified(true);

      if (absoluteX() < getLine().length())
      {
         char[] text = getLine().getText().toCharArray();
         
         int start = absoluteX();
         int end = start;
         
         // note that order of checks in both loops is important
         
         while (end < text.length && !Character.isLetterOrDigit(text[end]))
         {
            end++;
         }
         
         while (end < text.length && Character.isLetterOrDigit(text[end]))
         {
            end++;
         }
         
         if (end > start)
         {
            StringBuffer buff = new StringBuffer(content);
            
            buff.delete(getLine().getOffset() + start, 
                        getLine().getOffset() + end);
            
            parseContent(buff.toString(), absoluteY(), false);
         }
      }
   }

   /**
    * Move cursor one word left.
    */
   private void wordLeft()
   {
      if (absoluteX() == 0)
      {
         if (absoluteY() > 0)
            cursorEndOfPrevLine();
      }
      else
      {
         char[] text = getLine().getText().toCharArray();
         
         if (absoluteX() > text.length)
            cursorEndOfLine();
         
         cursorLeft();
         
         if (absoluteX() < text.length)
            while (absoluteX() >0 && text[absoluteX()] == ' ')
               cursorLeft();
         
         if (absoluteX() == 0)
         {
            if (absoluteY() > 0)
               cursorEndOfPrevLine();
         }
         else
         {
            while (absoluteX() > 0 && text[absoluteX() - 1] != ' ')
               cursorLeft();
         }
      }
   }

   /**
    * Move cursor one word right.
    */
   private void wordRight()
   {
      if (absoluteX() >= getLine().length())
      {
         if (absoluteY() < lastRow())
         {
            cursorDown();
            cursorBeginOfLine();
         }
      }
      else
      {
         char[] text = getLine().getText().toCharArray();

         cursorRight();
         
         while (absoluteX() < text.length && text[absoluteX()] != ' ')
            cursorRight();

         while (absoluteX() < text.length && text[absoluteX()] == ' ')
            cursorRight();
      }
   }

   /**
    * Insert empty line above current line. 
    */
   private void openLineAbove()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      config.setModified(true);

      StringBuffer buff = new StringBuffer(content);
      
      buff.insert(getLine().getOffset(), '\n');
      
      parseContent(buff.toString(), absoluteY(), false);
      
   }

   /**
    * Maximal value of the absolute Y coordinate.
    * 
    * @return  The last row index.
    */
   private int lastRow()
   {
      return parsedText.length - 1;
   }

   /**
    * Returns the number of lines the editor is holding.
    * 
    * @return  number of lines of text.
    */
   public int fullHeight()
   {
      int len = parsedText.length;

      // do not count the last line if it's empty
      if (len > 0)
      {
         Line line = parsedText[len - 1];
         if (line.getText().length() == 0 && !line.isNewLine())
         {
            len--;
         }
      }

      return len;
   }

   /**
    * Move cursor to beginning of the line.
    */
   private void cursorBeginOfLine()
   {
      cursorX = 0;
      scrollX = 0;
   }

   /**
    * Delete to end of line.
    */
   private void deleteToEndOfLine()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      config.setModified(true);
      
      if ((getLine().length() - absoluteX()) > 0)
      {
         StringBuffer buff = new StringBuffer(content);
         buff.delete(getLine().getOffset() + absoluteX(), 
                     getLine().getOffset() + getLine().length());
         
         parseContent(buff.toString(), absoluteY(), false);
      }
   }

   /**
    * Handle BELL event.
    */
   private void bell()
   {
      screen().beep();
   }

   /**
    * Handle BACKSPACE key.
    */
   private void backspace()
   {
      boolean checkMove = true;
      
      if (absoluteX() == 0 && cursorY > 0)
      {
         cursorEndOfPrevLine();
         checkMove = false;
      }

      int col = absoluteX();
      
      // move cursor first
      
      cursorLeft();
      
      if (checkMove && col == absoluteX())
         return;

      if (!checkMove && getLine().getText().charAt(absoluteX()) != ' ')
      {
         //We're about to merge lines
         cursorRight();
      }
      
      deleteCharacter();
   }

   /**
    * 
    */
   private void cursorEndOfPrevLine()
   {
      cursorUp();
      cursorEndOfLine();
   }

   /**
    * Move cursor to the start position. 
    */
   private void cursorHome()
   {
      cursorX = 0;
      cursorY = 0;
      scrollX = 0;
      scrollY = 0;
   }

   /**
    * Move cursor to the end of line.
    */
   private void cursorEndOfLine()
   {
      int len = getLine().length();
      
      if (scrollX < len && len < (scrollX + config.innerChars))
         cursorX = len - scrollX;
      else
      {
         if (len > config.innerChars)
         {
            cursorX = config.innerChars - 1;
            scrollX = len - cursorX;
         }
         else
         {
            cursorX = len;
            scrollX = 0;
         }
      }
   }

   /**
    * Get reference to the line under absolute cursor position.
    * 
    * @return  The instance of the <code>Line</code> at current position.
    */
   private Line getLine()
   {
      return parsedText[absoluteY()];
   }

   /**
    * Split absolute horizontal cursor position into two components - position
    * of the first visible character and relative cursor position in the
    * visible region.
    * 
    * @param   pos
    *          The new absolute cursor position.
    */
   private void cursorPos(int pos)
   {
      cursorX = pos % config.innerChars;
      scrollX = pos - cursorX;
   }

   /**
    * Move cursor one row down, if possible.
    */
   private void cursorDown()
   {
      int low = Math.min(parsedText.length, config.innerLines);
      
      // cursor does not reach boundary
      if (cursorY < (low - 1))
         cursorY++;
      else
      {
         // Cursor reached the low boundary
         if (parsedText.length > config.innerLines && 
             parsedText.length > (scrollY + cursorY + 1))
         {
            // scroll down
            scrollY++;
         }
      }
   }

   /**
    * Move cursor one row up, if possible.
    */
   private void cursorUp()
   {
      // move cursor up
      if (cursorY > 0)
         cursorY--;
      else
         // scroll up and keep cursor at the first row
         if (scrollY > 0)
            scrollY--;
   }

   /**
    * Move cursor one char right.
    */
   private void cursorRight()
   {
      if (cursorX < (config.innerChars - 1))
         cursorX++;
      else
         if (absoluteX() < config.bufferChars)
            scrollX++;
   }

   /**
    * Move cursor one char left, if possible.
    */
   private void cursorLeft()
   {
      if (cursorX > 0)
         cursorX--;
      else
         if (scrollX > 0)
            scrollX--;
   }

   /**
    * Place character at current cursor position. Depending on the current
    * insertion mode character can be either inserted into text or can replace
    * existing character.
    * 
    * @param   ch
    *          Character to place.
    */
   private void placeChar(char ch)
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      config.setModified(true);
      
      StringBuffer buff = new StringBuffer(content);
      Line current = getLine();
      
      int pos = current.getOffset();
      
      if (absoluteX() >= current.length())
      {
         int len = absoluteX() - current.length();
         
         if (!(insert || ch == '\n'))
            len++;

         if (len > 0)
            buff.insert(pos + current.length(), 
                        StringHelper.repeatChar(' ', len));
      }
      
      pos += absoluteX();
      
      if (insert || ch == '\n')
         buff.insert(pos, ch);
      else
         buff.setCharAt(pos, ch);
      
      parseContent(buff.toString(), absoluteY(), insert);
      
      if (absoluteX() >= getLine().length())
      {
         cursorPos(absoluteX() - getLine().length() + (ch == '\n' ? 0:1));
         cursorDown(); 
      }
      else
         cursorRight();
   }

   /**
    * Get absolute horizontal position in text.
    * 
    * @return  The horizontal position in the text.
    */
   private int absoluteX()
   {
      return cursorX + scrollX;
   }

   /**
    * Get absolute vertical position in text.
    * 
    * @return  The vertical position in the text.
    */
   private int absoluteY()
   {
      return cursorY + scrollY;
   }
   
   /**
    * Delete character at current cursor position. 
    */
   private void deleteCharacter()
   {
      if (config.readOnly)
      {
         bell();
         return;
      }
      
      config.setModified(true);
      
      Line current = getLine();
      
      if (absoluteX() >= current.length())
      {
         // at the end of the text, just ignore the request
         if (absoluteY() == lastRow())
            return;
         
         // move cursor at the end of the text
         cursorEndOfLine();
      }
      
      if (absoluteX() < current.length())
      {
         if (absoluteX() == current.length() && !current.isNewLine())
         {
            cursorLeft();
         }
      }

      StringBuffer buff = new StringBuffer(content);
      
      if (current.getOffset() + absoluteX() < content.length())
      {
         buff.deleteCharAt(current.getOffset() + absoluteX());

         parseContent(buff.toString(), absoluteY(), false);
         
         int newdelta = getLine().length() - absoluteX();
         
         if (newdelta < 0)
         {
            cursorDown();
            cursorBeginOfLine();
            
            for (int i = 0; i > newdelta; i--)
               cursorRight();   
         }
      }
   }
   /**
    * Calculate horizontal position of the scroll-bar thumb.
    * 
    * @param   size
    *          The width of the scroll-bar.
    * 
    * @return  The horizontal position of the scroll-bar thumb.
    */
   protected double calcHpos(double size)
   {
      double res = size / (config.bufferChars + 1);
      res *= absoluteX() + 0.5;
      
      return Coordinate.scale(res);
   }

   /**
    * Calculate vertical position of the scroll-bar thumb.
    * 
    * @param   size
    *          The height of the scroll-bar.
    * 
    * @return  The vertical position of the scroll-bar thumb.
    */
   protected double calcVpos(double size)
   {
      double res = size / parsedText.length;
      res *= absoluteY() + 0.5;
      
      return Coordinate.scale(res);
   }
 
   /**
    * Returns the insert mode state.
    * 
    * @return  The insert mode state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   private boolean getInsertMode()
   {
      return Window.getInsertMode() == Window.INSERT_MODE_INS;
   }

   /**
    * Changes the insert mode state in the window.
    * 
    * @param   mode
    *          New state.
    *          <ul>
    *             <li><code>false</code>- overwrite mode,
    *             <li><code>true</code>- insert mode.
    *          </ul>
    */
   protected void setInsertMode(boolean mode)
   {
      if (config.readOnly && fakeInsert)
      {
         Window.setInsertMode(Window.INSERT_MODE_OVR, true);
      }
      else
      {
         Window.setInsertMode(mode ? Window.INSERT_MODE_INS:
                                    Window.INSERT_MODE_OVR,
                              true);
      }
   }

   /**
    * Parse content of the string and prepare array of displayable values.
    * 
    * @param   source
    *          Source string to parse.
    * @param   row
    *          Row index to start parsing from.
    * @param   placeChar
    *          Set to <code>true</code> if called from
    *          {@link #placeChar(char)} method in insert mode.
    */
   private void parseContent(String source, int row, boolean placeChar)
   {
      ArrayList<Line> lines = new ArrayList<Line>();
      
      for (int i = 0; i < row; i++)
         lines.add(parsedText[i]);
      
      char[] text = source.toCharArray();
      
      int start  = 0;
      int offset = 0;
      int nchars = 0;
      
      if (row >= 0)
         start = offset = parsedText[row].getOffset();
      
      for (int i = offset; i < text.length; i++)
      {
         if (text[i] == '\n')
         {
            nchars = 0;
            lines.add(new Line(text, start, i - start, true));
            start = i + 1;
            continue;
         }

         if (i > start && text[i - 1] == ' ' && text[i] != ' ')
         {
            // check for word wrapping
            int j = i;

            while (j < text.length)
            {
               if (text[j] == ' '  || 
                   text[j] == '\n' || 
                   (j - start) > config.bufferChars)
               {
                  break;
               }
               j++;
            }
            
            if (j < text.length)
            {
               if (text[j] == ' ' ||
                   text[j] == '\n')
               {
                  j--;
               }
            }

            if ((j - start) >= config.bufferChars)
            {
               nchars = 1;
               lines.add(new Line(text, start, i - start, false));
               start = i;
               continue;
            }
         }

         // last check
         nchars++;

         if (nchars > config.bufferChars)
         {
            nchars = 1;
            lines.add(new Line(text, start, i - start, false));
            
            if (text[i] == ' ')
               i++;
            
            start = i;
         }
      }

      if (lines.size() != 0 && nchars == 0)
      {
         Line last = lines.get(lines.size() - 1);
         
         if (last.isNewLine())
         {
            lines.add(new Line("", start, false));
            placeChar = false;
         }
      }

      if (text.length - start > 0 || lines.size() == 0)
      {
         lines.add(new Line(text, start, text.length - start, false));
         start = text.length;
      }
      
      if (placeChar && absoluteX() == (config.innerChars - 1))
         lines.add(new Line("", start, false));

      parsedText = lines.toArray(new Line[lines.size()]);
      content = source;
   }
   
   /**
    * Some actions on focus hanging.
    */
   private class LocalFocusListener implements FocusListener
   {
      /**
       * Setups <code>FillIn</code> as it is just activated by Tab key.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusGained(FocusEvent e)
      {
         if (e == null)
            return;
       
         widgetFocused = true;
         
         ThinClient.setHelp(config.help);

         if (config.readOnly)
         {
            fakeInsert = true;
         }

         setInsertMode(insert);
         repaint();
      }

      /**
       * Reset SELECTED attribute when focus is lost.
       * 
       * @param   e
       *          FocusEvent instance (ignored).
       */
      @Override
      public void onFocusLost(FocusEvent e)
      {
         if (e == null)
            return;
         
         widgetFocused = false;
         Window.setInsertMode(Window.INSERT_MODE_OFF);
      }
   }
   
   /**
    * Internal text representation element.
    */
   private static class Line
   {
      /** Initial character index in the source text. */
      private int start;
      
      /** Line text. */
      private String text;
      
      /**
       * <code>true</code> if instance was created for '\n' in the source
       * text.
       */
      private boolean newLine;
      
      /**
       * Construct an instance from given character array, offset and length.
       * 
       * @param   text
       *          Source character array.
       * @param   start
       *          Initial character offset.
       * @param   length
       *          Line length.
       * @param   newLine
       *          <code>true</code> means that instance will is created for
       *          the text line with '\n' at the end. The character itself
       *          should not be included into <code>length</code> parameter.
       */
      Line(char[] text, int start, int length, boolean newLine)
      {
         this.start   = start;
         this.newLine = newLine;
         
         if (text != null && length > 0)
            this.text = new String(text, start, length);
         else
            this.text = "";
      }

      /**
       * Construct an instance for given string and offset.
       * 
       * @param   text
       *          Source string.
       * @param   start
       *          Initial offset.
       * @param   newLine
       *          <code>true</code> means that instance will is created for
       *          the text line with '\n' at the end.
       */
      Line(String text, int start, boolean newLine)
      {
         this.start   = start;
         this.newLine = newLine;
         
         if (text != null)
            this.text = text;
         else
            this.text = "";
      }
      
      /**
       * Return <code>true</code> if instance represents the text with the
       * '\n' at the end.
       * 
       * @return  The state of the flag.
       */
      public boolean isNewLine()
      {
         return newLine;
      }
      
      /**
       * Get text length.
       * 
       * @return  The text length.
       */
      public int length()
      {
         return text.length();
      }
      
      /**
       * Return text.
       * 
       * @return  The line text.
       */
      public String getText()
      {
         return text;
      }
      
      /**
       * Return offset of this instance in the source text.
       * 
       * @return  The index of the first line character in the source text.
       */
      public int getOffset()
      {
         return start;
      }
   }
}
