/*
** Module   : WidgetRegistry.java
** Abstract :
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20101015 Widget registry code from Screen moved into separate class. 
** 002 SIY 20111101 Moved from CHUI package. Minor cleanups/refactorings in preparation to GUI 
**                  client.
** 003 SIY 20111202 Updated imports.
** 004 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
** 005 EVL 20140812 In new widget ID allocation approach ID for default window is taking from
**                  WidgetId class static constant.
** 006 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 007 VIG 20140919 Changed target window to which a frame is added.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 HC  20141207 Improved config state client-server synchronization.
*/

package com.goldencode.p2j.ui.client;

import java.util.*;
import java.util.logging.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.widget.*;
import com.goldencode.p2j.util.*;

/**
 * Widget registry.
 * <p>
 * This class holds all widgets (frames and their inner widgets) which were 
 * pushed from server to client. Also, this class is responsible for building
 * client side widgets from widget definition.  
 * 
 * @author 	SIY
 * @version 1.0
 */
public class WidgetRegistry<O extends OutputManager<?>>
{
   /** Logger. */
   private static final Logger LOG = LogHelper.getLogger(WidgetRegistry.class);
   
   /** The list of the widgets in the current screen. */
   private LinkedHashMap<WidgetId, Widget<O>> widgetList = new LinkedHashMap<>();

   /**
    * Constructor. 
    */
   public WidgetRegistry()
   {
   }

   /**
    * Recreates the widget based on the information can be taken from the
    * source widget snapshot. This snapshot must contain all required
    * information different from the default one which is enough to
    * reconstruct the widget of the given type with the desirable behavior.
    * 
    * @param   config
    *          Widget snapshot taken from the server.
    */
   @SuppressWarnings("unchecked")
   public void reconstructWidget(WidgetConfig config)
   {
      WidgetId cid = config.id;
      Widget<O> comp = getComponent(cid);
      
      // the ID is no longer required - any change made to this config instance must not affect 
      // the associated widget!
      config.id = null;
      
      if (comp != null)
      {
         // existing component, just apply new config values
         Widget<O> frame = (Widget<O>) UiUtils.locateFrame(comp);
         
         if (frame instanceof Frame)
         {
            ((Frame<O>) frame).applyConfig(comp, config);
         }
         else
         {
            ConfigManager.getInstance().replaceConfig(comp.config(), config);
         }
      }
      else
      {
         // new component
         comp = (Widget<O>) UiUtils.reconstructComponent(cid, config);

         if (comp != null)
         {
            widgetList.put(cid, comp);
            
            comp.initialize(cid, config);

            WidgetConfig wcfg = comp.config();
            
            // Save the base state, during initialization the widget/config state may change.
            // Without the proper base state the changes would not be sent to the server.
            ConfigManager.getInstance().addWidgetConfig(wcfg, config);
         }
      }
      
      // restore the id
      config.id = cid;
   }

   /**
    * Add pre-built frame.
    * 
    * @param   frame
    *          Frame reference.
    */
   public void addFrame(Frame<O> frame)
   {
      if (frame == null)
         return;
   
      Widget<O>[] list = frame.getContentPane().widgets();
      
      widgetList.put(frame.config().id, frame);
   
      for (Widget<O> comp : list)
      {
         if (UiUtils.hasConfig(comp))
         {
            widgetList.put(comp.config().id, comp);
         }
      }
   }

   /**
    * Get all components defined in the screen.
    * 
    * @return  Array of component.
    */
   @SuppressWarnings("unchecked")
   public Widget<O>[] getComponents()
   {
      return widgetList.values().toArray(new Widget[widgetList.size()]);
   }

   /**
    * Remove the given frame and all contained components from the master
    * list of components.
    * 
    * @param    frame
    *           The frame to remove.
    */
   public void removeFrame(Frame<O> frame)
   {
      if (getComponent(frame.config().id) != null)
      {
         // remove the frame from its parent window first
         TitledWindow<O> pwin = frame.ancestor();
         if (pwin != null)
         {
            pwin.removeWidget(frame);
         }
         else
         {
            LOG.log(Level.WARNING, 
                    "Removing frame " + frame.toString() + " with no parent window.");
         }

         WidgetId base = frame.config().id;
         
         // and remove all contained widgets
         Widget<O>[] w = frame.getContentPane().widgets();

         // determine widget numbers to remove and protect from NullPointerException
         int wSize = (w != null) ? w.length : 0;

         // the down body is destroyed automatically as the widgets are added to the content pane
         
         for (int i = 0; i < wSize; i++)
         {
            WidgetId id = UiUtils.getWidgetId(w[i]);
            if (id != null)
            {
               Widget<?> comp = widgetList.remove(id);
               
               if (comp != null)
               {
                  comp.destroy();
               }
            }
         }

         // remove the frame itself
         widgetList.remove(base);
         frame.destroy();
      }
   }

   /**
    * Return <code>Component</code> instance for the given ID.
    * 
    * @param   id
    *          Component ID.
    * @return  Reference to the <code>Component</code> or <code>null</code>
    *          if no such component exists.
    */
   @SuppressWarnings("unchecked")
   public Widget<O> getComponent(WidgetId id)
   {
      if (id == null)
      {
         return null;
      }

      if (id.equals(WidgetId.DEFAULT_WINDOW_ID))
         return (Widget<O>) Window.getDefaultWindow();
      
      Widget<O> w = widgetList.get(id);
      if (w == null)
      {
         w = WindowManager.findWindow(id.getId());
      }

      return w;
   }

   /**
    * Return <code>Component</code> instance for the given ID.
    * <p>
    * If the widget is part of a frame and the frame's down body is initialized, then return the 
    * widget associated with the current frame line.
    * 
    * @param   id
    *          Component ID.
    * @return  Reference to the <code>Component</code> or <code>null</code>
    *          if no such component exists.
    */
   @SuppressWarnings("unchecked")
   public Widget<O> getComponent(int id)
   {
      if (id == WidgetId.DEFAULT_WINDOW_ID.getId())
         return (Widget<O>) Window.getDefaultWindow();
      
      Widget<O> w = WindowManager.findWindow(id);
      if (w != null)
      {
         return w;
      }

      int frameID = lookupFrameIdFromWidgetId(id);
      
      if (frameID != id && frameID != -1)
      {
         Frame<?> frame = (Frame<?>) getComponent(frameID);
         
         int downIdx = frame == null || !frame.isDown() || frame.isConfigUpdateActive()
                        ? 1 
                        : ThinClient.getInstance().getFrameLine(frameID, 0);
         
         downIdx = downIdx - 1;
         
         WidgetId wid = (downIdx == 0 ? new WidgetId(id) : new WidgetDownId(id, downIdx));
         Widget<O> widget = widgetList.get(wid);

         // search by ID only if not found... 
         // TODO: confirm this doesn't break anything else!
         return (widget == null ? widgetList.get(new WidgetId(id)) : widget);
      }
      else
      {
         return widgetList.get(new WidgetId(id));
      }
   }

   /**
    * Retrieve the ID of the frame containing the widget with a given ID.
    *
    * @param    widgetId
    *           The widget ID.
    *
    * @return   The frame ID if widget is found or -1 if not.
    */
   public int lookupFrameIdFromWidgetId(int widgetId)
   {
      int resultId = -1;
      
      Widget<?> wid = widgetList.get(new WidgetId(widgetId));
      
      if (wid != null)
      {
         Frame<?> widFrame = (Frame<?>) UiUtils.locateFrame(wid);
         
         if (widFrame != null)
         {
            resultId = UiUtils.lookupWidgetIdAsInt(widFrame);
         }
      }
      
      return resultId;
   }

   /**
    * Instantiate screen from the given array of definitions.
    * 
    * @param   sd
    *          Array of data containers.
    */
   public void pushDefinition(ScreenDefinition[] sd)
   {
      for (int i = 0; i < sd.length; i++)
      {
         ScreenDefinition screenDefinition = sd[i];
   
         int frameId = screenDefinition.getRootFrameID();
         Frame<O> frame = (Frame<O>) getComponent(frameId);
         
         // if frame already exists, notify it that processing begins
         if (frame != null)
         {
            frame.preprocess();
         }
         
         Iterator<WidgetConfig> itr = screenDefinition.getConfigsIterator();
         
         while (itr.hasNext())
         {
            WidgetConfig cfg = itr.next();
            
            reconstructWidget(cfg);
         }
         
         if (frame == null)
         {
            frame = (Frame<O>) getComponent(screenDefinition.getRootFrameID());
         }
         
         if (frame != null)
         {
            addWidgetsToFrame(screenDefinition.getRootFrameID(), screenDefinition);
            frame.postprocess(screenDefinition.isSuppressRedraw());
         }
      }
   }

   /**
    * Add widgets to the frame which is identified by the given frame ID.
    * 
    * @param    frameID
    *           The index of the frame to be populated.
    * @param    screen
    *           List of widget configurations.
    */
   private void addWidgetsToFrame(int frameID, ScreenDefinition screen)
   {
      ArrayList<Integer> widgetIndexes = screen.getWidgetsInFrame(frameID);
      Frame<O> frame = (Frame<O>) getComponent(frameID);
   
      for (int i = 0; i < widgetIndexes.size(); i++)
      {
         Integer wid = widgetIndexes.get(i);
         int widgetIndex = wid.intValue();
         
         WidgetConfig config = screen.getConfig(widgetIndex);
         Widget<O> nextWidget = getComponent(widgetIndex);
         
         // associate browse columns with their browse widget which is a
         // container (of sorts) itself
         if (config instanceof BrowseColumnConfig)
         {
            Browse<O> browse = (Browse<O>) getComponent(((BrowseColumnConfig) config).browseId);
            
            if (browse != null)
               ((BrowseColumn<O>) nextWidget).setBrowse(browse);
   
            continue;
         }
         
         // check if this widget has already had the frame assigned as the
         // parent, if so then we MUST NOT add it to the widget list again
         // since it is already there (the Container._addComponent() method
         // sets the parent reference after unconditionally adding the 
         // component to the components vector
         if (nextWidget.parent() != null)
            continue;
   
         // make widget disabled by default
         
         if (nextWidget.isEnabled())
            nextWidget.setEnabled(false);
   
         frame.addWidget(nextWidget);
      }
   }

   /**
    * Get IDs of all widgets in the current screen belonging to the given frame.
    * 
    * @param   frameId
    *          Frame ID whose widget ID info should be returned.
    * 
    * @return  Array of all IDs.
    */
   public WidgetId[] getIDs(int frameId)
   {
      // check if the given frame is in registry
      Frame<O> frame = (Frame<O>) getComponent(frameId);
      
      if (frame == null)
      {
         return null;
      }
      
      // get the list of widgets from frame
      Widget<O>[] list = frame.getContentPane().widgets();

      // determine widget numbers to remove and protect from NullPointerException
      int wSize = (list != null) ? list.length : 0;
      
      if (wSize == 0)
      {
         return null;
      }
      
      // enumerate all widgets and get ID
      WidgetId[] res = new WidgetId[wSize];
      
      for (int i = 0; i < wSize; i++)
      {
         res[i] = UiUtils.getWidgetId(list[i]);
      }
      
      return res;
   }

   /**
    * Add the specified widget to the registry.
    * 
    * @param    widget
    *           The new widget which needs to be registered.
    */
   public void addWidget(Widget<O> widget)
   {
      this.widgetList.put(widget.config().id, widget);
   }

   /**
    * Add the widget with the specified ID to the registry.
    * 
    * @param    id
    *           The widget ID.
    * @param    widget
    *           The new widget which needs to be registered.
    */
   public void addWidget(WidgetId id, Widget<O> widget)
   {
      this.widgetList.put(id, widget);
   }
   
   /**
    * Remove the specified widget from the registry, immediately.
    * 
    * @param    id
    *           The widget ID.
    */
   public void removeWidget(WidgetId id)
   {
      this.widgetList.remove(id);
   }
}
