/*
** Module   : WindowGuiImpl.java
** Abstract : implements window widget drawing on GUI interfaces
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support; refactoring related to MESSAGE-AREA widget.
** 003 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Reworked the window parts to make use of widgets and generic drawing.
** 004 MAG 20140902 Use GUI driver system color API to get default colors for widgets drawing.
** 005 MAG 20140912 Set default FGCOLOR and BGCOLOR.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 008 MAG 20140923 Added GUI message area.
** 009 MAG 20141009 Insert indicator widget is not used in GUI.
** 010 CA  20141021 Finished support for the window title bar, window maximize/restore/minimize
**                  and close.  Added support for legacy 4GL window and mouse events.  
**                  Improved window resize.  TODO: support for the Size and Move options.
** 011 MAG 20141020 Added GUI Frame support.
** 012 HC  20141021 Implemented window move/position support.
** 013 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 014 CA  20141123 Fixed ENTRY/LEAVE when window focus get changed.  Fixed default placement of  
**                  dynamic windows.  Fixed window stacking at OS level.
** 015 CA  20141125 Fixed widget deregistration when the window is terminated.  Also, the window 
**                  needs to be placed only after the layout has been calculated (and done only 
**                  once).
** 016 EVL 20141126 Content pane GUI fix.
** 017 CA  20141206 In GUI Window, INSERT mode is always ON (and can't be changed).
** 018 HC  20141207 Implemented runtime support for widget coordinate attributes. Improved window
**                  mouse drag and mouse resize behavior. Implemented window synchronization with
**                  the server coordinate attributes.
*/

package com.goldencode.p2j.ui.client.gui;

import java.awt.event.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.WindowEvent;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * This class implement WINDOW widget drawing on GUI interfaces.
 */
@SuppressWarnings("unchecked")
public class WindowGuiImpl
extends Window<GuiOutputManager>
{
   /** Minimal window width. */
   private static final int MIN_WIDTH_PIXELS = 104;
   
   /** Minimal window height. */
   private static final int MIN_HEIGHT_PIXELS = 100;
   
   /** The title-bar widget for this window. */
   WindowTitleBar titleBar;

   /** The window's workspace, where all frame drawing is done. */
   WindowWorkSpace workspace;
   
   /** Screen driver */
   GuiDriver gd;

   /** Border size in pixels */
   double border = 0;

   /** Flag indicating that window resizing is in progress. */
   boolean resizing;

   /** The physical window width, after resize. */
   int resizedWidth;
   
   /** The physical window height, after resize. */
   int resizedHeight;

   /** The last mouse pressed determined for this window. */
   private Widget<GuiOutputManager> mousePressedSource = null;

   /** Flag indicating that the window is currently being drawn. */
   private boolean isDrawing = false;
   
   /** Flag indicating if the window is currently maximized. */
   private boolean maximized = false;
   
   /** The dimension before the window maximization. */
   private Dimension restoreDimension = null;

   /** The restore menu option.*/
   private PopupOption restoreOption;
   
   /**
    * Default constructor.
    * 
    * @param    id
    *           The widget's ID.
    */
   public WindowGuiImpl(WidgetId id)
   {
      super(id);
      
      // default window colors
      this.setDefaultColors();

      gd = (GuiDriver) OutputManager.getDriver();
      
      this.setLayout(new WindowLayout());

      registerWidgetStateListener(new WidgetStateListener()
      {
         @Override
         public void notifyWidgetDestroyed(Widget<? extends OutputManager<?>> w)
         {
            FontManager.deregisterFontTable(WindowGuiImpl.this, OutputManager.getDriver());
            gd.deregisterWindow(w.config().id.getId());
         }
      });
   }

   /**
    * Construct a window widget based on a specific configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   config
    *          Window configuration.
    */
   public WindowGuiImpl(WidgetId id, WidgetConfig config)
   {
      super(id, (WindowConfig) config);
      
      // default window colors
      this.setDefaultColors();

      gd = (GuiDriver) OutputManager.getDriver();

      this.setLayout(new WindowLayout());      

      registerWidgetStateListener(new WidgetStateListener()
      {
         @Override
         public void notifyWidgetDestroyed(Widget<? extends OutputManager<?>> w)
         {
            FontManager.deregisterFontTable(WindowGuiImpl.this, OutputManager.getDriver());
            gd.deregisterWindow(w.config().id.getId());
         }
      });
   }
   
   /**
    * Hide this window.
    */
   @Override
   public void hide()
   {
      if (!isVisible())
      {
         return;
      }
      
      WindowGuiImpl oldTop = (WindowGuiImpl) WindowManager.topWindow();

      super.hide();
      
      if (oldTop == this)
      {
         // ENTRY/LEAVE events are not sent; the new top is always the current DEFAULT-WINDOW
         WindowManager.moveToTop(Window.getDefaultWindow());
         
         // if the OS window is made visible (from hidden), then it is automatically moved to top... 
         // need to preserve the window order!
         gd.stackWindows(WindowManager.windowIds());

         final WindowGuiImpl top = (WindowGuiImpl) WindowManager.topWindow();
         ThinClient.getInstance().independentEventDrawingBracket(top, new Runnable()
         {
            @Override
            public void run()
            {
               top.repaint();
            }
         });
      }
   }
   
   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (isVisible())
         return;

      super.show();
      
      // if the OS window is made visible (from hidden), then it is automatically moved to top... 
      // need to preserve the window order!
      gd.stackWindows(WindowManager.windowIds());
   }
   
   /**
    * Get widgets array from first available window.
    * 
    * @return  array of the widgets belonging to the resolved windows.
    */
   @Override
   public Widget<GuiOutputManager>[] getWidgets()
   {
      return workspace.getContentPane().widgets();
   }
   
   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      boolean placeWindow = !(isRunning || gd == null);
      init();

      if (!isRunning)
      {
         return;
      }
      
      super.doLayout();

      if (placeWindow)
      {
         // set the window position
         WindowManager.placeWindow(this);
      }
   }
   
   /**
    * Draw window
    */
   @Override
   public void draw()
   {
      // Select renderer
      gd.selectWindow(config().id.getId());
      
      init();
      
      try
      {
         this.isDrawing = true;

         gd.setDrawingArea(this.clipRectangle());
         gd.startBatch();
         
         if (gd.isWindowVisible() != isVisible())
         {
            // if visibility has changed, fix the simulated window
            gd.setWindowVisible(isVisible());
         }
         
         super.draw();
      }
      finally
      {
         gd.endBatch();
         gd.setDrawingArea(null);
         this.isDrawing = false;
      }
   }

   /**
    * Check if the window can currently handle mouse events.
    * 
    * @return   <code>true</code> only and only if the window is not currently being 
    *           {@link #isDrawing drawn} or {@link #resizing resized}.
    */
   @Override
   public boolean handleMouseEvents()
   {
      return !isDrawing && !resizing;
   }

   /**
    * Set container size.
    *
    * @param   width
    *          New width.
    * @param   height
    *          New height.
    */
   @Override
   public void setSize(double width, double height)
   {
      super.setSize(width, height);
      this.contentPane.setSize(width - border, height);
      
      // this size is the full-width-char
      config().fullWidthChar = width;
      config().fullHeightChar = height;
   }
   
   /**                                     
    * Get dimension of screen area available for the placement of frames.
    * 
    * @return  Instance of <code>Dimension</code> which describes size of the
    *          available screen area.
    */
   @Override
   public Dimension getScreenDimension()
   {
      Dimension sz = dimension();
      return new Dimension(sz.width, sz.height - RESERVED_LINES);
   }

   /**
    * Resolve the font associated with the window's title.
    * 
    * @return   See above.
    */
   @Override
   public int resolveTitleFont()
   {
      int font = config().titleFont;
      
      // TODO: does it check the window's font first?
      return font == FontManager.UNDEFINED_FONT ? FontManager.WINDOW_TITLE_FONT : font;
   }
      
   /**
    * Get the message area widget associated with this window.
    * 
    * @return   See above.
    */
   public MessageAreaGuiImpl getGuiMessageArea()
   {
      return (MessageAreaGuiImpl) messageArea;
   }
   
   /**
    * Get messages area rectangle by message line index.
    *
    * @param   line
    *          message line specification: either 0-based line number for
    *          specific line or -1 for the whole message area.
    * @return  Instance of <code>Rectangle</code> which describes the message
    *          area in screen coordinates.
    */
   public com.goldencode.p2j.ui.client.Rectangle getMessageLineRectangle(int line)
   {
      return getGuiMessageArea().bounds();
   }

   /**
    * Get status area rectangle.
    *
    * @return  Instance of <code>Rectangle</code> which describes the status
    *          area in screen coordinates.
    */
   public com.goldencode.p2j.ui.client.Rectangle getStatusLineRectangle()
   {
      return statusLine.bounds();
   }
   
   /**
    * Adds the frame to the window at the specified position.
    * 
    * @param    widget
    *           The handle of the frame to be added to the window.
    * @param    row
    *           The row number for the upper-left frame corner.
    * @param    column
    *           The column number for the upper-left frame corner.
    */
   public void addFrame(Widget<GuiOutputManager> widget, double row, double column)
   {
      row = Coordinate.scale(row);
      column = Coordinate.scale(column);
      
      if (widget == null && row == 0 && column == 0)
         return;
      
      workspace.add(widget);
   }
   
   /**
    * Detach this widget from this container.  The widget will not be destroyed.
    * 
    * @param   widget
    *          Widget to detach.
    */
   @Override
   public void detach(Widget<GuiOutputManager> widget)
   {
      workspace.detach(widget);
   }
   
   /**
    * Find the widget to which the given mouse event can be applied.
    * 
    * @param    evt
    *           The mouse event.
    * 
    * @return   A widget over which the mouse is positioned or {@code null} if none found.
    */
   @Override
   public Widget<GuiOutputManager> findMouseSource(MouseEvt evt)
   {
      Widget<GuiOutputManager> widget = null;
      
      int evtID = evt.getEvent().getID();
      
      if (evtID == MouseEvent.MOUSE_DRAGGED || evtID == MouseEvent.MOUSE_RELEASED)
      {
         return mousePressedSource;
      }
      
      try
      {
         java.awt.Point p = evt.getEvent().getPoint();
         
         // search the other widgets
         widget = contentPane.findMouseSource(new NativePoint(p.x, p.y));
         
         widget = (widget == null ? this : widget);
      }
      finally
      {
         if (widget != null && evtID == MouseEvent.MOUSE_PRESSED)
         {
            mousePressedSource = widget;
         }
         
         if (evtID == MouseEvent.MOUSE_RELEASED)
         {
            mousePressedSource = null;
         }
      }
      
      return widget;
   }
   
   /**
    * Iconify the window.
    */
   public void iconify()
   {
      gd.selectWindow(config().id.getId());

      // the window-minimize event is raised by the OS listener

      gd.iconifyWindow();
   }
   
   /**
    * Restore the window by raising the <code>WINDOW-RESTORE</code> event.
    */
   public void restore()
   {
      if (!maximized)
      {
         return;
      }
      
      gd.selectWindow(config().id.getId());

      maximized = false;
      this.restoreOption.setEnabled(false);

      // restore to the previous width/height
      CoordinatesConversion c = screen().coordinates();
      int nwidth = c.widthToNative(restoreDimension.width);
      int nheight = c.heightToNative(restoreDimension.height);
      
      NativeDimension dim = this.gd.getWindowDimension();
      int dx = dim.width - nwidth;
      int dy = dim.height - nheight;

      resize(dx, dy, false, false, false);
      
      // raise the window-restore event
      ThinClient.getInstance().eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_RESTORED);
            EventManager.postEvent(evt);
         }
      });
   }
   
   /**
    * Close this window.  In 4GL, the <code>WINDOW-CLOSE</code> event is raised. If there is a 
    * trigger associated with the event, the RETURN NO-APPLY can not prevent closing of this window.
    */
   public void close()
   {
      // RETURN NO-APPLY is ignored
      ThinClient.getInstance().eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_CLOSE);
            EventManager.postEvent(evt);
         }
      });
   }
   
   /**
    * Maximize this window.
    */
   public void maximize()
   {
      gd.selectWindow(config().id.getId());

      if (maximized)
      {
         restore();

         return;
      }
      
      Dimension d = this.dimension();
      this.restoreDimension = d;
      
      double width = config().maxWidthChars;
      double height = config().maxHeightChars;
      
      CoordinatesConversion c = screen().coordinates();
      
      // raise the window-maximize event first; RETURN NO-APPLY is ignored
      ThinClient.getInstance().eventBracket(new Runnable()
      {
         @Override
         public void run()
         {
            Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_MAXIMIZED);
            EventManager.postEvent(evt);
         }
      });
      
      // resize, but only if the dimension differs
      if (Double.compare(width, d.width) != 0 || Double.compare(height, d.height) != 0)
      {
         int nwidth = c.widthToNative(width);
         int nheight = c.heightToNative(height);
         
         NativeDimension dim = this.gd.getWindowDimension();
         int dx = dim.width - nwidth;
         int dy = dim.height - nheight;
               
         resize(dx, dy, false, false, false);
      }
      
      this.restoreOption.setEnabled(true);
      this.maximized = true;

      repaint();
   }

   /**
    * Move this widget on the X and Y axis, by the specified deltas.  Values are specified in
    * units depending on the UI type.
    * 
    * @param   dx
    *          The X-axis delta.
    * @param   dy
    *          The Y-axis delta.
    */
   @Override
   public void moveBy(int dx, int dy)
   {
      gd.selectWindow(config().id.getId());

      NativePoint topLeft = gd.getWindowLocation();
      topLeft = topLeft.translate(dx, dy);
      setPhysicalLocation(topLeft.x, topLeft.y);
      
      gd.setWindowLocation(topLeft.x, topLeft.y);
   }
   
   /**
    * Resize the widget to the specified dimension.  Values are specified in units depending on
    * the UI type.
    * 
    * @param    dx
    *           The X axis adjustment.
    * @param    dy
    *           The Y axis adjustment.
    * @param    moveX
    *           Flag indicating that the top-left corner needs to be moved on the X axis.
    * @param    moveY
    *           Flag indicating that the top-left corner needs to be moved on the Y axis.
    */
   @Override
   public void resize(int dx, int dy, boolean moveX, boolean moveY)
   {
      resize(dx, dy, moveX, moveY, true);
   }
   
   /**
    * Resize the widget to the specified dimension.  Values are specified in units depending on
    * the UI type.
    * 
    * @param    dx
    *           The X axis adjustment.
    * @param    dy
    *           The Y axis adjustment.
    * @param    moveX
    *           Flag indicating that the top-left corner needs to be moved on the X axis.
    * @param    moveY
    *           Flag indicating that the top-left corner needs to be moved on the Y axis.
    * @param    raise
    *           Flag indicating if the 4GL WINDOW-RESIZED event needs to be raised.
    */
   public void resize(int dx, int dy, boolean moveX, boolean moveY, boolean raise)
   {
      if (maximized)
      {
         return;
      }
      
      this.gd.selectWindow(this.config().id.getId());

      // we are resizing the workspace; everything else is relative to it
      NativeDimension dim = this.gd.getWindowDimension();

      this.resizing = true;
      this.resizedWidth = Math.max(dim.width - dx, MIN_WIDTH_PIXELS);
      this.resizedHeight = Math.max(dim.height - dy, MIN_HEIGHT_PIXELS);

      // before the resize
      NativeRectangle nb1 = this.physicalBounds();

      try
      {
         doLayout();

         // after the resize
         NativeRectangle nb2 = this.physicalBounds();
         
         // sync the right point (if resizing horizontally)
         dx = dx != 0 ? (moveX ? 1 : -1) * (nb1.right() - nb2.right()) : 0;
         // sync the bottom point (if resizing vertically)
         dy = dy != 0 ? (moveY ? 1 : -1) * (nb1.bottom() - nb2.bottom()) : 0;
         
         dx = (moveX ? dx : 0);
         dy = (moveY ? dy : 0);
         if (dx != 0 || dy != 0)
         {
            this.moveBy(dx, dy);
         }
         
         gd.resizeWindow(resizedWidth, resizedHeight);
      }
      finally
      {
         this.resizing = false;
      }

      repaint();
      
      if (raise)
      {
         // raise the window-resized event; RETURN NO-APPLY is ignored.
         ThinClient.getInstance().eventBracket(new Runnable()
         {
            @Override
            public void run()
            {
               Event evt = new WindowEvent(WindowGuiImpl.this, Keyboard.SE_WINDOW_RESIZED);
               EventManager.postEvent(evt);
            }
         });
      }
   }
      
   /**
    * Shows the new message text.
    * <p>
    * Colors are ignored in Windows (GUI)
    * 
    * @param   message
    *          Array of two text message which need to be displayed inside the
    *          message area.
    * @param   cs
    *          Array of two color specifications.
    */
   @Override
   public void setMessageText(String[] message, Color[] cs)
   {
      setMessageText(message, cs, false);
   }

   /**
    * Set message main worker routine.
    * <p>
    * Colors are ignored in Windows (GUI)
    * 
    * @param   msg
    *          Message text array.
    * @param   cs
    *          Message color array.
    * @param   force
    *          <code>true</code> means force redisplaying of messages even
    *          if message text is already <code>null</code>.
    */
   @Override
   public void setMessageText(String[] msg, Color[] cs, boolean force)
   {
      getGuiMessageArea().setMessageText(msg, cs, force);         
   }

   /**
    * Determine if the current window is on top of all windows, including OS windows.
    * 
    * @return   See above.
    */
   public boolean isTopWindow()
   {
      boolean top = WindowManager.topWindow().config().id.equals(config.id);
      
      return top && gd.isOnTop(config.id.getId());
   }

   /**
    * Destroy this window.  Notify all listeners about the widget being destroyed.
    */
   @Override
   public void destroy()
   {
      int wid = config().id.getId();
      
      gd.selectWindow(wid);
      
      for (Widget<GuiOutputManager> w : widgets())
      {
         if (w.config() != null)
         {
            gd.deregisterWidget(w.config().id.getId());
         }
      }
      
      gd.deregisterWidget(this.config.id.getId());
   
      super.destroy();
   }
   
   /**
    * Called by the framework to notify the widget that the associated config state 
    * may have changed.
    */
   @Override
   public void syncWithConfig()
   {
      super.syncWithConfig();

      if (isDisplayed())
      {
         NativePoint loc = physicalLocation();
         if (loc.x != config.x || loc.y != config.y)
         {
            moveBy(config.x - loc.x, config.y - loc.y);
         }

         NativeDimension dim = physicalDimension();
         if (dim.width != config.widthPixels || dim.height != config.heightPixels)
         {
            this.resize(dim.width - config.widthPixels, dim.height - config.heightPixels, false,
                  false, true);
         }
      }
   }
   
   /**
    * In GUI, the insert mode is always {@link Window#INSERT_MODE_INS}.
    * 
    * @return    Always {@link Window#INSERT_MODE_INS}.
    */
   protected int getInsMode()
   {
      return INSERT_MODE_INS;
   }

   /**
    * Set the contents of the status line. This method is private and
    * called at service level by another static public method within the
    * class.  The status line output is truncated to a maximum of 63
    * characters, only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param   status
    *          The new string to be displayed inside the status line area.
    * @param   noLengthLimit
    *          flag indicating the status message should not be trimmed
    */
   @Override
   protected void setStatusLineText(String status, boolean noLengthLimit)
   {
      getContentPane().moveToTop(statusLine);
      statusLine.setText(status);
      statusLine.repaint();
   }

   /**
    * Reset Window instance state.
    */
   @Override
   protected void reset()
   {
      super.reset();
      
      titleBar = new WindowTitleBar(WidgetId.nextID());
      // TODO: what is the layout mgr for the GUI window (the workspace)?
      workspace = new WindowWorkSpace(WidgetId.nextID(), ScrollBar.Bars.BOTH);
   }
         
   /**
    * Initialize this window.  This can be done only after the driver has selected it.
    */
   protected void init()
   {
      if (isRunning || gd == null)
      {
         return;
      }

      gd.selectWindow(config().id.getId());

      border = screen().coordinates().widthFromNative(GuiOptions.BORDER_SIZE * 2);
      
      // the physical location is always for the OS window
      Point loc = location();
      setPhysicalLocation((int) loc.x, (int) loc.y); // these are always pixels
      
      reset();
      
      isRunning = true;

      add(contentPane);
      
      // initialize the window's font-table
      FontManager.registerFontTable(this, OutputManager.getDriver());
      
      // title-bar
      contentPane.add(titleBar);
      
      // workspace
      contentPane.add(workspace);
      
      // message area with variable lines of the text
      messageArea.attach(this);
      contentPane.add(getGuiMessageArea());
      
      // status line with default text
      statusLine.attach(this);      
      contentPane.add(statusLine);
      
      // add the widgets to the registry
      registerWidgets(contentPane);
      
      // TODO: do this conditionally?
      gd.registerMoveableWidget(titleBar.config().id.getId());
      gd.registerResizeableWidget(this.config().id.getId());
      
      PopupMenu popup = createPopup();
      registerWidgets(popup);
      moveToBottom(popup);
      
      gd.registerWidgetPopup(((AbstractWidget<GuiOutputManager>) titleBar.icon).config().id.getId(), 
                             MouseEvent.BUTTON1,
                             popup.config().id.getId());
      gd.registerWidgetPopup(titleBar.config().id.getId(), 
                             MouseEvent.BUTTON3, 
                             popup.config().id.getId());
      gd.registerWidgetPopup(((AbstractWidget<GuiOutputManager>) titleBar.title).config().id.getId(),
                             MouseEvent.BUTTON3, 
                             popup.config().id.getId());
      
      gd.setWindowVisible(config.visible);
      
      if (Window.getDefaultWindow() == this)
      {
         // if this is the default window, set the deltas
         WindowManager.setPlacementDeltas(titleBar.icon.physicalBounds().right(),
                                          titleBar.physicalBounds().bottom());
      }
   }
      
   /**
    * The private function to set up the global insert mode value.
    * <p>
    * On GUI no INSERT indicator is used to display insert status. 
    * 
    * @param   insMode
    *          New integer value for insert mode, but is ignored. 
    * @param   force
    *          Whether we have to repaint the indicator even if the new mode
    *          is the same as the old one.
    */
   @Override
   protected void setInsMode(int insMode, boolean force)
   {  
      insertMode = INSERT_MODE_INS;
   }

   /**
    * Resolve the font associated with the status-area.
    * 
    * @return   See above.
    */
   int resolveStatusAreaFont()
   {
      int font = config().statusAreaFont;
      
      // TODO: does it check the window's font first?
      return font == FontManager.UNDEFINED_FONT ? FontManager.DEFAULT_FONT : font;
   }

   /**
    * Resolve the font associated with the message-area.
    * 
    * @return   See above.
    */
   int resolveMessageAreaFont()
   {
      int font = config().messageAreaFont;
      
      // TODO: does it check the window's font first?
      return font == FontManager.UNDEFINED_FONT ? FontManager.DEFAULT_FONT : font;
   }

   /**
    * Determine if this window is being resized.
    * 
    * @return   The state of the {@link #resizing} flag.
    */
   boolean isResizing()
   {
      return resizing;
   }
   
   /**
    * Check if the window is already maximized - it can't be resized.
    * 
    * @return   The state of the {@link #maximized} flag.
    */
   boolean isMaximized()
   {
      return maximized;
   }
   
   /**
    * Set default window BGCOLOR and FGCOLOR. Windows are always use system colors.
    */
   private void setDefaultColors()
   {
      WindowConfig config = config();
      if (config != null)
      {
         // set default BGCOLOR
         config.sysbgcolor = SysColor.COLOR_3DFACE;
         // set default FGCOLOR         
         config.sysfgcolor = SysColor.COLOR_WINDOWTEXT;
      }
   }
      
   /**
    * Create a popup with the window caption actions.
    * 
    * @return   A popup widget.
    */
   private PopupMenu createPopup()
   {
      PopupMenu menu = new PopupMenu(WidgetId.nextID());
      
      restoreOption = menu.addOption("Restore",  new Runnable()
      {
         @Override
         public void run()
         {
            WindowGuiImpl.this.restore();
         }
      });
      restoreOption.setEnabled(false);
      
      menu.addOption("Move",     new Runnable()
      {
         @Override
         public void run()
         {
            // TODO: implement this
            // WindowGuiImpl.this.move();
         }
      }).setEnabled(false);
      
      menu.addOption("Size",     new Runnable()
      {
         @Override
         public void run()
         {
            // TODO: implement this
            // WindowGuiImpl.this.size();
         }
      }).setEnabled(false);

      menu.addOption("Minimize", new Runnable()
      {
         @Override
         public void run()
         {
            WindowGuiImpl.this.iconify();
         }
      });
      
      menu.addOption("Maximize", new Runnable()
      {
         @Override
         public void run()
         {
            WindowGuiImpl.this.maximize();
         }
      });
      
      menu.addOption(null,       null);
      
      menu.addOption("Close",    new Runnable()
      {
         @Override
         public void run()
         {
            WindowGuiImpl.this.close();
         }
      });
      
      this.contentPane.add(menu);
      menu.setVisible(false);
      
      return menu;
   }

   /**
    * Fix the IDs of this widget and all its containing widgets, if the widget is a container,
    * recursively.
    * 
    * @param    widget
    *           The root widget.
    */
   private void registerWidgets(Widget<GuiOutputManager> widget)
   {
      WidgetConfig cfg = widget.config();
      if (cfg != null)
      {
         // not all OS widgets need a config

         screen().getRegistry().addWidget(widget);
      }

      if (widget instanceof AbstractContainer)
      {
         Widget<GuiOutputManager>[] widgets = 
            ((AbstractContainer<GuiOutputManager>) widget).widgets();

         for (Widget<GuiOutputManager> w : widgets)
         {
            registerWidgets(w);
         }
      }
   }              
}
