/*
** Module   : WindowLayout.java
** Abstract : Layout the elements of a window.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 MAG 20140912 Set a new BitmapScreen object when window is resized.
** 003 HC  20140911 GUI coordinate support - data types refactoring.
** 004 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 005 MAG 20140923 GUI message area layout.
** 006 CA  20141021 Fixes related to window layout during resize.
** 007 MAG 20141021 Clear window borders whenever is resized.
** 008 HC  20141021 Implemented window move/position support.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 CA  20141123 Fixed initial dimension for dynamic windows.
** 011 HC  20141207 Improved window layout algorithm to honor the requested window size.
*/

package com.goldencode.p2j.ui.client.gui;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.layout.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * Layout the elements of a legacy-style window. This includes the window title, the workspace,
 * message area and status area.
 */
public class WindowLayout
implements LayoutManager<GuiOutputManager>
{
   /**
    * Get the minimum size of the specified container.
    * 
    * @param    container
    *           The container.
    *          
    * @return   Always {@code null}.
    */
   @Override
   public Dimension minimumSize(Container<GuiOutputManager> container)
   {
      return null;
   }

   /**
    * Place components in specified container.
    * 
    * @param   container
    *          Container to process - must be a window.
    */
   @Override
   public void doLayout(Container<GuiOutputManager> container)
   {
      // TODO: only use pixel/native coordinates, syncing of the characters values
      //       must be handled by the framework
      
      WindowGuiImpl window = (WindowGuiImpl) container;
      
      CoordinatesConversion c = window.screen().coordinates();

      // select renderer
      window.gd.selectWindow(window.config().id.getId());

      // 0. prerequisites for computing full width/height:
      boolean resizing = window.isResizing();

      // - the message area height
      int mheight = c.heightToNative(window.getGuiMessageArea().height());
      
      // - the status area height
      int sheight = c.heightToNative(window.getStatusArea().height()); 

      // - window size
      int height;
      int width;
      double cheight;
      double cwidth;
      
      if (resizing)
      {
         height = window.resizedHeight;
         width = window.resizedWidth;
         cheight = c.heightFromNative(height);
         cwidth = c.widthFromNative(width);
      }
      else
      {
         cheight = window.config().heightChars;
         cwidth = window.config().widthChars;
         height = c.heightToNative(cheight);
         width = c.widthToNative(cwidth);
      }
      
      // - layout the titlebar now, is needed to compute the window height
      int tbwidth = width - GuiOptions.BORDER_SIZE * 2;
      double ctbwidth = c.widthFromNative(tbwidth);
      window.titleBar.setWidth(ctbwidth);
      window.titleBar.doLayout();
      int tbheight = c.heightToNative(window.titleBar.height());

      int wheight = height - GuiOptions.BORDER_SIZE * 2 - tbheight - mheight - sheight; 
      int wwidth = width - GuiOptions.BORDER_SIZE * 2;
      double cwheight = c.heightFromNative(wheight);
      double cwwidth = c.widthFromNative(wwidth); 
      
      window.workspace.setSize(cwwidth, cwheight);
      window.config().setWidthChars(cwidth);
      window.config().setHeightChars(cheight);
      window.config().widthPixels = width;
      window.config().heightPixels = height;
      
      if (window.isResizing())
      {
         // clear window border
         clearBorder(window.gd, new NativeDimension(width, height));
      }
      else
      {
         window.gd.resizeWindow(width, height);
      }

      window.setSize(cwidth, cheight);
      
      // update screen bitmap
      window.setBitmap(new ScreenBitmap(height, width));
      
      // all coordinates for the following widgets are relative to the OS window (the container)
      int x = GuiOptions.BORDER_SIZE;
      int y = GuiOptions.BORDER_SIZE;
      
      // 1. set the title into position:
      window.titleBar.setPhysicalLocation(x, y);
      window.titleBar.setLocation(c.widthFromNative(x), c.heightFromNative(y));
      
      // 2. set the window workarea (this is the legacy 4GL window, where frames are placed)
      // - bellow the title and above the message area
      // - its size does not include the borders
      // - its size needs to be computed first, as the general window size depends on it
      y = y + tbheight;
      window.workspace.setPhysicalLocation(x, y);
      window.workspace.setLocation(c.widthFromNative(x), c.heightFromNative(y)); 
      
      // 3. set the message area
      y = y + wheight;
      window.getGuiMessageArea().setPhysicalLocation(x, y);
      window.getGuiMessageArea().setLocation(c.widthFromNative(x), c.heightFromNative(y)); 
      window.getGuiMessageArea().doLayout();
      
      // 4. set the status area
      y = y + mheight;
      window.getStatusArea().setPhysicalLocation(x, y);
      window.getStatusArea().setLocation(c.widthFromNative(x), c.heightFromNative(y));
   }

   /**
    * Whenever window is resized the right and bottom borders became dirty.  In order to prevent
    * dirty borders right and bottom borders are cleaned.
    * 
    * @param   gd
    *          GUI device.
    * @param   d
    *          Window dimension.
    */
   private void clearBorder(GuiDriver gd, NativeDimension d)
   {
      // main window border size
      int border = GuiOptions.BORDER_SIZE;
      // clear left and bottom borders
      gd.setColor(gd.getSysColor(SysColor.COLOR_3DFACE));
      gd.fillRect(d.width - border, 0, border, d.height);
      gd.fillRect(0, d.height - border, d.width, border);      
   }
}
