/*
** Module   : GuiDriver.java
** Abstract : define GUI primitives.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support; some other new APIs were added.
** 003 CA  20140805 Added mouse event support and misc primitives for window actions/support.
** 004 MAG 20140902 Integrate COLOR-TABLE. Added support for BGCOLOR and FGCOLOR color conversion.
**                  Added system colors used solely for widget drawing.  
** 005 MAG 20140912 Implement GUI status bar. 
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Added new APIs to allow:
**                  - drawing of nested containers (translatedOrigin)
**                  - notification of any mouse event (listenAnyMouseEvent)
**                  - batch drawing (start/endBatch).
** 009 HC  20141031 Implemented support for drawing highlighted areas.
** 010 MAG 20141029 Added clip rectangle parameter to draw method.
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 EVL 20141116 Adding round rectangles draw and fill.  Adding api to get the UI theme name.
** 013 CA  20141123 Added APIs to allow enforcement of window stacking at OS level.
** 014 CA  20141125 Added API to conditionally enable/disable OS-level events.
** 015 EVL 20141125 Adding dotted line draw capability.
** 016 CA  20141205 Added API to draw relative to window's real origin (top-left corner), 
**                  regardless of the current translated origin.
**                  Added API to register editable widgets (thus the driver can adjust the mouse 
**                  cursor accordingly).
** 017 CA  20141220 Added APIs for clipboard access.  Added APIs to allow text drawing and font 
**                  selection matching legacy sizes.
** 018 HC  20141207 Implemented runtime support for several SESSION attributes.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.awt.*;
import java.awt.Image;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.driver.*;
import com.goldencode.p2j.ui.client.gui.driver.swing.*;

/**
 * GUI screen driver contract.
 */
public interface GuiDriver
extends ScreenDriver<GuiOutputManager>
{
   /**
    * Translates the origin of the <code>Graphics2D</code> context to the
    * point (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
    * Modifies the <code>Graphics2D</code> context so that its new origin
    * corresponds to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * <code>Graphics2D</code> context's former coordinate system.  
    * All coordinates used in subsequent rendering operations on this graphics
    * context are relative to this new origin.
    *
    * @param   x 
    *          the specified x coordinate
    * @param   y 
    *          the specified y coordinate
    */
   public void translate(int x, int y);
   
   /**
    * The real location where the origin was translated.
    * 
    * @return   See above.
    */
   public NativePoint translatedOrigin();
   
   /**
    * Draw a string at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   public void drawString(String text, int x, int y);

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   legacyWidth
    *          The legacy width, which needs to be matched when drawing.
    * @param   legacyHeight
    *          The legacy width, which needs to be matched when drawing.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y);

   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param   x1  
    *          the first point's <i>x</i> coordinate.
    * @param   y1  
    *          the first point's <i>y</i> coordinate.
    * @param   x2  
    *          the second point's <i>x</i> coordinate.
    * @param   y2  
    *          the second point's <i>y</i> coordinate.
    */
   public void drawLine(int x1, int y1, int x2, int y2);

   /**
    * Draw a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */
   public void drawRect(int x, int y, int width, int height);

   /**
    * Draw a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter);

   /**
    * Fill a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   public void fillRect(int x, int y, int width, int height);

   /**
    * Fill a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */   
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter);

   /**
    * Fill polygon.
    * 
    * @param   xPoints
    *          X coordinates.
    * @param   yPoints
    *          Y coordinates.
    */   
   public void fillPolygon(int[] xPoints, int[] yPoints);
   
   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>Color</code>.  This method uses
    * the current <code>Color</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be filled.
    * @param   y 
    *          the y coordinate of the rectangle to be filled.
    * @param   width
    *          the width of the rectangle to be filled.
    * @param   height 
    *          the height of the rectangle to be filled.
    * @param   raised 
    *          a boolean value that determines whether the rectangle appears 
    *          to be raised above the surface or etched into the surface.
    */
   public void fill3DRect(int x, int y, int width, int height, boolean raised);
   
   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>Color</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be drawn.
    * @param   y 
    *          the y coordinate of the rectangle to be drawn.
    * @param   width 
    *          the width of the rectangle to be drawn.
    * @param   height 
    *          the height of the rectangle to be drawn.
    * @param   raised 
    *          a boolean that determines whether the rectangle appears 
    *          to be raised above the surface or sunk into the surface.
    */
   public void draw3DRect(int x, int y, int width, int height, boolean raised);
   
   /**
    * Draw a image at the specified location.
    * 
    * @param   img
    *          Image to draw.
    * @param   x
    *          left location.
    * @param   y
    *          top location.
    * @param   w
    *          the image width.
    * @param   h
    *          the image height.
    */
   public void drawImage(Image img, int x, int y, int w, int h);

   /**
    * Draws a highlighted rectangular area of the specified color and bounds.
    * The color is alpha-blended with the content of the screen buffer.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          area width
    * @param   height
    *          area height
    * @param   color
    *          the highlight color
    */
   public void drawHighlight(int x, int y, int width, int height, Color color);

   /**
    * Clears the highlighted area drawn by the previous call of 
    * {@link #drawHighlight(int, int, int, int, Color)}.
    */
   public void clearHighlight();

   /**
    * Set current color used on draw operations.
    * 
    * @param   color
    *          Drawing color.
    */
   public void setColor(Color color);
   
   /**
    * Set current font used on draw operations.
    * 
    * @param   font
    *          Drawing font.
    */
   public void setGuiFont(Font font);
   
   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.
    * @param    legacyWidth
    *           The legacy width which needs to be matched.
    * @param    legacyHeight
    *           The legacy height which needs to be matched.
    *           
    * @return   The scaled font.
    */
   public Font scaleFont(Font font, int legacyWidth, int legacyHeight);
   
   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param   style
    *          Font style.
    */
   public void setFontStyle(int style);

   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param   stroke
    *          Line style.
    */
   public void setLineStroke(Stroke stroke);

   /**
    * Resize window.
    * 
    * @param   width
    *          new width.
    * @param   height
    *          new height
    */
   public void resizeWindow(int width, int height);
   
   /**
    * Register this window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    sim
    *           The simulator in charge of this window.
    */
   public void registerWindow(int windowId, GuiSimulator sim);
   
   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   public void deregisterWindow(int windowId);
   
   /**
    * On a GUI interface multiple windows could be created. Each window has it own renderer.
    * When a window is draw first the window must call this method to setup it renderer as 
    * the current graphical canvas.  
    * 
    * @param   windowId
    *          Window ID.
    */
   public void selectWindow(int windowId);   
   
   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top);
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   public NativePoint getWindowLocation();
   
   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    name
    *           The font face name.
    * @param    size
    *           The font's point size.
    * @param    bold
    *           Flag indicating if the font is bold.
    * @param    italic
    *           Flag indicating if the font is italic.
    * @param    underline
    *           Flag indicating if the font is underline.
    * @param    definition
    *           The font's definition.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * 
    * @return   A font instance.
    */
   public Font createFont(String  name,
                          int     size,
                          boolean bold,
                          boolean italic,
                          boolean underline,
                          byte[]  definition,
                          int     legacyWidth,
                          int     legacyHeight);
   
   /**
    * Get the width of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's width.
    */
   public int getFontWidth(Font font);
   
   /**
    * Get the height of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's height.
    */
   public int getFontHeight(Font font);
   
   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   public int getTextWidth(String text, Font font);

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   public int getTextHeight(String text, Font font);
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible);

   /**
    * Iconify the window.
    */
   public void iconifyWindow();

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type, one of the {@link Cursor} constants.
    */
   public void setCursor(int type);

   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   public NativeDimension getWindowDimension();

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    * 
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   public void draw(NativePoint origin, NativeRectangle clip, Runnable core);

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the window's real (0,0) origin (top-left corner) and then translating to the
    * specified location, before executing the drawing.
    * <p>
    * The translated origin (as before this was called) needs to be restored, before this call
    * ends.
    * 
    * @param    origin
    *           The drawing origin (relative to window's real origin).
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   public void drawFromOrigin(NativePoint origin, NativeRectangle clip, Runnable core);

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(NativeRectangle bounds);
   
   /**
    * Register the specified widget as a moveable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void registerMoveableWidget(int wid);

   /**
    * Register the specified widget as a resizeable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void registerResizeableWidget(int wid);
   
   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   public void deregisterWidget(int wid);

   /**
    * Register the specified widget for popup events, via mouse.
    * <p>
    * The popup with the given ID will be shown when the specified mouse button is pressed inside
    * the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   public void registerWidgetPopup(int wid, int btn, int pid);

   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * 
    * @param    wid
    *           The widget id.
    */
   public void registerEditableWidget(int wid);

   /**
    * Create a new GUI simulator.
    * 
    * @param    options
    *           The window options.
    *           
    * @return   See above.
    */
   public GuiSimulator newSimulator(GuiOptions options);
   
   /**
    * Get BGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index and convert the
    * value to a {@link java.awt.Color#Color(int)} instance.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of Color
    */
   public Color getBgColor(int index);

   /**
    * Get FGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index and convert the
    * value to a {@link java.awt.Color#Color(int)} instance.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of Color
    */
   public Color getFgColor(int index);
 
   /**
    * Get a system color.
    * 
    * @param   index
    *          Color index.
    *          
    * @return  see above.
    */
   public Color getSysColor(int index);
   
   /**
    * Get foreground color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   public Color getFgStdColor(String stdColor);

   /**
    * Get background color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   public Color getBgStdColor(String stdColor);
   
   /**
    * Force screen buffer draw.
    */
   public void flush();
   
   /**
    * Set status text message of current window.
    * 
    * @param   message
    *          Text message.
    */
   public void setStatusText(String message);

   /**
    * Set status text message of specified window.
    * 
    * @param   message
    *          Text message.
    * @param   window
    *          Window instance.
    */
   public void setStatusText(Window<?> window, String message);
   
   /**
    * Check if window has renderer.
    * 
    * @param   windowId
    *          Window ID.
    * 
    * @return  <code>true</code> if window has renderer <code>true</false> otherwise.
    */
   public boolean hasWindowRenderer(int windowId);

   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    id
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   public void listenAnyMouseEvent(int id, long when);
   
   /** 
    * Start a batch drawing routine.
    */
   public void startBatch();
   
   /**
    * End a batch drawing routine.
    */
   public void endBatch();

   /**
    * Check if the currently selected window is visible.
    * 
    * @return   See above.
    */
   public boolean isWindowVisible();

   /**
    * Get the window UI theme name.  Can be either xp or classic.
    * 
    * @return   The currently used theme name for OS UI interface elemens.
    */
   public String getUiTheme();

   /**
    * Force the GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   public void stackWindows(int[] windowIds);

   /**
    * Determine if the given window is on top of all windows, including OS.
    * 
    * @param    windowId
    *           The window ID to check.
    *           
    * @return   See above.
    */
   public boolean isOnTop(int windowId);

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    process
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   public void enableEvents(int[] windowIds, boolean process);
   
   /**
    * Returns display size.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The dimensions in pixels.
    */
   public NativeDimension getDisplaySize(int windowId);
   
   /**
    * Returns an application-usable display area. For example the task bar is
    * not counted as the usable area.
    * 
    * @param    windowId
    *           A valid window id.
    *           
    * @return   The dimensions in pixels.
    */
   public NativeRectangle getDisplayWorkArea(int windowId);

   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   public void copyToCliboard(String txt);

   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   See above.
    */
   public String clipboardContents();
}
