/*
** Module   : GuiPrimitives.java
** Abstract : GUI draw primitives.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Added mouse event support.  More misc changes related to window behavior.
** 004 MAG 20140905 Added fillPoly method.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Changed drawImage, to specify the image's width and height.
** 009 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 010 EVL 20141116 Adding round rectangles draw and fill.
** 011 EVL 20141125 Adding dotted line draw capability.
** 012 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 013 HC  20141207 Implemented runtime support for several SESSION attributes.
*/

package com.goldencode.p2j.ui.client.gui.driver;

import java.awt.*;
import java.awt.Image;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.swing.*;

/** 
 * This is a low level GUI-specific drawing interface that abstracts the widget-specific drawing
 * from the low-level rendering implementations.
 */
public interface GuiPrimitives
extends OutputPrimitives
{   
   /**
    * Perform shutdown.
    */
   public void shutdown();

   /**
    * Register this window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    sim
    *           The simulator in charge of this window.
    */
   public void registerWindow(int windowId, GuiSimulator sim);

   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   public void deregisterWindow(int windowId);

   /**
    * Translates the origin of the <code>Graphics2D</code> context to the
    * point (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
    * Modifies the <code>Graphics2D</code> context so that its new origin
    * corresponds to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * <code>Graphics2D</code> context's former coordinate system.  
    * All coordinates used in subsequent rendering operations on this graphics
    * context are relative to this new origin.
    *
    * @param   x 
    *          the specified x coordinate
    * @param   y 
    *          the specified y coordinate
    */
   public void translate(int x, int y);

   /**
    * Draw a string at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   public void drawString(String text, int x, int y);

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   legacyWidth
    *          The legacy width, which needs to be matched when drawing.
    * @param   legacyHeight
    *          The legacy width, which needs to be matched when drawing.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y);

   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param   x1  
    *          the first point's <i>x</i> coordinate.
    * @param   y1  
    *          the first point's <i>y</i> coordinate.
    * @param   x2  
    *          the second point's <i>x</i> coordinate.
    * @param   y2  
    *          the second point's <i>y</i> coordinate.
    */
   public void drawLine(int x1, int y1, int x2, int y2);

   /**
    * Draw a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */
   public void drawRect(int x, int y, int width, int height);
   
   /**
    * Draw a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter);

   /**
    * Fill a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   public void fillRect(int x, int y, int width, int height);
   
   /**
    * Fill a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */   
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter);

   /**
    * Fill polygon.
    * 
    * @param   xPoints
    *          X coordinates.
    * @param   yPoints
    *          Y coordinates.
    */   
   public void fillPolygon(int[] xPoints, int[] yPoints);

   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>Color</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be drawn.
    * @param   y 
    *          the y coordinate of the rectangle to be drawn.
    * @param   width 
    *          the width of the rectangle to be drawn.
    * @param   height 
    *          the height of the rectangle to be drawn.
    * @param   raised 
    *          a boolean that determines whether the rectangle appears 
    *          to be raised above the surface or sunk into the surface.
    */
   public void draw3DRect(int x, int y, int width, int height, boolean raised);

   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>Color</code>.  This method uses
    * the current <code>Color</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be filled.
    * @param   y 
    *          the y coordinate of the rectangle to be filled.
    * @param   width
    *          the width of the rectangle to be filled.
    * @param   height 
    *          the height of the rectangle to be filled.
    * @param   raised 
    *          a boolean value that determines whether the rectangle appears 
    *          to be raised above the surface or etched into the surface.
    */
   public void fill3DRect(int x, int y, int width, int height, boolean raised);
   
   /**
    * Draw a image at the specified location.
    * 
    * @param   img
    *          Image to draw.
    * @param   x
    *          left location.
    * @param   y
    *          top location.
    * @param   w
    *          the image width.
    * @param   h
    *          the image height.
    */
   public void drawImage(Image img, int x, int y, int w, int h);  

   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    name
    *           The font face name.
    * @param    size
    *           The font's point size.
    * @param    bold
    *           Flag indicating if the font is bold.
    * @param    italic
    *           Flag indicating if the font is italic.
    * @param    underline
    *           Flag indicating if the font is underline.
    * @param    definition
    *           The font's definition.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * 
    * @return   A font instance.
    */
   public Font createFont(String  name,
                                   int     size,
                                   boolean bold,
                                   boolean italic,
                                   boolean underline,
                                   byte[]  definition,
                                   int     legacyWidth,
                                   int     legacyHeight);

   /**
    * Set current color used on draw operations.
    * 
    * @param   color
    *          Drawing color.
    */
   public void setColor(Color color);

   /**
    * Set current font used on draw operations.
    * 
    * @param   font
    *          Drawing font.
    */
   public void setGuiFont(Font font);

   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.
    * @param    legacyWidth
    *           The legacy width which needs to be matched.
    * @param    legacyHeight
    *           The legacy height which needs to be matched.
    *           
    * @return   The scaled font.
    */
   public Font scaleFont(Font font, int legacyWidth, int legacyHeight);

   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param   style
    *          Font style.
    */
   public void setFontStyle(int style);
   
   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param   stroke
    *          Line style.
    */
   public void setLineStroke(Stroke stroke);

   /**
    * Resize window.
    * 
    * @param   width
    *          new width.
    * @param   height
    *          new height
    */
   public void resizeWindow(int width, int height);

   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top);
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   public NativePoint getWindowLocation();

   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   public NativeDimension getWindowDimension();

   /**
    * Get the width of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's width.
    */
   public int getFontWidth(Font font);

   /**
    * Get the height of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's height.
    */
   public int getFontHeight(Font font);

   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   public int getTextWidth(String text, Font font);

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   public int getTextHeight(String text, Font font);
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible);

   /**
    * Iconify the window.
    */
   public void iconifyWindow();

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    c
    *           The cursor type, one of the {@link Cursor} constants.
    */
   public void setCursor(Cursor c);

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(NativeRectangle bounds);
   
   /**
    * Force screen buffer draw.
    */
   public void flush();   
   
   /**
    * Check if window has renderer.
    * 
    * @param   windowId
    *          Window ID.
    * 
    * @return  <code>true</code> if window has renderer <code>true</false> otherwise.
    */
   public boolean hasWindowRenderer(int windowId);

   /**
    * Get the currently selected simulator.
    * 
    * @return   See above.
    */
   public GuiSimulator getSimulator();

   /**
    * Returns display size.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The dimensions in pixels.
    */
   public abstract NativeDimension getDisplaySize(int windowId);
   
   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in pixels.
    */
   public abstract NativeRectangle getDisplayWorkArea(int windowId);
}
