/*
** Module   : GuiSimulator.java
** Abstract : GUI renderer.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support - removed fontSize.
** 003 CA  20140806 Added mouse support and other misc APIs related to window state.
** 004 MAG 20140912 Added fill polygon.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140924 Create graphical object when buffered image is created.
** 007 CA  20141021 Added notification of OS window events.  Added support for batch drawing: the 
**                  current AWT repaint processing will wait until the last paint primitive is
**                  received, before returning the control to the AWT event processing thread, to
**                  avoid flickering.  Do not allow window maximize via OS means.
**                  Added implementation for new paint primitives: SET_LOCATION, RESIZE_WINDOW.
** 008 HC  20141031 Implemented support for drawing highlighted areas.
** 009 MAG 20141020 Improve clipping rectangle management. Added mouse wheel listener.
** 010 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
** 011 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 012 EVL 20141116 Adding round rectangles draw and fill.
** 013 CA  20141123 Fixed processing of window focus gained/lost.  Added APIs to allow movement of
**                  a window to top (at OS level).  When the window is destroyed, first remove the
**                  listeners, so no key/mouse/window events is processed while the window is being
**                  destroyed.
** 014 CA  20141125 Do not accept negative dimensions, when resizing the screen buffers.
**                  Added API to conditionally enable/disable OS-level events.
** 015 EVL 20141125 Adding dotted line draw capability.
** 016 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 017 HC  20141207 Implemented runtime support for several SESSION attributes.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.event.*;
import java.awt.font.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.List;
import java.util.concurrent.*;

import javax.swing.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;
import com.goldencode.p2j.ui.client.driver.swing.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.WindowEvent;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.swing.*;

/**
 * The GUI renderer. Because the paint should be permanent a BufferedImage is used to draw
 * and keep the last screen image. A pool of drawing operations is used to queue drawing 
 * primitives. The operations are enqueued from poll one by one and draw on the buffered image. 
 * When pool is empty the screen is updated with the image from the stored buffered image.  
 */
@SuppressWarnings("serial")
public class GuiSimulator 
extends AbstractSimulator
{
   /** Map of saved event listeners, which can be restored at a later time. */
   private Map<Class<? extends EventListener>, List<EventListener>> savedListeners = null;

   /** Screen buffer */
   private BufferedImage screenBuffer;
   
   /** Graphical object */
   private Graphics2D g2;

   /** Draw operations pools */
   private Deque<PaintStructure> graphicPool = new ConcurrentLinkedDeque<>();

   /** The window ID associated with this simulator. */
   private int windowId;
   
   /** A stack of clipping rectangles. */
   private Deque<java.awt.Rectangle> clippings = new ConcurrentLinkedDeque<>();

   /** The registered mouse handler. */
   private final MouseHandler mouseHandler;
   
   /** The bitmap associated with this simulator. */
   private ScreenBitmap bitmap;
   
   /** The X translated location of the drawing area, on the physical canvas. */
   private int translateX = 0;
   
   /** The Y translated location of the drawing area, on the physical canvas. */
   private int translateY = 0;
   
   /** Screen buffer for clearing an active highlight */
   private BufferedImage highlightRestoreBuffer;
   
   /** The left location of the last highlighted widget. */
   private int highlightX;
   
   /** The top location of the last highlighted widget. */
   private int highlightY;

   /** The window focus listeners for this simulator. */
   private WindowFocusListener[] savedFocusListeners;
   
   /**
    * Constructor.
    * 
    * @param   frame
    *          Application container.
    * @param   options
    *          GUI options.
    * @param   keyReader
    *          The {@link SwingKeyboardReader key reader} which listens for incoming keys.
    */
   public GuiSimulator(JFrame frame, GuiOptions options, SwingKeyboardReader keyReader)
   {
      super(frame, null, keyReader);

      frame.setUndecorated(true);

      // register listeners for OS window events
      final ThinClient tc = ThinClient.getInstance();
      frame.addWindowListener(new WindowAdapter()
      {
         @Override
         public void windowIconified(java.awt.event.WindowEvent e)
         {
            final WindowGuiImpl window = (WindowGuiImpl) tc.getWidget(windowId);

            tc.independentEventDrawingBracket(window, new Runnable()
            {
               @Override
               public void run()
               {
                  EventManager.postEvent(new WindowEvent(window, Keyboard.SE_WINDOW_MINIMIZED));
               }
            });
         }
         
         @Override
         public void windowDeiconified(java.awt.event.WindowEvent e)
         {
            final WindowGuiImpl window = (WindowGuiImpl) tc.getWidget(windowId);

            tc.independentEventDrawingBracket(window, new Runnable()
            {
               @Override
               public void run()
               {
                  EventManager.postEvent(new WindowEvent(window, Keyboard.SE_WINDOW_RESTORED));
               }
            });
         }
      });
      
      frame.addWindowFocusListener(new WindowFocusListener()
      {
         @Override
         public void windowLostFocus(java.awt.event.WindowEvent e)
         {
            // a LEAVE event is not raised; it will be raised by the 4GL window which received
            // the focus.  If focus is received by another OS window, then no LEAVE needs to be
            // sent.
            
            WindowGuiImpl window = (WindowGuiImpl) WindowManager.findWindow(windowId);

            boolean toOS = (e.getOppositeWindow() == null);
            if (toOS)
            {
               // if focus is lost to a non-4GL window, the titlebar needs to be aware of this
               ThinClient.getInstance().postOSEvent(new WindowActivated(window, false, toOS));
            }
         }
         
         @Override
         public void windowGainedFocus(java.awt.event.WindowEvent e)
         {
            WindowGuiImpl window = (WindowGuiImpl) WindowManager.findWindow(windowId);

            boolean fromOS = (e.getOppositeWindow() == null);
            ThinClient.getInstance().postOSEvent(new WindowActivated(window, true, fromOS));
         }
      });

      frame.addWindowStateListener(new WindowStateListener()
      {
         public void windowStateChanged(java.awt.event.WindowEvent e)
         {
            if ((e.getNewState() & Frame.MAXIMIZED_BOTH) == Frame.MAXIMIZED_BOTH)
            {
               resizeWindow(getWidth(), getHeight());
            }
         }
      });
   
      // setup mouse listeners
      mouseHandler = new MouseHandler(this);
      addMouseListener(mouseHandler);
      addMouseMotionListener(mouseHandler);
      addMouseWheelListener(mouseHandler);
      
      // TODO: disable the popup of the window actions (min/max/close/etc) via key? 
   }

   /**
    * Cause the outer frame to be disposed, as the window is being terminated.
    */
   public void quit()
   {
      captureMouseEvents(false);

      for (MouseListener l : getMouseListeners())
      {
         removeMouseListener(l);
      }
      for (MouseMotionListener l : getMouseMotionListeners())
      {
         removeMouseMotionListener(l);
      }
      for (MouseWheelListener l : getMouseWheelListeners())
      {
         removeMouseWheelListener(l);
      }
      for (WindowListener l : frame.getWindowListeners())
      {
         frame.removeWindowListener(l);
      }
      for (WindowFocusListener l : frame.getWindowFocusListeners())
      {
         frame.removeWindowFocusListener(l);
      }
      for (WindowStateListener l : frame.getWindowStateListeners())
      {
         frame.removeWindowStateListener(l);
      }
      for (KeyListener l : frame.getKeyListeners())
      {
         frame.removeKeyListener(l);
      }
      
      super.quit();
   }

   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   public boolean captureMouseEvents(boolean state)
   {
      return mouseHandler.captureMouseEvents(state);
   }
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible)
   {
      frame.pack();
      frame.setVisible(visible);
   }
   
   /**
    * Check if the simulated window is visible.
    * 
    * @return   See above.
    */
   public boolean isWindowVisible()
   {
      return frame.isVisible();
   }

   /**
    * Set the window ID for this simulator.
    * 
    * @param   windowId
    *          The window ID.
    */
   public void setWindowId(int windowId)
   {
      this.windowId = windowId;
   }

   /**
    * Get the window ID for this simulator.
    */
   public int getWindowId()
   {
      return windowId;
   }

   /** 
    * Get the native screen size, in pixel units.
    * 
    * @return   See above.
    */
   public NativeDimension getScreenSize()
   {
      Dimension dim = Toolkit.getDefaultToolkit().getScreenSize();
      NativeDimension ndim = new NativeDimension(dim.width, dim.height);
      
      return ndim;
   }

   /**
    * Paint the entire component to the given graphics instance, using
    * previously set state.  Depending on the repaint request, the graphics
    * object's clipping region may have been set to a subset of the full
    * rectangle. This method makes no such distinction, it just blindly
    * draws everything and the stuff outside of the clipping region never
    * gets sent to the hardware.  This is the core where all drawing occurs
    * and it draws the screen text (including switching out the fonts per
    * character as needed to get the right colors/attributes), the current
    * cursor (if it is visible) and the any selection box (when a selection
    * is active).
    *
    * @param    graphics
    *           The object into which to draw.
    */
   @Override
   public void paintComponent(Graphics graphics)
   {
      super.paintComponent(graphics);

      // draw
      drawOnScreenBitmap(graphics);
   }  

   /**
    * Add a new <code>PaintStructure</code> to the drawing pool.
    * The item is enqueued at the end of the pool.
    * 
    * @param   ps
    *          A new PaintStructure instance.          
    */      
   public void offer(PaintStructure ps)
   {
      // to prevent deadlocks, synchronize on the frame's layout lock
      synchronized (getTreeLock())
      {
         graphicPool.offerLast(ps);

         getTreeLock().notifyAll();
      }
   }

   /**
    * Resize window.
    * 
    * @param   width
    *          new width.
    * @param   height
    *          new height
    */
   public void resizeWindow(int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.RESIZE_WINDOW);
      ps.width = width;
      ps.height = height;
      
      offer(ps);
      repaint();

      this.bitmap.setRows(height);
      this.bitmap.setColumns(width);
   }

   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LOCATION);
      ps.x = left;
      ps.y = top;
      
      offer(ps);
      repaint();
   }
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   public NativePoint getWindowLocation()
   {
      java.awt.Point p = frame.getLocation();
      
      return new NativePoint(p.x, p.y);
   }

   /**
    * Iconify the window.
    */
   public void iconifyWindow()
   {
      frame.setState(Frame.ICONIFIED);
   }

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   public void setDrawingArea(NativeRectangle bounds)
   {
      PaintStructure ps = new PaintStructure(bounds == null ? PaintPrimitives.NO_CLIP  
                                                            : PaintPrimitives.CLIP);
      if (bounds != null)
      {
         ps.x = bounds.left();
         ps.y = bounds.top();
         ps.width = bounds.right() - bounds.left();
         ps.height = bounds.bottom() - bounds.top();
      }
      
      offer(ps);
   }

   /**
    * Get the screen width, in character units.
    * 
    * @return    See above.
    */
   public int screenWidth()
   {
      return bitmap.cols();
   }

   /**
    * Get the screen height, in character units.
    * 
    * @return    See above.
    */
   public int screenHeight()
   {
      return bitmap.rows();
   }

   /**
    * Set the active screen bitmap for the current window.
    * 
    * @param    bitmap
    *           The screen bitmap.
    */
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      ScreenBitmap prev = bitmap;

      this.bitmap = bitmap;
      
      if (prev == null || bitmap.cols() != prev.cols() || bitmap.rows() != prev.rows())
      {
         resizeWindow(bitmap.cols(), bitmap.rows());
      }
   }

   /**
    * Get the screen bitmap for the current window.
    * 
    * @return   See above.
    */
   public ScreenBitmap getScreenBitmap()
   {
      return bitmap;
   }

   /**
    * Start a batch drawing routine.
    */
   public void startBatch()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.START_BATCH);
      offer(ps);
   }

   /**
    * End a batch drawing routine.
    */
   public void endBatch()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.END_BATCH);
      offer(ps);
      
      repaint();
      
      synchronized (getTreeLock())
      {
         getTreeLock().notifyAll();
      }
   }
   
   /**
    * Clear the entire screen, with the default window color.
    */
   @Override
   public void clear()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR);
      ps.x = 0;
      ps.y = 0;
      ps.width = getWidth();
      ps.height = getHeight();
      ps.color = null; // TODO: color???
      
      offer(ps);
   }

   /**
    * Clear specified area at the screen.
    * 
    * @param   left
    *          Left coordinate of the area.
    * @param   top
    *          Top coordinate of the area.
    * @param   right
    *          Right coordinate of the area.
    * @param   bottom
    *          Bottom coordinate of the area.
    * @param   color
    *          Color.
    */
   public void clearArea(int left, int top, int right, int bottom, com.goldencode.p2j.ui.Color color)
   {
      // this is a no-op for now; it doesn't really do anything
      /*PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR_AREA);
      ps.x = left;
      ps.y = right;
      ps.width = right - left;
      ps.height = bottom - top;
      ps.color = null; // TODO: color???
      
      offer(ps);*/
   }
   
   /**
    * Draws a highlighted rectangular area of the specified color and bounds.
    * The color is alpha-blended with the content of the screen buffer.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          area width
    * @param   height
    *          area height
    * @param   color
    *          the highlight color
    */
   public void drawHighlight(int x, int y, int width, int height, Color color)
   {
      if (width <= 0 || height <= 0)
      {
         return;
      }
      
      clearHighlight();
      
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_HIGHLIGHT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.color = color;
      
      offer(ps);
   }
   
   /**
    * Restores the screen area previously highlighted by 
    * {@link #drawHighlight(int, int, int, int, Color)}.
    */
   public void clearHighlight()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.CLEAR_HIGHLIGHT);
      offer(ps);
   }
   
   /**
    * Save (to {@link #savedFocusListeners} and remove the OS window's focus listeners.
    */
   public void saveFocusListeners()
   {
      WindowFocusListener[] flisteners = frame.getWindowFocusListeners();
      for (WindowFocusListener l : flisteners)
      {
         frame.removeWindowFocusListener(l);
      }
      
      this.savedFocusListeners = flisteners;
   }

   /**
    * Restore the OS window's focus listeners, from the {@link #saveFocusListeners()} list.
    */
   public void restoreFocusListeners()
   {
      for (WindowFocusListener l : savedFocusListeners)
      {
         frame.addWindowFocusListener(l);
      }
      this.savedFocusListeners = null;
   }

   /**
    * Move this window to top, at the OS level.
    */
   public void moveToTop()
   {
      if (!isWindowVisible())
      {
         return;
      }
   
      try
      {
         SwingUtilities.invokeAndWait(new Runnable()
         {
            @Override
            public void run()
            {
               frame.setAlwaysOnTop(true);
               frame.toFront();
               frame.requestFocus();
               frame.setAlwaysOnTop(false);
            }
         });
      }
      catch (InvocationTargetException | InterruptedException e)
      {
         throw new RuntimeException(e);
      }
   }
   
   /**
    * Determine if the window is on top of all windows, including OS.
    * 
    * @return   See above.
    */
   public boolean isOnTop()
   {
      return frame.isActive();
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   public void enableEvents(boolean capture)
   {
      if (capture)
      {
         if (savedListeners == null)
         {
            return;
         }
         
         List<EventListener> events;
         
         events = savedListeners.remove(MouseListener.class);
         for (EventListener l : events)
         {
            addMouseListener((MouseListener) l);
         }
         
         events = savedListeners.remove(MouseMotionListener.class);
         for (EventListener l : events)
         {
            addMouseMotionListener((MouseMotionListener) l);
         }
   
         events = savedListeners.remove(MouseWheelListener.class);
         for (EventListener l : events)
         {
            addMouseWheelListener((MouseWheelListener) l);
         }
   
         events = savedListeners.remove(WindowListener.class);
         for (EventListener l : events)
         {
            frame.addWindowListener((WindowListener) l);
         }
   
         events = savedListeners.remove(WindowFocusListener.class);
         for (EventListener l : events)
         {
            frame.addWindowFocusListener((WindowFocusListener) l);
         }
   
         events = savedListeners.remove(WindowStateListener.class);
         for (EventListener l : events)
         {
            frame.addWindowStateListener((WindowStateListener) l);
         }
   
         events = savedListeners.remove(KeyListener.class);
         for (EventListener l : events)
         {
            frame.addKeyListener((KeyListener) l);
         }
         
         savedListeners = null;
      }
      else
      {
         if (savedListeners != null)
         {
            return;
         }
   
         savedListeners = new HashMap<>();
   
         List<EventListener> events;
         
         events = new ArrayList<>();
         for (MouseListener l : getMouseListeners())
         {
            removeMouseListener(l);
            events.add(l);
         }
         savedListeners.put(MouseListener.class, Collections.unmodifiableList(events));
         
         events = new ArrayList<>();
         for (MouseMotionListener l : getMouseMotionListeners())
         {
            removeMouseMotionListener(l);
            events.add(l);
         }
         savedListeners.put(MouseMotionListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (MouseWheelListener l : getMouseWheelListeners())
         {
            removeMouseWheelListener(l);
            events.add(l);
         }
         savedListeners.put(MouseWheelListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (WindowListener l : frame.getWindowListeners())
         {
            frame.removeWindowListener(l);
            events.add(l);
         }
         savedListeners.put(WindowListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (WindowFocusListener l : frame.getWindowFocusListeners())
         {
            frame.removeWindowFocusListener(l);
            events.add(l);
         }
         savedListeners.put(WindowFocusListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (WindowStateListener l : frame.getWindowStateListeners())
         {
            frame.removeWindowStateListener(l);
            events.add(l);
         }
         savedListeners.put(WindowStateListener.class, Collections.unmodifiableList(events));
   
         events = new ArrayList<>();
         for (KeyListener l : frame.getKeyListeners())
         {
            frame.removeKeyListener(l);
            events.add(l);
         }
         savedListeners.put(KeyListener.class, Collections.unmodifiableList(events));
      }
   }

   /**
    * Get the metrics of the given font.  If the {@link #g2 graphics} have been obtained, call the
    * {@link #getFontMetrics} and the graphics object.  Otherwise, use the simulator's graphics.
    *  
    * @param    font
    *           The font for which font metrics is to be obtained.
    *           
    * @return   the font metrics for <code>font</code>.
    */
   @Override
   public FontMetrics getFontMetrics(Font font)
   {
      synchronized (getTreeLock())
      {
         FontMetrics fm = (g2 == null ? super.getFontMetrics(font) : g2.getFontMetrics(font));
         
         return fm;
      }
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   public int getTextHeight(String text, Font font)
   {
      synchronized (getTreeLock())
      {
         FontMetrics fm = getFontMetrics(font);
         LineMetrics lm = fm.getLineMetrics(text, g2 == null ? getGraphics() : g2);
         
         return (int) lm.getHeight();
      }
   }

   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.
    * @param    legacyWidth
    *           The legacy width which needs to be matched.
    * @param    legacyHeight
    *           The legacy height which needs to be matched.
    *           
    * @return   The scaled font.
    */
   public Font scaleFont(Font font, int legacyWidth, int legacyHeight)
   {
      FontMetrics fm = getFontMetrics(font);
   
      int avgWidth = 0;
      for (int ch = 0x20; ch <= 0xff; ch++)
      {
         avgWidth += fm.charWidth(ch);
      }
      avgWidth = (int) Math.ceil((double) avgWidth / (double) (0xff - 0x20 + 1));
      int height = fm.getHeight();
      
      Font scaled = font;
      
      if (height != legacyHeight || avgWidth != legacyWidth)
      {
         double scaleX = (double) legacyWidth / (double) avgWidth;
         double scaleY = (double) legacyHeight / (double) height;
   
         scaled = font.deriveFont(AffineTransform.getScaleInstance(scaleX, scaleY));
      }
      
      return scaled;
   }

   /**
    * Resize the screen buffers so that they will have the new width/height.
    * <p>
    * This will be a no-op if either the width or the height is not a positive number.
    * 
    * @param   width
    *          The new width.
    * @param   height
    *          The new height.
    */
   private void resizeScreenBuffers(int width, int height)
   {
      if (width <= 0 || height <= 0)
      {
         return;
      }
      
      synchronized (getTreeLock())
      {
         BufferedImage resizedImage = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         if (screenBuffer != null)
         {
            Graphics2D g = resizedImage.createGraphics();
            g.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
            g.drawImage(screenBuffer, 0, 0, null);
            g.dispose();
         }
         
         screenBuffer = resizedImage;
         if (g2 != null)
         {
            g2.dispose();
         }
   
         g2 = screenBuffer.createGraphics();
         g2.translate(translateX, translateY);
         g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_OFF);
         if (!clippings.isEmpty())
         {
            g2.clip(clippings.peek());
         }
   
         getTreeLock().notifyAll();
         repaint();
      }
   }

   /**
    * Update screen. Iterate over the pool and draw.
    * 
    * @param   graphics
    *          Graphical context.
    */
   private void drawOnScreenBitmap(Graphics graphics)
   {
      // ensure the drawing will be performed by only one thread at a time
      synchronized (getTreeLock())
      {
         int waitForBatch = 0;
         while (!graphicPool.isEmpty() || waitForBatch != 0)
         {
            // if a batch was started, wait for it to end.
            if (graphicPool.isEmpty() && waitForBatch != 0)
            {
               try
               {
                  getTreeLock().wait();
               }
               catch (InterruptedException e)
               {
                  // terminate painting
                  break;
               }
            }
            
            PaintStructure ps = graphicPool.pollFirst();

            if (ps != null)
            {
               switch (ps.id)
               {
                  case FLUSH:
                     graphics.drawImage(screenBuffer, 0, 0, this);
                     break;
                  
                  case START_BATCH:
                     waitForBatch++;
                     break;

                  case END_BATCH:
                     waitForBatch--;
                     break;
                     
                  default:
                     draw(g2, ps);
                     break;
               }
            }
         }
   
         graphics.drawImage(screenBuffer, 0, 0, this);
      }
   }
   
   /**
    * Draw using a paint operation.
    * 
    * @param   g2
    *          Graphical context.
    * @param   ps
    *          Paint operation parameters.
    */
   private void draw(Graphics2D g2, PaintStructure ps)
   {
      switch (ps.id)
      {
         case DRAW_STRING:
            drawString(g2, ps);
            break;
         case DRAW_STRING_SCALED:
            drawStringScaled(g2, ps);
            break;
         case DRAW_LINE:
            g2.drawLine(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_RECT:
            g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case DRAW_ROUND_RECT:
            g2.drawRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter, ps.arcDiameter);
            break;
         case DRAW_IMAGE:
            g2.drawImage(ps.img, ps.x, ps.y, ps.width, ps.height, this);
            break;
         case FILL_RECT:
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case FILL_ROUND_RECT:
            g2.fillRoundRect(ps.x, ps.y, ps.width, ps.height, ps.arcDiameter, ps.arcDiameter);
            break;
         case FILL_POLYGON:
            g2.fillPolygon(ps.xPoints, ps.yPoints, ps.nPoints);
            break;
         case DRAW_3D_RECT:
            g2.draw3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case FILL_3D_RECT:
            g2.fill3DRect(ps.x, ps.y, ps.width, ps.height, ps.raised);
            break;
         case SET_COLOR:
            g2.setColor(ps.color);
            break;
         case SET_LINE_STROKE:
            g2.setStroke(ps.stroke);
            break;
         case SET_FONT:
            g2.setFont(ps.font);
            break;
         case SET_FONT_STYLE:
            g2.setFont(g2.getFont().deriveFont(ps.x));
            break;
         case TRANSLATE:
            g2.translate(ps.x, ps.y);
            translateX += ps.x;
            translateY += ps.y;
            break;
         case CLEAR_AREA:
            if (ps.color != null)
            {
               g2.setColor(ps.color);
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLEAR:
            if (ps.color != null)
            {
               g2.setColor(ps.color);
            }
            g2.fillRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case CLIP:
            java.awt.Rectangle prevClip = g2.getClipBounds();
            if (prevClip != null)
            {
               clippings.push(prevClip);
            }
            
            // debugging aid: enable these lines to outline the clipped area
            if (false)
            {
               g2.setColor(Color.green);
               g2.drawRect(ps.x, ps.y, ps.width, ps.height);
            }
            
            g2.clipRect(ps.x, ps.y, ps.width, ps.height);
            break;
         case NO_CLIP:
            java.awt.Rectangle clip = clippings.isEmpty() ? null : clippings.pop();
            g2.setClip(clip);
            break;
         case SET_LOCATION:
            frame.setLocation(ps.x, ps.y);
            break;
         case RESIZE_WINDOW:
            Dimension sz = new Dimension(ps.width, ps.height);

            setPreferredSize(sz);
            setMinimumSize(sz);
            setMaximumSize(sz);
            
            frame.setSize(sz);
            setSize(sz);
            
            frame.pack();

            resizeScreenBuffers(ps.width, ps.height);
            break;
         case DRAW_HIGHLIGHT:
            drawHighlightWorker(ps.x, ps.y, ps.width, ps.height, ps.color);
            break;
         case CLEAR_HIGHLIGHT:
            clearHighlightWorker();
            break;
         default:
            break;
      }      
   }
   
   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * Once this API is finished, the graphic's font is restored to its state as before the call.
    *  
    * @param   g2
    *          The graphics used to draw the text.
    * @param   ps
    *          The structure with the drawing details.
    */
   private void drawStringScaled(Graphics2D g2, PaintStructure ps)
   {
      Font origFont = g2.getFont();
      
      FontMetrics fm = getFontMetrics(origFont);
      
      int txtWidth = fm.stringWidth(ps.text);
      LineMetrics lm = fm.getLineMetrics(ps.text, g2);
      int txtHeight = (int) lm.getHeight();
      
      if (txtWidth == ps.width && txtHeight == ps.height)
      {
         drawString(g2, ps);
         return;
      }

      double scaleX = (double) ps.width / txtWidth;
      double scaleY = (double) ps.height / txtHeight;

      Font scaled = origFont.deriveFont(AffineTransform.getScaleInstance(scaleX, scaleY));
      g2.setFont(scaled);
      try
      {
         fm = getFontMetrics(scaled);
         lm = fm.getLineMetrics(ps.text, g2);

         ps.y = ps.y - lm.getBaselineIndex() - Math.round(lm.getDescent());
         drawString(g2, ps);
      }
      finally
      {
         g2.setFont(origFont);
      }
   }

   /**
    * Draw a string using the current font.
    *  
    * @param   g2
    *          The graphics used to draw the text.
    * @param   ps
    *          The structure with the drawing details.
    */
   private void drawString(Graphics2D g2, PaintStructure ps)
   {
      g2.drawString(ps.text, ps.x, ps.y);
   }

   /**
    * This is the worker method for drawing a highlighted rectangular area of the specified
    * color and bounds. The color is alpha-blended with the content of the screen buffer.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          area width
    * @param   height
    *          area height
    * @param   color
    *          the highlight color
    */
   private void drawHighlightWorker(int x, int y, int width, int height, Color color)
   {
      // keep the original composite and color to restore later
      Composite prevComposite = g2.getComposite();
      Color prevColor = g2.getColor();
      
      try 
      {
         // save the area for later to allow clear
         highlightRestoreBuffer = new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
         Graphics2D gh = highlightRestoreBuffer.createGraphics();
         int sx = x - translateX;
         int sy = y - translateY;
         gh.drawImage(screenBuffer, 0, 0, width, height, sx, sy, sx+width, sy+height, this);
         
         // create and set new composite for the highlight
         AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC_OVER, 0.5f);
         g2.setComposite(ac);
   
         // draw the highlight
         g2.setColor(color);
         g2.fillRect(x - translateX, y - translateY, width, height);
         highlightX = x;
         highlightY = y;
      }
      finally
      {
         g2.setColor(prevColor);;
         g2.setComposite(prevComposite);
      }
   }
   
   /**
    * This is the worker method for restoring the screen area previously highlighted by
    * {@link #drawHighlightWorker(int, int, int, int, Color)}.
    */
   private void clearHighlightWorker()
   {
      if (highlightRestoreBuffer == null)
      {
         // nothing to restore
         return;
      }
      
      int w = highlightRestoreBuffer.getWidth();
      int h = highlightRestoreBuffer.getHeight();
      
      int dx = highlightX - translateX;
      int dy = highlightY - translateY;
      
      Composite prevComposite = g2.getComposite();
      AlphaComposite ac = AlphaComposite.getInstance(AlphaComposite.SRC, 1f);
      g2.setComposite(ac);
      g2.drawImage(highlightRestoreBuffer, dx, dy, dx+w, dy+h, 0, 0, w, h, this);
      g2.setComposite(prevComposite);
      
      highlightRestoreBuffer = null;
   }
   
   /**
    * Returns display size.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The dimensions in pixels.
    */
   public NativeDimension getDisplaySize()
   {
      Rectangle displayBounds = frame.getGraphicsConfiguration().getBounds(); 
      return new NativeDimension(displayBounds.width, displayBounds.height);
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in pixels.
    */
   public NativeRectangle getDisplayWorkArea()
   {
      Insets is = frame.getToolkit().getScreenInsets(frame.getGraphicsConfiguration());
      return new NativeRectangle(is.top, is.left, is.bottom, is.right);
   }
}
