/*
** Module   : MouseMoveable.java
** Abstract : Implementation of a mouse moveable widget.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20140805 Created initial version.
** 002 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 003 HC  20141207 Improved window mouse drag and mouse resize behavior.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.event.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.*;

/**
 * Listener for mouse actions which can trigger a widget to be moved.
 */
class MouseMoveable
extends MouseWidgetAction
{
   /** The last captured mouse event, for this widget. */
   private MouseEvent lastMouseEvent;
   
   /** The saved event of left mouse button down. */
   private MouseEvent leftButtonDownEvent;
   
   /** The screen location of the associated widget when left button down. */
   private NativePoint leftButtonDownWidgetLocation;

   /**
    * Create a new mouse moveable action, for the specified widget.
    * 
    * @param    wid
    *           The widget ID.
    */
   public MouseMoveable(int wid)
   {
      super(wid);
   }

   /**
    * Notification of a mouse exited event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      if (lastMouseEvent == null)
      {
         WindowGuiImpl window = (WindowGuiImpl) widget.window();
         gd.selectWindow(window.config().id.getId());
      }
   }
   
   /**
    * Notification of a mouse dragged event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      if (lastMouseEvent == null)
      {
         return;
      }
      
      if (lastMouseEvent.getButton() != MouseEvent.NOBUTTON &&
          lastMouseEvent.getButton() != MouseEvent.BUTTON1)
      {
         // drag is only possible by left button
         return;
      }

      int evtID = lastMouseEvent.getID();
      if (evtID == MouseEvent.MOUSE_DRAGGED || evtID == MouseEvent.MOUSE_PRESSED)
      {
         //get the absolute mouse drag delta
         int dx = e.getXOnScreen() - leftButtonDownEvent.getXOnScreen();
         int dy = e.getYOnScreen() - leftButtonDownEvent.getYOnScreen();
         
         NativePoint currLoc = widget.displayPhysicalLocation();
         int currWidgetDx = currLoc.x - leftButtonDownWidgetLocation.x; 
         int currWidgetDy = currLoc.y - leftButtonDownWidgetLocation.y;
         
         widget.moveBy(dx - currWidgetDx, dy - currWidgetDy);
      }
      
      this.lastMouseEvent = e;
   }

   /**
    * Notification of a mouse pressed event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      if (e.getButton() == MouseEvent.BUTTON1)
      {
         this.lastMouseEvent = e;
         this.leftButtonDownEvent = e;
         this.leftButtonDownWidgetLocation = widget.displayPhysicalLocation();
      }
   }

   /**
    * Notification of a mouse released event for this {@link MouseWidgetAction#widget}.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      this.lastMouseEvent = null;
      this.leftButtonDownEvent = null;
      this.leftButtonDownWidgetLocation = null;
   }
}
