/*
** Module   : SwingGuiDriver.java
** Abstract : Swing GUI screen driver.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140508 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140805 Refactored to implement new GenericScreenDriver APIs and the support provided 
**                  by it.  Added mouse support and other misc APIs related to window management.
** 004 MAG 20140902 Integrate COLOR-TABLE. Added support for BGCOLOR and FGCOLOR color conversion.
**                  Added system colors used solely for widget drawing.  
** 005 MAG 20140912 Implement GUI status bar.
** 006 HC  20140911 GUI coordinate support - data types refactoring.
** 007 MAG 20140925 Check if window has renderer.
** 008 CA  20141021 Added support to allow drawing of nested containers (the real location of the
**                  translated origin is required).  Misc changes related to image drawing and
**                  mouse processing.
** 009 VIG 20141022 TC.screenSize replaced with window.size.
** 010 HC  20141031 Implemented support for drawing highlighted areas.
** 011 MAG 20141029 Added clip rectangle parameter to draw method.
** 012 CA  20141029 Each simulator needs to receive the swing keyboard reader implementation in 
**                  use.  It's the callers responsibility to ensure that the same instance is used
**                  for all simulators, if the client is ran in a multi-window mode (i.e. GUI).
**                  The GUI driver ensures that only one instance is used, for each client.
** 013 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 014 EVL 20141116 Adding round rectangles draw and fill.
** 015 CA  20141123 Added APIs to allow enforcement of window stacking at OS level.
** 016 CA  20141125 Invalidate the current simulator, if the currently selected window is being 
**                  destroyed.  Added API to conditionally enable/disable OS-level events.
** 017 EVL 20141125 Adding dotted line draw capability.
** 018 CA  20141205 Added API to draw relative to window's real origin (top-left corner), 
**                  regardless of the current translated origin.
**                  Added API to register editable widgets (thus the driver can adjust the mouse 
**                  cursor accordingly).
** 019 CA  20141220 Added APIs for clipboard access.  Added APIs to allow text drawing and font 
**                  selection matching legacy sizes.
** 020 HC  20141207 Implemented runtime support for several SESSION attributes.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Color;
import java.awt.Image;
import java.awt.datatransfer.*;
import java.awt.event.*;
import java.io.*;
import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.Window;
import com.goldencode.p2j.ui.client.chui.driver.console.*;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.swing.*;
import com.goldencode.terminal.*;

/**
 * Swing-backed GUI driver.
 */
public class SwingGuiDriver
extends GuiScreenDriver<GuiSimulator>
{
   /**
    * The key reader.  This is global per the entire application, no matter how many frames are
    * alive.
    */
   private static final ContextLocal<SwingKeyboardReader> keyReader = 
   new ContextLocal<SwingKeyboardReader>()
   {
      protected SwingKeyboardReader initialValue()
      {
         return new WinKeyboardReader();
      };
   };

   /** Default terminal type name. */
   private String termType = "win3";

   /** System colors cache */
   private Map<Integer, Color> systemColorsCache = new HashMap<>();

   /** The real location of the translated origin. */
   private NativePoint translatedOrigin = new NativePoint(0, 0);

   /**
    * Convert bootstrap options.
    * 
    * @param   cfg
    *          Bootstrap configuration.
    *          
    * @return  Converted options.
    */
   public static GuiOptions convertToOptions(BootstrapConfig cfg)
   {
      return new GuiOptions();
   }
   
   /**
    * Read single keystroke from keyboard. Returned code should be either proper
    * Progress 4GL/P2J key code or -1 if environment-specific input cannot be 
    * properly translated.
    * <p>
    * The list of codes and detailed description is provided in {@link Keyboard}
    * class. The {@link KeyProcessor} can serve as an example of the key 
    * translation implementation.
    * 
    * @return  key code or -1 if no keys were pressed since last call.
    */
   @Override
   public int readKey()
   {
      return keyReader.get().read();
   }

   /**
    * Create a new GUI simulator.
    * 
    * @param    options
    *           The window options.
    *           
    * @return   See above.
    */
   @Override
   public GuiSimulator newSimulator(GuiOptions options)
   {
      final GuiSimulator sim;

      try
      {
         sim = GuiClient.load(options, keyReader.get());
         // the real window size is unknown at this time; it will be set later, when the layout
         // is performed
         sim.setScreenBitmap(new ScreenBitmap(0, 0));
      }
      catch (InvocationTargetException | InterruptedException e)
      {
         throw new IllegalStateException("Could not create simulator!", e);
      }
      
      return sim;
   }
   
   /**
    * Translates the origin of the <code>Graphics2D</code> context to the
    * point (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
    * Modifies the <code>Graphics2D</code> context so that its new origin
    * corresponds to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * <code>Graphics2D</code> context's former coordinate system.  
    * All coordinates used in subsequent rendering operations on this graphics
    * context are relative to this new origin.
    *
    * @param   x 
    *          the specified x coordinate
    * @param   y 
    *          the specified y coordinate
    */
   @Override
   public void translate(int x, int y)
   {
      direct.translate(x, y);

      // this is mutable...
      translatedOrigin.translate(x, y);
   }  
   
   /**
    * Draw a string at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   @Override
   public void drawString(String text, int x, int y)
   {
      direct.drawString(text, x, y);
   }      

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   legacyWidth
    *          The legacy width, which needs to be matched when drawing.
    * @param   legacyHeight
    *          The legacy width, which needs to be matched when drawing.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   @Override
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y)
   {
      direct.drawStringScaled(text, legacyWidth, legacyHeight, x, y);
   }

   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param   x1  
    *          the first point's <i>x</i> coordinate.
    * @param   y1  
    *          the first point's <i>y</i> coordinate.
    * @param   x2  
    *          the second point's <i>x</i> coordinate.
    * @param   y2  
    *          the second point's <i>y</i> coordinate.
    */
   @Override
   public void drawLine(int x1, int y1, int x2, int y2)
   {
      direct.drawLine(x1, y1, x2, y2);
   }      

   /**
    * Draw a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   @Override
   public void drawRect(int x, int y, int width, int height)
   {
      direct.drawRect(x, y, width, height);
   }      

   /**
    * Draw a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */
   @Override
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      direct.drawRoundRect(x, y, width, height, arcDiameter);
   }      

   /**
    * Fill a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   @Override
   public void fillRect(int x, int y, int width, int height)
   {
      direct.fillRect(x, y, width, height);
   }      

   /**
    * Fill a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */   
   @Override
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      direct.fillRoundRect(x, y, width, height, arcDiameter);
   }      

   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>Color</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be drawn.
    * @param   y 
    *          the y coordinate of the rectangle to be drawn.
    * @param   width 
    *          the width of the rectangle to be drawn.
    * @param   height 
    *          the height of the rectangle to be drawn.
    * @param   raised 
    *          a boolean that determines whether the rectangle appears 
    *          to be raised above the surface or sunk into the surface.
    */
   @Override
   public void draw3DRect(int x, int y, int width, int height, boolean raised)
   {
      direct.draw3DRect(x, y, width, height, raised);
   }      

   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>Color</code>.  This method uses
    * the current <code>Color</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be filled.
    * @param   y 
    *          the y coordinate of the rectangle to be filled.
    * @param   width
    *          the width of the rectangle to be filled.
    * @param   height 
    *          the height of the rectangle to be filled.
    * @param   raised 
    *          a boolean value that determines whether the rectangle appears 
    *          to be raised above the surface or etched into the surface.
    */   
   @Override
   public void fill3DRect(int x, int y, int width, int height, boolean raised)
   {
      direct.fill3DRect(x, y, width, height, raised);
   }      
   
   /**
    * Fill polygon.
    * 
    * @param   xPoints
    *          X coordinates.
    * @param   yPoints
    *          Y coordinates.
    */   
   @Override
   public void fillPolygon(int[] xPoints, int[] yPoints)
   {
      direct.fillPolygon(xPoints, yPoints);      
   }
   
   /**
    * Draw a image at the specified location.
    * 
    * @param   img
    *          Image to draw.
    * @param   x
    *          left location.
    * @param   y
    *          top location.
    * @param   w
    *          the image width.
    * @param   h
    *          the image height.
    */
   @Override
   public void drawImage(Image img, int x, int y, int w, int h)
   {
      direct.drawImage(img, x, y, w, h);
   }

   /**
    * Draws a highlighted rectangular area of the specified color and bounds.
    * The color is alpha-blended with the content of the screen buffer.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          area width
    * @param   height
    *          area height
    * @param   color
    *          the highlight color
    */
   @Override
   public void drawHighlight(int x, int y, int width, int height, Color color)
   {
      sim.drawHighlight(x, y, width, height, color); 
   }

   /**
    * Clears the highlighted area drawn by the previous call of 
    * {@link #drawHighlight(int, int, int, int, Color)}.
    */
   @Override
   public void clearHighlight()
   {
      sim.clearHighlight();
   }

   /**
   * Set current color used on draw operations.
   * 
   * @param   color
   *          Drawing color.
   */    
   @Override
   public void setColor(Color color)
   {
      direct.setColor(color);      
   }
   
   /**
    * Set current font used on draw operations.
    * 
    * @param   font
    *          Drawing font.
    */
   @Override
   public void setGuiFont(Font font)
   {
      direct.setGuiFont(font);      
   }
   
   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.
    * @param    legacyWidth
    *           The legacy width which needs to be matched.
    * @param    legacyHeight
    *           The legacy height which needs to be matched.
    *           
    * @return   The scaled font.
    */
   @Override
   public Font scaleFont(Font font, int legacyWidth, int legacyHeight)
   {
      return direct.scaleFont(font, legacyWidth, legacyHeight);
   }
   
   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param   style
    *          Font style.
    */
   @Override
   public void setFontStyle(int style)
   {
      direct.setFontStyle(style);
   }
   
   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param   stroke
    *          Line style.
    */
   @Override
   public void setLineStroke(Stroke stroke)
   {
      direct.setLineStroke(stroke);
   }

   /**
    * Resize window.
    * 
    * @param   width
    *          new width.
    * @param   height
    *          new height
    */
   public void resizeWindow(int width, int height)
   {
      direct.resizeWindow(width, height);
   }   
   
   /**
    * On a GUI interface multiple windows could be created. Each window has it own renderer.
    * When a window is draw first the window must call this method to setup it renderer as 
    * the current graphical canvas.  
    * 
    * @param   windowId
    *          Window ID.
    */
   public void selectWindow(int windowId)
   {
      if (sim == null || sim.getWindowId() != windowId)
      {
         direct.selectWindow(windowId);
         
         sim = direct.getSimulator();
      }
   }   
   
   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   public void setWindowLocation(int left, int top)
   {
      direct.setWindowLocation(left, top);
   }
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      return direct.getWindowLocation();
   }

   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWindowDimension()
   {
      return direct.getWindowDimension();
   }

   /**
    * Register this window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    sim
    *           The simulator in charge of this window.
    */
   @Override
   public void registerWindow(int windowId, GuiSimulator sim)
   {
      direct.registerWindow(windowId, sim);
   }

   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void deregisterWindow(int windowId)
   {
      direct.deregisterWindow(windowId);
      
      if (sim != null && windowId == sim.getWindowId())
      {
         sim = null;
      }
   }   

   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    name
    *           The font face name.
    * @param    size
    *           The font's point size.
    * @param    bold
    *           Flag indicating if the font is bold.
    * @param    italic
    *           Flag indicating if the font is italic.
    * @param    underline
    *           Flag indicating if the font is underline.
    * @param    definition
    *           The font's definition.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * 
    * @return   A font instance.
    */
   @Override
   public Font createFont(String  name,
                          int     size,
                          boolean bold,
                          boolean italic,
                          boolean underline,
                          byte[]  definition,
                          int     legacyWidth,
                          int     legacyHeight)
   {
      return direct.createFont(name, size, bold, italic, underline, definition, 
                                   legacyWidth, legacyHeight);
   }

   /**
    * Get the width of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's width.
    */
   @Override
   public int getFontWidth(Font font)
   {
      return direct.getFontWidth(font);
   }

   /**
    * Get the height of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's height.
    */
   @Override
   public int getFontHeight(Font font)
   {
      return direct.getFontHeight(font);
   }

   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   @Override
   public int getTextWidth(String text, Font font)
   {
      return direct.getTextWidth(text, font);
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, Font font)
   {
      return direct.getTextHeight(text, font);
   }

   /**
    * Change terminal type to specified one. The terminal-based implementation
    * should try to change terminal type to specified one. It is assumed that 
    * passed terminal string is specific to particular driver implementation 
    * and recognized by it. 
    * 
    * @param   name
    */
   @Override
   public void setTerminalType(String name)
   {
      termType = name;
   }

   /**
    * Obtain system/implementation specific terminal type string.
    * <p>
    * At present this is used to implement 4GL TERMINAL function and do some 
    * terminal-specific workarounds. At present workarounds exist for "vt320" 
    * terminal type. If implementation requires no special workarounds, then 
    * it should avoid returning "vt320" as terminal type string.  
    * 
    * @return  terminal type string.
    */
   @Override
   public String terminalType()
   {
      return termType;
   }

   /**
    * Get the default WINDOW-SYSTEM used by this driver. See {@link ClientParameters} for the list
    * of possible values.
    * 
    * @return   See above.
    */
   @Override
   public String getWindowSystem()
   {
      return ClientParameters.WINDOW_SYSTEM_MS_WINDOWS;
   }
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible)
   {
      direct.setWindowVisible(visible);
   }
   
   /**
    * Iconify the window.
    */
   @Override
   public void iconifyWindow()
   {
      direct.iconifyWindow();
   }

   /**
    * Create a shell emulator and execute the given command.
    * 
    * @param    cmdList
    *           The command.
    * @param    overlay
    *           Flag identifying if the emulator window will overlay the current window or will
    *           open in a new window.
    *           
    * @return   A VT100 emulator executing the given command.
    */
   @Override
   protected VT100Emulator createShellEmulator(String[] cmdList, int overlay)
   {
      // TODO: this requires a ChUI screen driver, to execute the emulator in it.

      return null;
   }

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    type
    *           The cursor type, one of the {@link Cursor} constants.
    */
   @Override
   public void setCursor(int type)
   {
      direct.setCursor(Cursor.getPredefinedCursor(type));
   }

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the specified point, execute the drawing and restore the origin of the graphics
    * context.
    * 
    * @param    origin
    *           The drawing origin.
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void draw(NativePoint origin, NativeRectangle clip, Runnable core)
   {
      translate(origin.x, origin.y);

      if (clip != null)
      {
         setDrawingArea(clip);
      }
      
      try
      {
         core.run();
      }
      finally
      {
         if (clip != null)
         {
            setDrawingArea(null);
         }
         // restore context
         translate(-origin.x, -origin.y);
      }
   }

   /**
    * Execute a widget drawing bracket.  This includes translating the origin of the graphics
    * context to the window's real (0,0) origin (top-left corner) and then translating to the
    * specified location, before executing the drawing.
    * <p>
    * The translated origin (as before this was called) needs to be restored, before this call
    * ends.
    * 
    * @param    origin
    *           The drawing origin (relative to window's real origin).
    * @param    clip
    *           Clip rectangle.
    * @param    core
    *           The drawing worker.
    */
   @Override
   public void drawFromOrigin(NativePoint origin, NativeRectangle clip, Runnable core)
   {
      NativePoint save = new NativePoint(this.translatedOrigin);
      translate(-save.x, -save.y);
      
      try
      {
         draw(origin, clip, core);
      }
      finally
      {
         // move current translated back to screen (0,0)
         translate(-this.translatedOrigin.x, -this.translatedOrigin.y);
         // translate to the saved origin
         translate(save.x, save.y);
      }
   }
   
   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   @Override
   public void setDrawingArea(NativeRectangle bounds)
   {
      direct.setDrawingArea(bounds);
   }

   /**
    * Register the specified widget as a moveable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerMoveableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new MouseMoveable(wid));
   }

   /**
    * Register the specified widget as a resizeable widget, via mouse.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void registerResizeableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new MouseResizeable(wid));
   }

   /**
    * Register the specified widget for popup events, via mouse.
    * <p>
    * The popup with the given ID will be shown when the specified mouse button is pressed inside
    * the specified widget.
    * 
    * @param    wid
    *           The widget id.
    * @param    btn
    *           The button to be pressed, when showing the popup.
    * @param    pid
    *           The popup widget ID.
    */
   @Override
   public void registerWidgetPopup(int wid, int btn, int pid)
   {
      MouseHandler.registerWidgetAction(wid, new MousePopupable(wid, btn, pid));
   }

   /**
    * Register this widget as being editable, so that the mouse cursor will adjust properly when
    * the widget is entered.
    * 
    * @param    wid
    *           The widget id.
    */
   @Override
   public void registerEditableWidget(int wid)
   {
      MouseHandler.registerWidgetAction(wid, new EditableWidget(wid));
   }

   /**
    * Deregister the specified widget for mouse events.
    * 
    * @param    wid
    *           The widget ID.
    */
   @Override
   public void deregisterWidget(int wid)
   {
      MouseHandler.deregisterWidget(wid);
   }

   /**
    * Enable or disable capture of mouse events.
    * 
    * @param    state
    *           {@code true} to capture mouse events, {@code false} to ignore them.
    *           
    * @return   The previous state.
    */
   @Override
   public boolean captureMouseEvents(boolean state)
   {
      return sim.captureMouseEvents(state);
   }

   /**
    * Handle the specified event, in the context of the specified widget.
    * 
    * @param    sourceID
    *           The source widget ID.
    * @param    evt
    *           The mouse event.
    */
   @Override
   public void handleMouseEvent(int sourceID, MouseEvent evt)
   {
      MouseHandler.handleMouseEvent(sourceID, evt);
   }
      
   /**
    * Get BGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index and convert the
    * value to a {@link java.awt.Color#Color(int)} instance.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of Color
    */
   @Override
   public Color getBgColor(int index)
   {
      int rgb = -1;

      int windowId = sim.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      if (window != null)
      {
         rgb = ColorManager.getBgColor(window, index);
      }
      
      return new Color(rgb);
   }

   /**
    * Get FGCOLOR color. Get a RGB value from COLOR-TABLE at the specified index and convert the
    * value to a {@link java.awt.Color#Color(int)} instance.
    * 
    * @param   index
    *          Color number.
    *          
    * @return  instance of Color
    */
   @Override
   public Color getFgColor(int index)
   {
      int rgb = -1;
      
      int windowId = sim.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      if (window != null)
      {
         rgb = ColorManager.getFgColor(window, index);
      }
      
      return new Color(rgb);
   }
 
   /**
    * Get a system color.  Because system colors cannot be change at runtime a cache is used in
    * order to optimize the creation of <code>Color</code> instances.  
    * 
    * @param   index
    *          Color index.
    *          
    * @return  see above.
    */
   @Override
   public Color getSysColor(int index)
   {
      // first lookup in cache
      Color color = systemColorsCache.get(index);
      if (color == null)
      {
         // if not found in cache get from system colors table.
         color = new Color(ColorManager.getSysColor(index));
         // cache value
         systemColorsCache.put(index, color);
      }
      return color;
   }

   /**
    * Get foreground color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   @Override
   public Color getFgStdColor(String stdColor)
   {
      int rgb = -1;
      
      int windowId = sim.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      if (window != null)
      {
         rgb = ColorManager.getStdColor(window, stdColor, true);
      }
      
      return new Color(rgb);
   }

   /**
    * Get background color for a standard color like NORMAL, INPUT or MESSAGES.
    * 
    * @param   stdColor
    *          Color name.
    *          
    * @return  see above.
    */
   @Override
   public Color getBgStdColor(String stdColor)
   {
      int rgb = -1;
      
      int windowId = sim.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      if (window != null)
      {
         rgb = ColorManager.getStdColor(window, stdColor, false);
      }
      
      return new Color(rgb);
   }
   
   /**
    * Force screen buffer draw.
    */
   public void flush()
   {
      direct.flush();
   }
   
   /**
    * Set status text message of current window.
    * 
    * @param   message
    *          Text message.
    */
   public void setStatusText(String message)
   {
      int windowId = sim.getWindowId();
      Window<?> window = (Window<?>) WindowManager.findWindow(windowId);
      setStatusText(window, message);
   }

   /**
    * Set status text message of specified window.
    * 
    * @param   message
    *          Text message.
    * @param   window
    *          Window instance.
    */
   @SuppressWarnings("rawtypes") 
   public void setStatusText(Window window, String message)
   {
      if (window != null)
      {
         window.getStatusArea().setText(message);
      }      
   }
   
   /**
    * Check if window has renderer.
    * 
    * @param   windowId
    *          Window ID.
    * 
    * @return  <code>true</code> if window has renderer <code>true</false> otherwise.
    */
   public boolean hasWindowRenderer(int windowId)
   {
      return direct.hasWindowRenderer(windowId);
   }

   /**
    * The real location where the origin was {@link #translatedOrigin translated}.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint translatedOrigin()
   {
      return new NativePoint(translatedOrigin);
   }

   /**
    * Notify the widget of any mouse events, generated after the given timestamp (in millis).
    * 
    * @param    wid
    *           The widget to be notified.
    * @param    when
    *           Only events generated after this timestamp will be notified.
    */
   @Override
   public void listenAnyMouseEvent(int wid, long when)
   {
      MouseHandler.listenAnyMouseEvent(wid, when);
   }
   
   /**
    * Check if the currently selected window is visible.
    * 
    * @return   See above.
    */
   @Override
   public boolean isWindowVisible()
   {
      return sim.isWindowVisible();
   }
   
   /**
    * Force the GUI windows to be stacked as in the specified order.
    * 
    * @param    windowIds
    *           The stack order to be enforced.
    */
   @Override
   public void stackWindows(int[] windowIds)
   {
      GuiSimulator oldSim = (sim == null ? null : sim);
      
      for (int wid : windowIds)
      {
         selectWindow(wid);
         sim.saveFocusListeners();
      }

      for (int wid : windowIds)
      {
         selectWindow(wid);
         sim.moveToTop();
      }

      for (int wid : windowIds)
      {
         selectWindow(wid);
         sim.restoreFocusListeners();
      }
      
      if (oldSim != null)
      {
         selectWindow(oldSim.getWindowId());
      }
   }

   /**
    * Determine if the given window is on top of all windows, including OS.
    * 
    * @param    windowId
    *           The window ID to check.
    *           
    * @return   See above.
    */
   public boolean isOnTop(int windowId)
   {
      int oldW = sim.getWindowId();
      
      selectWindow(windowId);
      
      try
      {
         return sim.isOnTop();
      }
      finally
      {
         selectWindow(oldW);
      }
   }

   /**
    * Enable or Disable OS-level events for the specified windows.
    * 
    * @param    windowIds
    *           The window IDs which need to be processed.
    * @param    capture
    *           Flag indicating OS level events are captured (when <code>true</code>).
    */
   @Override
   public void enableEvents(int[] windowIds, boolean capture)
   {
      GuiSimulator oldSim = (sim == null ? null : sim);
      
      for (int wid : windowIds)
      {
         selectWindow(wid);

         sim.enableEvents(capture);
      }
      
      if (oldSim != null)
      {
         selectWindow(oldSim.getWindowId());
      }
   }
   
   /**
    * Returns display size.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The dimensions in pixels.
    */
   @Override
   public NativeDimension getDisplaySize(int windowId)
   {
      return direct.getDisplaySize(windowId);
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea(int windowId)
   {
      return direct.getDisplayWorkArea(windowId);
   }

   /**
    * Copy the given text to the system clipboard.
    * 
    * @param    txt
    *           The text to be copied.
    */
   @Override
   public void copyToCliboard(String txt)
   {
      StringSelection stringSelection = new StringSelection(txt);
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
      clipboard.setContents(stringSelection, stringSelection);
   }
   
   /**
    * Get the current contents from the OS clipboard.
    * 
    * @return   See above.
    */
   @Override
   public String clipboardContents()
   {
      Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();

      Transferable content = clipboard.getContents(null);
      boolean hasTransferableText = 
         (content != null && content.isDataFlavorSupported(DataFlavor.stringFlavor));
      
      String result = null;
      if (hasTransferableText)
      {
         try
         {
           result = (String) content.getTransferData(DataFlavor.stringFlavor);
         }
         catch (UnsupportedFlavorException | IOException ex)
         {
            // ignore
         }
       }
      
      return result;
   }
}
