/*
** Module   : SwingGuiPrimitives.java
** Abstract : Swing GUI implementation of drawing primitives.
**
** Copyright (c) 2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description---------------------------------
** 001 MAG 20140523 First version.
** 002 CA  20140711 Enhanced font support.
** 003 CA  20140806 Added misc APIs related to window behavior/state.
** 004 MAG 20140905 Added fillPoly method.
** 005 HC  20140911 GUI coordinate support - data types refactoring.
** 006 MAG 20140925 Check if window has renderer.
** 007 CA  20141021 Changed the drawImage API to receive the image's width and height.
** 008 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 009 EVL 20141116 Adding round rectangles draw and fill.
** 010 EVL 20141125 Adding dotted line draw capability.
** 011 CA  20141220 Added APIs to allow text drawing and font selection matching legacy sizes.
** 012 HC  20141207 Implemented runtime support for several SESSION attributes.
*/

package com.goldencode.p2j.ui.client.gui.driver.swing;

import java.awt.*;
import java.awt.Image;
import java.util.*;

import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.gui.driver.*;

import java.awt.Color;
import java.awt.font.*;
import java.awt.geom.*;

import javax.swing.*;

/**
 * Swing GUI implementation of drawing primitives.
 */
public class SwingGuiPrimitives
implements GuiPrimitives
{
   /** A special map containing the attributes of a underlined font. */
   private static Map<TextAttribute, Integer> UNDERLINE_FONT = new HashMap<>();

   /** Windows renderer. */
   private Map<Integer, GuiSimulator> renderers = new HashMap<>();
   
   /** Current GUI renderer. */
   private GuiSimulator sim = null;

   static 
   {
      UNDERLINE_FONT.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
      UNDERLINE_FONT = Collections.unmodifiableMap(UNDERLINE_FONT);
   }

   /**
    * Register this window with the driver.
    * 
    * @param    windowId
    *           The window ID.
    * @param    sim
    *           The simulator in charge of this window.
    */
   @Override
   public void registerWindow(int windowId, GuiSimulator sim)
   {
      if (sim == null)
      {
         throw new NullPointerException("Null simulator is not allowed!");
      }
      
      renderers.put(windowId, sim);
      
      sim.setWindowId(windowId);
   }

   /**
    * Deregister this window from the driver.  This needs to be called only when the 
    * window widget is being destroyed.
    * 
    * @param    windowId
    *           The window ID.
    */
   @Override
   public void deregisterWindow(int windowId)
   {
      GuiSimulator sim = renderers.remove(windowId);
      sim.quit();
   }
   
   /**
    * On a GUI interface multiple windows could be created. Each window has it own renderer.
    * When a window is draw first the window must call this method to setup it renderer as 
    * the current graphical canvas.  
    * 
    * @param   windowId
    *          Window ID.
    */
   @Override
   public void selectWindow(int windowId)
   {
      sim = getSimulator(windowId);
   }
   
   /**
    * Translates the origin of the <code>Graphics2D</code> context to the
    * point (<i>x</i>,&nbsp;<i>y</i>) in the current coordinate system.
    * Modifies the <code>Graphics2D</code> context so that its new origin
    * corresponds to the point (<i>x</i>,&nbsp;<i>y</i>) in the
    * <code>Graphics2D</code> context's former coordinate system.  
    * All coordinates used in subsequent rendering operations on this graphics
    * context are relative to this new origin.
    *
    * @param   x 
    *          the specified x coordinate
    * @param   y 
    *          the specified y coordinate
    */
   @Override
   public void translate(int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.TRANSLATE);
      ps.x = x;
      ps.y = y;
      
      sim.offer(ps);
   }

   /**
    * Draw a string at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   @Override
   public void drawString(String text, int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING);
      ps.text = text;
      ps.x = x;
      ps.y = y;
      
      sim.offer(ps);
   }

   /**
    * Draw a string using the current font, which will be scaled so that the final string will
    * have the specified legacy width and height.
    * <p>
    * The string will be drawn at the specified location.
    * 
    * @param   text
    *          text string to draw.
    * @param   legacyWidth
    *          The legacy width, which needs to be matched when drawing.
    * @param   legacyHeight
    *          The legacy width, which needs to be matched when drawing.
    * @param   x
    *          left position.
    * @param   y
    *          top position.
    */
   @Override
   public void drawStringScaled(String text, int legacyWidth, int legacyHeight, int x, int y)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_STRING_SCALED);
      ps.text = text;
      ps.x = x;
      ps.y = y;
      ps.width = legacyWidth;
      ps.height = legacyHeight;
      
      sim.offer(ps);
   }
   
   /**
    * Draws a line, using the current color, between the points
    * <code>(x1,&nbsp;y1)</code> and <code>(x2,&nbsp;y2)</code>
    * in this graphics context's coordinate system.
    * 
    * @param   x1  
    *          the first point's <i>x</i> coordinate.
    * @param   y1  
    *          the first point's <i>y</i> coordinate.
    * @param   x2  
    *          the second point's <i>x</i> coordinate.
    * @param   y2  
    *          the second point's <i>y</i> coordinate.
    */
   @Override
   public void drawLine(int x1, int y1, int x2, int y2)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_LINE);
      ps.x = x1;
      ps.y = y1;
      ps.width = x2;
      ps.height = y2;
      
      sim.offer(ps);
   }

   /**
    * Draw a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */
   @Override
   public void drawRect(int x, int y, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      
      sim.offer(ps);
   }

   /**
    * Draw a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */
   @Override
   public void drawRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_ROUND_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.arcDiameter = arcDiameter;
      
      sim.offer(ps);
   }      

   /**
    * Fill a rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    */   
   @Override
   public void fillRect(int x, int y, int width, int height)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      
      sim.offer(ps);
   }
   
   /**
    * Fill a rounded rectangle.
    * 
    * @param   x
    *          left position
    * @param   y
    *          top position
    * @param   width
    *          rectangle width.
    * @param   height
    *          rectangle height.
    * @param   arcDiameter
    *          diameter of the arc at corners.
    */   
   @Override
   public void fillRoundRect(int x, int y, int width, int height, int arcDiameter)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_ROUND_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.arcDiameter = arcDiameter;
      
      sim.offer(ps);
   }      

   /**
    * Fill polygon.
    * 
    * @param   xPoints
    *          X coordinates.
    * @param   yPoints
    *          Y coordinates.
    */   
   public void fillPolygon(int[] xPoints, int[] yPoints)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_POLYGON);
      ps.xPoints = xPoints;
      ps.yPoints = yPoints;
      ps.nPoints = Math.min(xPoints.length, yPoints.length);
      
      sim.offer(ps);
   }
   
   /**
    * Force screen buffer draw.
    */
   public void flush()
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FLUSH);
      
      sim.offer(ps);      
   }
      
   /**
    * Draws a 3-D highlighted outline of the specified rectangle.
    * The edges of the rectangle are highlighted so that they
    * appear to be beveled and lit from the upper left corner.
    * <p>
    * The colors used for the highlighting effect are determined
    * based on the current color.
    * The resulting rectangle covers an area that is
    * <code>width&nbsp;+&nbsp;1</code> pixels wide
    * by <code>height&nbsp;+&nbsp;1</code> pixels tall.  This method
    * uses the current <code>Color</code> exclusively and ignores
    * the current <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be drawn.
    * @param   y 
    *          the y coordinate of the rectangle to be drawn.
    * @param   width 
    *          the width of the rectangle to be drawn.
    * @param   height 
    *          the height of the rectangle to be drawn.
    * @param   raised 
    *          a boolean that determines whether the rectangle appears 
    *          to be raised above the surface or sunk into the surface.
    */
   @Override
   public void draw3DRect(int x, int y, int width, int height, boolean raised)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_3D_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.raised = raised;
      
      sim.offer(ps);
   }

   /**
    * Paints a 3-D highlighted rectangle filled with the current color.
    * The edges of the rectangle are highlighted so that it appears
    * as if the edges were beveled and lit from the upper left corner.
    * The colors used for the highlighting effect and for filling are
    * determined from the current <code>Color</code>.  This method uses
    * the current <code>Color</code> exclusively and ignores the current
    * <code>Paint</code>.
    * 
    * @param   x 
    *          the x coordinate of the rectangle to be filled.
    * @param   y 
    *          the y coordinate of the rectangle to be filled.
    * @param   width
    *          the width of the rectangle to be filled.
    * @param   height 
    *          the height of the rectangle to be filled.
    * @param   raised 
    *          a boolean value that determines whether the rectangle appears 
    *          to be raised above the surface or etched into the surface.
    */
   @Override
   public void fill3DRect(int x, int y, int width, int height, boolean raised)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.FILL_3D_RECT);
      ps.x = x;
      ps.y = y;
      ps.width = width;
      ps.height = height;
      ps.raised = raised;
      
      sim.offer(ps);
   }
   
   /**
    * Draw a image at the specified location.
    * 
    * @param   img
    *          Image to draw.
    * @param   x
    *          left location.
    * @param   y
    *          top location.
    * @param   w
    *          the image width.
    * @param   h
    *          the image height.
    */
   @Override
   public void drawImage(Image img, int x, int y, int w, int h)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.DRAW_IMAGE);
      ps.x = x;
      ps.y = y;
      ps.width = w;
      ps.height = h;
      ps.img = img;

      sim.offer(ps);
   }
   
   /**
    * Set current color used on draw operations.
    * 
    * @param   color
    *          Drawing color.
    */
   @Override
   public void setColor(Color color)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_COLOR);
      ps.color = color;

      sim.offer(ps);
   }

   /**
    * Set current font used on draw operations.
    * 
    * @param   font
    *          Drawing font.
    */
   @Override
   public void setGuiFont(Font font)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_FONT);
      ps.font = font;

      sim.offer(ps);
   }

   /**
    * Scale the font so its width and height match the legacy values.
    * 
    * @param    font
    *           The font to be scaled.
    * @param    legacyWidth
    *           The legacy width which needs to be matched.
    * @param    legacyHeight
    *           The legacy height which needs to be matched.
    *           
    * @return   The scaled font.
    */
   @Override
   public Font scaleFont(Font font, int legacyWidth, int legacyHeight)
   {
      return sim.scaleFont(font, legacyWidth, legacyHeight);
   }

   /**
    * Set the style of the current font used on draw operations.
    * 
    * @param   style
    *          Font style.
    */
   @Override
   public void setFontStyle(int style)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_FONT_STYLE);
      ps.x = style;
      
      sim.offer(ps);
   }
   
   /**
    * Set the style of the current line drawing primitives.
    * 
    * @param   stroke
    *          Line style.
    */
   @Override
   public void setLineStroke(Stroke stroke)
   {
      PaintStructure ps = new PaintStructure(PaintPrimitives.SET_LINE_STROKE);
      ps.stroke = stroke;
      
      sim.offer(ps);
   }

   /**
    * Resize window.
    * 
    * @param   width
    *          new width.
    * @param   height
    *          new height
    */
   @Override
   public void resizeWindow(int width, int height)
   {
      sim.resizeWindow(width, height);
   }

   /**
    * Set window location.
    * 
    * @param   left
    *          left coordinate
    * @param   top
    *          top coordinate
    */
   @Override
   public void setWindowLocation(int left, int top)
   {
      sim.setWindowLocation(left, top);
   }
   
   /**
    * Get the window location on screen.
    * 
    * @return   See above.
    */
   @Override
   public NativePoint getWindowLocation()
   {
      return sim.getWindowLocation();
   }

   /**
    * Get the physical window dimension, in pixel units.
    * 
    * @return   See above.
    */
   @Override
   public NativeDimension getWindowDimension()
   {
      return new NativeDimension(sim.getWidth(), sim.getHeight());
   }

   /**
    * Create a font with the given details and register it with the driver.
    * 
    * @param    name
    *           The font face name.
    * @param    size
    *           The font's point size.
    * @param    bold
    *           Flag indicating if the font is bold.
    * @param    italic
    *           Flag indicating if the font is italic.
    * @param    underline
    *           Flag indicating if the font is underline.
    * @param    definition
    *           The font's definition.
    * @param    legacyWidth
    *           The legacy width which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * @param    legacyHeight
    *           The legacy height which needs to be matched by this font.  If no legacy details
    *           exist, then this argument will be negative.
    * 
    * @return   A font instance.
    */
   @Override
   public Font createFont(String  name,
                          int     size,
                          boolean bold,
                          boolean italic,
                          boolean underline,
                          byte[]  definition,
                          int     legacyWidth,
                          int     legacyHeight)
   {
      int style = 0;
      style += (bold ? Font.BOLD : 0);
      style += (italic ? Font.ITALIC : 0);
      
      // TODO: bitmap fonts
      
      Font f = new Font(name, style, size);
      if (underline)
      {
         f = f.deriveFont(UNDERLINE_FONT);
      }
      
      // scale the font to match the legacy metrics
      int width = getFontWidth(f);
      int height = getFontHeight(f);

      if ((legacyWidth > 0 && width != legacyWidth) || 
          (legacyHeight > 0 && height != legacyHeight))
      {
         double scaleX = (double) legacyWidth / (double) width;
         double scaleY = (double) legacyHeight / (double) height;
         f = f.deriveFont(AffineTransform.getScaleInstance(scaleX, scaleY));
      }
      
      // TODO: if a new font is created from the byte array, then register it with the graphics
      // for all windows

      return f;
   }

   /**
    * Get the width of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's width.
    */
   @Override
   public int getFontWidth(Font font)
   {
      FontMetrics fm = sim.getFontMetrics(font);
      
      int avgWidth = 0;
      int charno = 255;
      for (char ch = 1; ch <= 255; ch++)
      {
         avgWidth += fm.charWidth(ch);
      }
      avgWidth = (int) Math.ceil((avgWidth + 1.0) / (double) charno);

      return avgWidth;
   }

   /**
    * Get the height of the specified font.
    * 
    * @param    font
    *           The font.
    *           
    * @return   The fonts's height.
    */
   @Override
   public int getFontHeight(Font font)
   {
      FontMetrics fm = sim.getFontMetrics(font);
      
      return fm.getHeight();
   }

   /**
    * Get the width of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's width.
    */
   @Override
   public int getTextWidth(String text, Font font)
   {
      FontMetrics fm = sim.getFontMetrics(font);
      
      return fm.stringWidth(text);
   }

   /**
    * Get the height of the specified text, using the given font.
    * 
    * @param    text
    *           The text.
    * @param    font
    *           The font used to draw and measure the text.
    *           
    * @return   The text's height.
    */
   @Override
   public int getTextHeight(String text, Font font)
   {
      return sim.getTextHeight(text, font);
   }
   
   /**
    * Hide or show the window, depending on the {@code visible} flag.
    * 
    * @param    visible
    *           The window visibility state.
    */
   public void setWindowVisible(boolean visible)
   {
      sim.setWindowVisible(visible);
   }

   /**
    * Iconify the window.
    */
   @Override
   public void iconifyWindow()
   {
      sim.iconifyWindow();
   }

   /**
    * Get the screen width, in character units.
    * 
    * @return    See above.
    */
   @Override
   public int screenWidth()
   {
      return sim.screenWidth();
   }

   /**
    * Get the screen height, in character units.
    * 
    * @return    See above.
    */
   @Override
   public int screenHeight()
   {
      return sim.screenHeight();
   }
   
   /**
    * Set the active screen bitmap for the current window.
    * 
    * @param    bitmap
    *           The screen bitmap.
    */
   @Override
   public void setScreenBitmap(ScreenBitmap bitmap)
   {
      sim.setScreenBitmap(bitmap);
   }

   /**
    * Get the screen bitmap for the current window.
    * 
    * @return   See above.
    */
   @Override
   public ScreenBitmap getScreenBitmap()
   {
      return sim.getScreenBitmap();
   }

   /**
    * Set the mouse cursor to the given type.
    * 
    * @param    c
    *           The cursor type, one of the {@link Cursor} constants.
    */
   @Override
   public void setCursor(Cursor c)
   {
      sim.setCursor(c);
   }

   /**
    * Limit the available drawing area to the specified boundaries.
    * 
    * @param    bounds
    *           The bounds for the drawing area.
    */
   @Override
   public void setDrawingArea(NativeRectangle bounds)
   {
      sim.setDrawingArea(bounds);
   }

   /**
    * Clear specified area at the screen.
    * 
    * @param   left
    *          Left coordinate of the area.
    * @param   top
    *          Top coordinate of the area.
    * @param   right
    *          Right coordinate of the area.
    * @param   bottom
    *          Bottom coordinate of the area.
    * @param   color
    *          Color.
    */
   @Override
   public void clearArea(int left, int top, int right, int bottom, com.goldencode.p2j.ui.Color color)
   {
      sim.clearArea(left, top, right, bottom, color);
   }

   /**
    * Refresh specified area.
    * <p>
    * This is a no-op.
    * 
    * @param   rect
    *          Area to refresh. 
    */
   @Override
   public void refresh(NativeRectangle rect)
   {
      // no-op
   }

   /**
    * Synchronize screen by forcing a repaint.
    */
   @Override
   public void sync()
   {
      SwingUtilities.invokeLater(new Runnable()
      {
         @Override
         public void run()
         {
            sim.repaint();
         }
      });
   }

   /**
    * Perform shutdown.
    */
   public void shutdown()
   {
      for (GuiSimulator sim : renderers.values())
      {
         sim.quit();
      }
      
      renderers.clear();
   }
   
   /**
    * Check if window has renderer.
    * 
    * @param   windowId
    *          Window ID.
    * 
    * @return  <code>true</code> if window has renderer <code>true</false> otherwise.
    */
   public boolean hasWindowRenderer(int windowId)
   {
      return renderers.get(windowId) != null;
   }

   /**
    * Get the ID of the window associated with the current simulator.
    * 
    * @return   See above.
    */
   @Override
   public int getWindowId()
   {
      return sim.getWindowId();
   }

   /**
    * Get the currently selected simulator.
    * 
    * @return   See above.
    */
   public GuiSimulator getSimulator()
   {
      return sim;
   }
   
   /**
    * Returns display size.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    * 
    * @return   The dimensions in pixels.
    */
   public NativeDimension getDisplaySize(int windowId)
   {
      GuiSimulator sim = getSimulator(windowId);
      return sim.getDisplaySize();
   }

   /**
    * Returns an application-usable display area. For example, a task bar is
    * not counted as the usable area.
    * 
    * For multiple-display environments the display is selected based on the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *
    * @return   The area in pixels.
    */
   @Override
   public NativeRectangle getDisplayWorkArea(int windowId)
   {
      GuiSimulator sim = getSimulator(windowId);
      return sim.getDisplayWorkArea();
   }

   /**
    * Returns the gui simulator registered for the supplied window id.
    * 
    * @param    windowId
    *           A valid window id.
    *           
    * @return   A gui simulator reference.
    * 
    * @throws   RuntimeException
    *           When no renderer is registered for the supplied window id.
    */
   private GuiSimulator getSimulator(int windowId)
   {
      GuiSimulator sim = renderers.get(windowId);
      
      if (sim == null)
      {
         throw new RuntimeException("No renderer is registered for id = " + windowId);
      }

      return sim;
   }
}
