/*
** Module   : AbstractWidget.java
** Abstract : Base class for widgets.
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100822 Added base implementation of new interface methods config() and isFocusable().
** 004 SIY 20101122 Switching from KeyEvent to KeyInput.
** 005 SIY 20111115 Added physical widget location.
** 006 CA  20131121 The hidden/visible state changes must be collected when the widget's attribute
**                  is actually set.
** 007 EVL 20140617 Added method to check if the widget needs the cursor to be on.
** 008 CA  20140711 Enhanced font support. Added physicalDimension, in preparation for GUI support.
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added mouse event processing.  Added APIs to work with physical units, 
**                  depending on the UI type.
** 010 MAG 20140912 Added insets member used on GUI only.
** 011 HC  20140911 GUI coordinate support - data types refactoring.
** 012 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 013 MAG 20140917 Added parent width and height inheritance.
**                  Added mouse position translate.
** 014 CA  20141021 Changed the resize() API to specify the X and Y deltas, plus flags to indicate
**                  if the top-left corner needs to move on the X or Y axis.
**                  Added API to allow a widget to raise (or not) legacy mouse events, when a 
**                  mouse action was performed on it; by default all widgets can raise these events.
** 015 VIG 20141016 Added window() method to resolving widgets Window.
** 016 MAG 20141020 Added clipping rectangle computation.
** 017 HC  20141021 Implemented window move/position support.
** 018 HC  20141118 Added screenPhysicalLocation().
** 019 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 020 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui.client.widget;

import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.lang.ref.*;
import java.util.*;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;

/**
 * Base widget class. 
 */
public abstract class AbstractWidget<O extends OutputManager<?>>
implements Widget<O>
{
   /** Widget enabled flag. */
   private boolean enabled = true;
   
   /** Widget hidden flag. */
   private boolean hidden = false;
   
   /** Widget highlighted flag. */
   private boolean highlighted = false;
   
   /** Widget visible flag. */
   private boolean visible = true;
   
   /** Inherit parent width */
   private boolean inheritWidth = false;

   /** Inherit parent height */
   private boolean inheritHeight = false;

   /** Storage for focus listeners. */
   private ArrayList<FocusListener> focusListeners = new ArrayList<FocusListener>();
   
   /** Storage for key listeners. */
   private ArrayList<KeyListener> keyListeners = new ArrayList<KeyListener>();

   /** Widget name. */
   private String name = "";
   
   /** Widget origin (upper left corner) relative to container expressed in
    * character cell units. */
   private Point origin = new Point(0, 0);

   /** Widget physical origin (upper left corner) relative to container. 
       expressed in units native to UI type. */
   private NativePoint physicalOrigin = new NativePoint(0, 0);
   
   /** Reference to parent container. */
   private WeakReference<Container<O>> parent = null;

   /** List of listeners which require notifications about the state of this widget. */
   private List<WidgetStateListener> listeners = new ArrayList<>(); 

   /** Reference to screen. */
   private O screen;
   
   /** Widget insets */
   protected Insets insets = Insets.ZERO_INSETS;
   
   /** Widget native insets */
   protected NativeInsets nativeInsets = NativeInsets.ZERO_INSETS;

   /**
    * Constructor
    */
   @SuppressWarnings("unchecked")
   public AbstractWidget()
   {
      screen = (O) OutputManager.instance();

      ClientConfigManager mgr = ConfigManager.getInstance();
      this.registerWidgetStateListener(mgr);
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, WidgetConfig cfg)
   {
      // no-op by default
   }
   
   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   @Override
   public O screen() 
   {
      return screen;
   }

   /**
    * Add focus listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addFocusListener(FocusListener listener)
   {
      focusListeners.add(listener);
   }

   /**
    * Add key listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   @Override
   public void addKeyListener(KeyListener listener)
   {
      keyListeners.add(listener);
   }

   /**
    * Get ancestor window for this widget.
    * 
    * @return  ancestor window.
    */
   @Override
   public TitledWindow<O> ancestor()
   {
      if (this instanceof TitledWindow)
         return (TitledWindow<O>) this;

      Container<O> ancestor;
      Container<O> nextancestor;

      for (ancestor = parent(); !(ancestor instanceof TitledWindow); ancestor = nextancestor)
      {
         if (ancestor == null)
            return null;

         nextancestor = ancestor.parent();
         
         if (nextancestor == null)
            return null;
      }
      
      return (TitledWindow<O>) ancestor;
   }
   
   /**
    * Get the ID of the ancestor window for this widget.
    * 
    * @return  ancestor window ID or <code>null</code> if no valid ancestor exists.
    */
   public WidgetId ancestorId()
   {
      Widget<?> ancestor = ancestor();
      if (ancestor == null)
      {
         return null;
      }
      
      return ancestor.config().id;
   }

   /**
    * Get ancestor Window for this widget. This method must be called only for widgets
    * that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   @Override
   public Window<O> window()
   {
      TitledWindow<O> wnd = ancestor();
      
      if (!(wnd instanceof Window)) 
      {
         Widget<O> parent = (wnd == null ? null : wnd.parent());
         while (parent != null)
         {
            // get the real parent window
            if (parent instanceof Window)
            {
               break;
            }
            
            parent = parent.parent();
         }
         
         if (parent instanceof Window)
         {
            wnd = (Window<O>) parent;
         }
         else
         {
            int wid = UiUtils.getWidgetIdAsInt(this);
            throw new IllegalStateException("Widget " + name() + " " + wid + 
                                            " is not attached to a Window instance.");
         }
      }
      
      return (Window<O>) wnd;
   }
   
   /**
    * Get widget bounding rectangle.
    * 
    * @return  bounding rectangle.
    */
   @Override
   public Rectangle bounds()
   {
      Rectangle result;
      
      if (screen().isChui())
      {
         result = new Rectangle(location(), dimension());
      }
      else
      {
         // TODO: #2323 The base coordinate value differs for ChUI and GUI modes.
         //       Remove this branch when Rectangle offsets the correct
         //       base value.
         Point l = location();
         Dimension d = dimension();
         result = new Rectangle(l.y, l.x, l.y + d.height - 0.01, l.x + d.width - 0.01);
      }
      
      return result;
   }

   /**
    * Get widget insets.
    * 
    * @return  widget insets.
    */
   @Override
   public Insets insets()
   {
      return insets;
   }

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public abstract Dimension dimension();

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   @Override
   public NativeDimension physicalDimension()
   {
      return screen.coordinates().dimensionToNative(dimension());
   }

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   @Override
   public boolean isDisplayed()
   {
      return (parent() == null) ? false : parent().isDisplayed();
   }

   /**
    * Draw widget.
    */
   @Override
   public abstract void draw();

   /**
    * Draw caret.
    */
   @Override
   public void drawCaret()
   {
   }

   /**
    * Check if widget is enabled.
    * 
    * @return  <code>true</code> if widget is enabled.
    */
   @Override
   public boolean isEnabled()
   {
      return enabled;
   }

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   @Override
   public boolean focusTraversable()
   {
      return (enabled && visible && !hidden);
   }

   /**
    * Force widget visibility change without associated repainting.
    * 
    * @param   visibility
    *          New visibility state.
    */
   @Override
   public void forceVisibility(boolean visibility)
   {
      _setVisible(visibility); // direct access
   }

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   @Override
   public boolean hasFocus()
   {
      return (parent() == null || parent().focus() != this) ? 
                                                         false : 
                                                         parent().hasFocus();
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public abstract double height();

   /**
    * Get current hidden state.
    * 
    * @return   hidden state.
    */
   @Override
   public boolean hidden()
   {
      return hidden;
   }

   /**
    * Hide widget.
    */
   @Override
   public void hide()
   {
      if (isVisible())
      {
         _setVisible(false); // direct access
         repaint();
      }
   }

   /**
    * Get current highlighted state.
    * 
    * @return   highlighted state.
    */
   @Override
   public boolean highlighted()
   {
      return highlighted;
   }

   /**
    * Get widget location relative to container.
    * 
    * @return  relative widget location.
    */
   @Override
   public Point location()
   {
      return new Point(origin);
   }

   /**
    * Get widget physical location relative to container, expressed in units native to UI type.
    * 
    * @return  relative widget location.
    */
   @Override
   public NativePoint physicalLocation()
   {
      return new NativePoint(physicalOrigin);
   }
   
   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public abstract Dimension minimumSize();

   /**
    * Get widget name.
    * 
    * @return  widget name.
    */
   @Override
   public String name()
   {
      return name;
   }

   /**
    * Get widget parent.
    * 
    * @return  widget parent.
    */
   @Override
   public Container<O> parent()
   {
      return (parent == null) ? null : parent.get();
   }

   /**
    * Process focus event.
    * 
    * @param   event
    *          Event to process.
    */
   @Override
   public void processFocusEvent(FocusEvent event)
   {
      for (FocusListener listener : focusListeners)
      {
         event.dispatch(listener);
      }
   }

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      processKeyListeners(event);
   }

   /**
    * Process key listeners.
    * 
    * @param    event
    *           Keyboard event.
    */
   @Override
   public void processKeyListeners(KeyInput event)
   {
      for (KeyListener listener : keyListeners)
      {
         event.dispatch(listener);

         if (event.isConsumed())
            break;
      }
   }

   /**
    * Request widget repainting.
    */
   @Override
   public void repaint()
   {
      if (!isDisplayed())
         return;

      repaint(bounds());
   }

   /**
    * Request widget repainting.
    * 
    * Only the given update rectangle will be repainted (note that it is
    * clipped by all parent containers). No event is posted if the widget
    * itself or any of the parent containers are hidden.
    * 
    * @param    ur
    *           The update rectangle (relative to the origin of its parent
    *           Container) which will be repainted. Note that it is clipped by
    *           all parent containers.
    */
   @Override
   public void repaint(Rectangle ur)
   {
      if (!isDisplayed())
         return;

      Widget<O> cm = this;
      Container<O> ct = parent();

      // do not readjust coordinates if the root window is reached
      while (ct != null && !(ct instanceof Window))
      {
         if (!ct.isVisible())
         {
            return;
         }

         if (ct.clipping())
         {
            ur = ur.intersection(ct.getClipRect());

            if (ur == null || ur.empty())
            {
               return;
            }
         }

         cm = ct;
         ct = cm.parent();
         ur = ur.translate(cm.location());
      }

      EventManager.postEvent(new PaintEvent(this, ur));
   }

   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      Widget<O> currentFocus = ancestor().currentFocus();
      
      if (currentFocus == this)
         return;
      
      if (currentFocus != null)
         EventManager.postEvent(new FocusEvent(currentFocus, EventType.FOCUS_LOST));

      EventManager.postEvent(new FocusEvent(this, EventType.FOCUS_GAINED));

      parent().setFocus(this);
      repaint();
   }

   /**
    * Request screen synchronization.  
    */
   @Override
   public void requestSync()
   {
      EventManager.postEvent(new SyncEvent(this));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      if (parent() == null) {
         throw new IllegalStateException("widget is not in the container");
      }

      return Coordinate.scale(parent().screenLocation().translate(location()));
   }

   /**
    * Get absolute location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  absolute widget location.
    */
   @Override
   public NativePoint screenPhysicalLocation()
   {
      if (parent() == null) 
      {
         throw new IllegalStateException("widget is not in the container");
      }

      return parent().screenPhysicalLocation().translate(physicalLocation());
   }
   
   /**
    * Get display location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  display widget location.
    */
   @Override
   public NativePoint displayPhysicalLocation()
   {
      Window<?> window = window();
      if (window == null)
      {
         return screenPhysicalLocation();
      }
      
      // window physical location represents a location on the display, 
      // whereas screen physical location is the offset from the parent window
      return window.physicalLocation().translate(screenPhysicalLocation());
   }

  /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   @Override
   public void setEnabled(boolean enabled)
   {
      _setEnabled(enabled);

      repaint();
   }

   /**
    * Tracks the hidden state of the widget.
    * 
    * @param    hidden
    *           New hidden state.
    */
   @Override
   public void setHidden(boolean hidden)
   {
      _setHidden(hidden);
   }

   /**
    * Tracks the highlighted state of the widget.
    * 
    * @param    highlighted
    *           New highlighted state.
    */
   @Override
   public void setHighlighted(boolean highlighted)
   {
      this.highlighted = highlighted;
   }

   /**
    * Set widget relative location.
    * 
    * The x value corresponds to COLUMN while
    * y value corresponds to ROW 4GL attribute.
    * Note however that the values are expected to be
    * 0-based.
    * 
    * @param   x
    *          Widget X.
    * @param   y
    *          Widget Y.
    */
   @Override
   public void setLocation(double x, double y)
   {
      x = Coordinate.scale(x);
      y = Coordinate.scale(y);

      if (isVisible())
         repaint(); 
      
      origin.x = x;
      origin.y = y;
      
// TODO: setting the config column/row causes layout of FillIn fail.
//      WidgetConfig config = config();
//      if (config != null && config instanceof BaseConfig)
//      {
//         BaseConfig bconfig = (BaseConfig) config;
//         bconfig.column = x;
//         bconfig.row = y;
//      }
      
      if (isVisible())
         repaint();
   }

   /**
    * Set widget relative location.
    * 
    * @param   origin
    *          Location to set.
    */
   @Override
   public void setLocation(Point origin)
   {
      setLocation(origin.x, origin.y);
   }

   /**
    * Set widget relative physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   @Override
   public void setPhysicalLocation(int x, int y)
   {
      // note that these values not checked nor processed in any way,
      // widget just holds them as a container
      physicalOrigin.x = x;
      physicalOrigin.y = y;
      
      // config.x and config.y hold pixel values, but in chui the method parameters
      // really represent native units (terminal characters for example)
      if (!screen().isChui())
      {
         WidgetConfig config = config();
         if (config != null && config instanceof BaseConfig)
         {
            BaseConfig bconfig = (BaseConfig) config;
            bconfig.x = x;
            bconfig.y = y;
         }
      }
   }
   
   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   @Override
   public NativeRectangle physicalBounds()
   {
      return new NativeRectangle(physicalLocation(), physicalDimension());
   }
   
   /**
    * Move this widget on the X and Y axis, by the specified deltas.  Values are specified in
    * units depending on the UI type.
    * 
    * @param   dx
    *          The X-axis delta.
    * @param   dy
    *          The Y-axis delta.
    */
   @Override
   public void moveBy(int dx, int dy)
   {
      // no-op by default
   }
   
   /**
    * Resize the widget to the specified dimension.  Values are specified in units depending on
    * the UI type.
    * 
    * @param    dx
    *           The X axis adjustment.
    * @param    dy
    *           The Y axis adjustment.
    * @param    moveX
    *           Flag indicating that the top-left corner needs to be moved on the X axis.
    * @param    moveY
    *           Flag indicating that the top-left corner needs to be moved on the Y axis.
    */
   @Override
   public void resize(int dx, int dy, boolean moveX, boolean moveY)
   {
      // no-op by default
   }
   
   /**
    * Set widget name.
    * 
    * @param   name
    *          New widget name.
    */
   @Override
   public void setName(String name)
   {
      this.name = name;
   }

   /**
    * Set widget parent.
    * 
    * @param   container
    *          Parent to set.
    */
   @Override
   public void setParent(Container<O> container)
   {
      parent = (container == null) ? 
               null : 
               new WeakReference<Container<O>>(container);
   }

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      if (visible)
         show();
      else
         hide();
   }

   /**
    * Show widget.
    */
   @Override
   public void show()
   {
      if (!isVisible())
      {
         _setVisible(true); // direct access
         repaint();
      }
   }

   /**
    * Check if widget is totally obscured.
    * 
    * @return  <code>true</code> if widget is totally obscured.
    */
   @Override
   public boolean isTotallyObscured()
   {
      return WindowManager.checkRectangle(bounds(), ancestor());
   }

   /**
    * Check if widget is visible.
    * 
    * @return  <code>true</code> if widget is visible.
    */
   @Override
   public boolean isVisible()
   {
      return visible;
   }
   
   /**
    * Check if widget is part of a header.
    * 
    * @return  <code>true</code> if widget is part of header.
    */
   @Override
   public boolean isHeader()
   {
      return config().header;
   }

   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return true;
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public abstract double width();

   /**
    * Disable widget.
    */
   protected void disable()
   {
      _setEnabled(false);
   }

   /**
    * Enable widget.
    */
   protected void enable()
   {
      _setEnabled(true);
   }

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   evt
    *          Event to process.
    */
   @Override
   public void processEvent(Event evt)
   {
      if (evt instanceof KeyInput)
      {
         Widget<O> ancestor = ancestor();

         if (ancestor != null)
            ancestor.processKeyEvent((KeyInput) evt);
      }
      else if (evt instanceof FocusEvent)
         processFocusEvent((FocusEvent) evt);
   }

   /**
    * By default widget has no config.
    */
   @Override
   public <W extends WidgetConfig> W config()
   {
      return null;
   }
   
   /**
    * Called by the framework to notify the widget that the associated config state 
    * may have changed.
    */
   @Override
   public void syncWithConfig()
   {
      // no-op
   }

   /**
    * Checks whether this widget can receive focus generally.
    * 
    * @return  <code>true</code> if this widget can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          widget can never receive focus.    
    */
   @Override
   public boolean isFocusable()
   {
      return true;
   }
   
   /**
    * This API is responsible for resolving the font associated with this widget's title.  
    * This can be one of the default fonts, a special system font, or an entry in the font-table.
    * 
    * @return   See above.
    */
   public int resolveTitleFont()
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * This API is responsible for resolving the widget's font.  This can be one of the default 
    * fonts, a special system font, or an entry in the font-table.
    * 
    * @return   See above.
    */
   public int resolveFont()
   {
      throw new RuntimeException("Not implemented.");
   }
   
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   @Override
   public Widget<O> findMouseSource(NativePoint p)
   {
      return isVisible() && this.physicalBounds().contains(p) ? this : null;
   }
   
   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse entered event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      // no-op
   }

   /**
    * Notification of a mouse wheel moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseWheelMoved(MouseWheelEvent e)
   {
      // no-op
   }
   
   /**
    * Destroy this widget.  Notify all listeners about the widget being destroyed.
    */
   public void destroy()
   {
      for (WidgetStateListener l : listeners)
      {
         l.notifyWidgetDestroyed(this);
      }
   }

   /**
    * Return whether width is inherited from parent widget.
    * 
    * @return  see above.
    */
   public boolean isInheritWidth()
   {
      return inheritWidth;
   }

   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritWidth
    *          <code>true</code> if width is inherited from parent. 
    */
   public void setInheritWidth(boolean inheritWidth)
   {
      this.inheritWidth = inheritWidth;
   }

   /**
    * Return whether height is inherited from parent widget.
    * 
    * @return  see above.
    */
   public boolean isInheritHeight()
   {
      return inheritHeight;
   }

   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritHeight
    *          <code>true</code> if height is inherited from parent. 
    */
   public void setInheritHeight(boolean inheritHeight)
   {
      this.inheritHeight = inheritHeight;
   }

   /**
    * Register a widget state listener for this widget.
    * 
    * @param    l
    *           The widget state listener.
    */
   public void registerWidgetStateListener(WidgetStateListener l)
   {
      if (l != null)
      {
         this.listeners.add(l);
      }
   }
   
   /**
    * Translate mouse position relative to widget. Widget location is relative to container on
    * which is layout. The mouse position is relative to window content pane which is the mouse
    * listener. This function translate mouse position coordinates relative to widget.
    *  
    * @param   widget
    *          Widget instance.
    * @param   evt
    *          Mouse event.
    *          
    * @return  see above.
    */
   public NativePoint translate(Widget<O> widget, MouseEvent evt)
   {
      // get widget physical location wich is relative to it container. 
      NativePoint p = widget.physicalLocation();

      // get inner containers and calculate position
      Widget<O> parent = widget.parent();
      
      while (!(parent instanceof Window<?>))
      {
         p.translate(parent.physicalLocation());
         parent = parent.parent();
      }
      
      // mouse position relative to widget position
      return new NativePoint(evt.getX() - p.x, evt.getY() - p.y);
   }   

   /**
    * Gets the current state of this component's data.
    * 
    * @return   The state of this widget's data  (for example 
    *           <code>ScreenBuffer.UNINITIALIZED</code>).
    */
   @Override
   public byte getState()
   {
      return config().state;
   }

   /**
    * Stores the new state of the component's data. 
    * 
    * @param    state
    *           The new state (for example <code>ScreenBuffer.CHANGED</code>).
    */
   @Override
   public void setState(byte state)
   {
      if (state < ScreenBuffer.STATE_MIN || state > ScreenBuffer.STATE_MAX)
         return;

      WidgetConfig config = this.config();
      if (config != null)
      {
         config.state = (byte) state;
      }
   }

   /**
    * Build widget clipped rectangle.
    * 
    * @return  see above.
    */
   protected NativeRectangle clipRectangle()
   {
      NativePoint origin = physicalLocation();
      NativeDimension dimension = physicalDimension();
      return clipRectangle(origin, dimension);
   }
   
   /**
    * Build clipped rectangle for given coordinates.
    * 
    * @param   origin
    *          Physical location.
    * @param   dimension
    *          Physical dimension.
    * 
    * @return  see above.
    */
   protected NativeRectangle clipRectangle(NativePoint origin, NativeDimension dimension)
   {
      // corrections for scrolling 
      int x = dimension.width  - (origin.x < 0 ? origin.x : 0);
      int y = dimension.height - (origin.y < 0 ? origin.y : 0);
      // clip rectangle
      return new NativeRectangle(0, 0, y, x);     
   }
   
   /**
    * Update the {@link #enabled} flag in the configuration and in the widget's state.
    * 
    * @param    enabled
    *           The widget's enabled state.
    */
   private void _setEnabled(boolean enabled)
   {
      this.enabled = enabled;
      if (config() != null)
      {
         config().enabled = enabled;
      }
   }
   
   /**
    * Update the {@link #visible} flag in the configuration and in the widget's state.
    * 
    * @param    visible
    *           The widget's visible state.
    */
   public void _setVisible(boolean visible)
   {
      this.visible = visible;
      WidgetConfig config = config();
      if (config != null)
      {
         config.visible = visible;
         if (visible && !config.realized)
         {
            config.realized = true;
         }
      }
   }
   
   /**
    * Update the {@link #hidden} flag in the configuration and in the widget's state.
    * 
    * @param    hidden
    *           The widget's hidden state.
    */
   private void _setHidden(boolean hidden)
   {
      this.hidden = hidden;
      if (config() != null)
      {
         config().hidden = hidden;
      }
   }

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   public boolean raiseLegacyMouseEvents()
   {
      return true;
   }
   
   /**
    * Determine the widget's window and synchronize output with physical terminal.
    * 
    * @see   OutputManager#sync(int)
    */
   public void sync()
   {
      int wid = window().config().id.getId();
      
      screen().sync(wid);
   }
   
   /**
    * Determine the widget's window and mark a rectangle as needing repainting. 
    * 
    * @param   rect
    *          rectangle to mark for repainting
    *          
    * @see     OutputManager#refresh(int, Rectangle)
    */
   public void refresh(Rectangle rect)
   {
      int wid = window().config().id.getId();
      
      screen().refresh(wid, rect);
   }
   
   /**
    * Determine the widget's window and clear specified area of the screen.
    * 
    * @param   origin
    *          Upper left corner of the area to clear.
    * @param   size
    *          Dimensions of the area.
    * @param   color
    *          Color.
    * 
    * @see     OutputManager#clear(int, NativePoint, NativeDimension, Color)
    */
   public void clear(Point origin, Dimension size, Color color)
   {
      int wid = window().config().id.getId();
      
      screen().clear(wid, origin, size, color);
   }
   
   /**
    * Determine the widget's window, get a copy of the current screen bitmap and reset clipping 
    * entirely. All screen positions become available for output. The previous state can be
    * restored using setBitmap() method.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getAndResetBitmap(int)
    */
   public ScreenBitmap getAndResetBitmap()
   {
      int wid = window().config().id.getId();

      return screen().getAndResetBitmap(wid);
   }
   
   /**
    * Determine the widget's window and reset the screen bitmap.
    * 
    * @see     OutputManager#resetBitmap(int)
    */
   public void resetBitmap()
   {
      int wid = window().config().id.getId();

      screen().resetBitmap(wid);
   }

   /**
    * Determine the widget's window, get a copy of the current screen bitmap and clip output 
    * further so it is guaranteed not to go outside of the provided rectangle.
    * 
    * @param   rect
    *          The extra clipping rectangle. <code>null</code> has no effect
    *          on clipping.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getAndSetBitmap(int, Rectangle)
    */
   public ScreenBitmap getAndSetBitmap(Rectangle rect)
   {
      int wid = window().config().id.getId();

      return screen().getAndSetBitmap(wid, rect);
   }

   /**
    * Determine the widget's window and get a copy of the current screen bitmap.
    * 
    * @return  a copy of the current screen bitmap.
    * 
    * @see     OutputManager#getBitmapCopy(int)
    */
   public ScreenBitmap getBitmapCopy()
   {
      int wid = window().config().id.getId();

      return screen().getBitmapCopy(wid);
   }
   
   /**
    * Determine the widget's window and restore the bitmap to a previous state.
    * 
    * @param   bitmap
    *          The bitmap instance to set as the output primitive's bitmap.
    * 
    * @see     OutputManager#setBitmap(int, ScreenBitmap)
    */
   public void setBitmap(ScreenBitmap bitmap)
   {
      int wid = window().config().id.getId();

      screen().setBitmap(wid, bitmap);
   }

   /**
    * Determine the widget's window and add a new clipping rectangle to the bitmap. If the 
    * rectangle is null, fill the entire bitmap with <code>true</code> values. Else, if the bitmap
    * was just reset, fill it with <code>false</code> values and set the clipping region to true.
    * 
    * @param   rect
    *          The clipping rectangle to be added.
    * 
    * @see     OutputManager#clipWith(int, Rectangle)
    */
   public void clipWith(Rectangle rect)
   {
      int wid = window().config().id.getId();

      screen().clipWith(wid, rect);
   }
   
   /**
    * Determine the widget's window and get the number of rows on the screen.
    * 
    * @return  The number of rows (lines of characters).
    */
   public double screenHeight()
   {
      int wid = window().config().id.getId();

      return screen.screenHeight(wid);
   }

   /**
    * Determine the widget's window and get the number of columns on the screen.
    *
    * @return  The number of columns (in characters).
    */
   public double screenWidth()
   {
      int wid = window().config().id.getId();

      return screen.screenWidth(wid);
   }

   /**
    * Determine the widget's window and get the screen dimensions.
    * 
    * @return screen dimension.
    */
   public Dimension screenDimension()
   {
      int wid = window().config().id.getId();

      return screen.screenDimension(wid);
   }
}
