/*
** Module   : Widget.java
** Abstract : Common interface for all widgets. 
**
** Copyright (c) 2010-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 SIY 20100629 Created initial version
** 002 SIY 20100804 Changes in imports, some refactorings.
** 003 SIY 20100817 Fixed return type for ancestor() method.
** 004 SIY 20100822 Added config() and isFocusable() methods.
** 005 SIY 20101122 Switching from KeyEvent to KeyInput.
** 006 SIY 20111115 Added getter/setter for physical widget location.
** 007 EVL 20140617 Added method to check if the widget needs the cursor to be on.
** 008 CA  20140711 Added physicalDimension, in preparation for GUI support.
** 009 CA  20140805 Added generics to force the abstract widget classes to not use implementation-
**                  specific code: the abstract widgets will have access only to the base APIs
**                  provided by the OutputManager.
**                  Added mouse event support.  More APIs to ease GUI implementation.
** 010 MAG 20140912 Added insets() function used on GUI only.
** 011 HC  20140911 GUI coordinate support - data types refactoring.
** 012 CA  20140926 Added shared widget configuration support, to allow GUI/ChUI concrete
**                  implementation for the same widget ID.  Refactored the widget configuration
**                  classes: all fields were made public; these classes need to be as dumb as
**                  possible, all logic related to setting/getting a certain field should be at
**                  the widget or at the caller. Refs #2254
** 013 MAG 20140917 Added parent width and height inheritance.
** 014 CA  20141021 Changed the resize() API to specify the X and Y deltas, plus flags to indicate
**                  if the top-left corner needs to move on the X or Y axis.
**                  Added API to allow a widget to raise (or not) legacy mouse events, when a 
**                  mouse action was performed on it.
** 015 HC  20141021 Implemented window move/position support.
** 016 HC  20141118 Added screenPhysicalLocation().
** 017 CA  20141119 Refactored drawing code to enable multiple OS windows, for GUI support.  Misc
**                  changes/fixes related to drawing on multiple windows.
** 018 HC  20141207 Implemented runtime support for widget coordinate attributes.
*/

package com.goldencode.p2j.ui.client.widget;

import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelListener;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.event.listener.*;

/**
 * Common interface implemented by all widgets.
 */
public interface Widget<O extends OutputManager<?>>
extends MouseListener,
        MouseMotionListener,
        MouseWheelListener
{
   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   public void initialize(WidgetId id, WidgetConfig cfg);

   /**
    * Get access to {@link OutputManager} instance.
    * 
    * @return  reference to {@link OutputManager}.
    */
   public O screen();

   /**
    * Add focus listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   public void addFocusListener(FocusListener listener);

   /**
    * Add key listener.
    * 
    * @param   listener
    *          Listener to add.
    */
   public void addKeyListener(KeyListener listener);

   /**
    * Get ancestor window for this widget.
    * 
    * @return  ancestor window.
    */
   public Widget<O> ancestor();

   /**
    * Get ancestor Window for this widget. This method must be called only for widgets
    * that are already attached to Window.
    * 
    * @return  ancestor Window.
    * 
    * @throws  IllegalStateException
    *          When this widget is not attached to window.
    */
   public Window<O> window();

   /**
    * Get the ID of the ancestor window for this widget.
    * 
    * @return  ancestor window ID.
    */
   public WidgetId ancestorId();
   
   /**
    * Get widget bounding rectangle.
    * 
    * @return  bounding rectangle.
    */
   public Rectangle bounds();

   /**
    * Get widget insets.
    * 
    * @return  widget insets.
    */
   public Insets insets();

   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   public Dimension dimension();

   /**
    * Get widget physical dimension, expressed in units native to UI type.
    *  
    * @return  widget dimension.
    */
   public abstract NativeDimension physicalDimension();

   /**
    * Check if widget is displayed.
    * 
    * @return  <code>true</code> if widget is displayed.
    */
   public boolean isDisplayed();

   /**
    * Draw widget.
    */
   public void draw();

   /**
    * Draw caret.
    */
   public void drawCaret();

   /**
    * Check if widget is enabled.
    * 
    * @return  <code>true</code> if widget is enabled.
    */
   public boolean isEnabled();

   /**
    * Check if widget is focus traversable.
    * 
    * @return  <code>true</code> if widget is focus traversable.
    */
   public boolean focusTraversable();

   /**
    * Force widget visibility change without associated repainting.
    * 
    * @param   visibility
    *          New visibility state.
    */
   public void forceVisibility(boolean visibility);

   /**
    * Check if widget has focus.
    * 
    * @return  <code>true</code> if widget has focus.
    */
   public boolean hasFocus();

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   public double height();

   /**
    * Get current hidden state.
    * 
    * @return   hidden state.
    */
   public boolean hidden();

   /**
    * Hide widget.
    */
   public void hide();

   /**
    * Get current highlighted state.
    * 
    * @return   highlighted state.
    */
   public boolean highlighted();

   /**
    * Get widget location relative to container.
    * 
    * @return  relative widget location.
    */
   public Point location();

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   public Dimension minimumSize();

   /**
    * Get widget name.
    * 
    * @return  widget name.
    */
   public String name();

   /**
    * Get widget parent.
    * 
    * @return  widget parent.
    */
   public Container<O> parent();

   /**
    * Process focus event.
    * 
    * @param   event
    *          Event to process.
    */
   public void processFocusEvent(FocusEvent event);

   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   public void processKeyEvent(KeyInput event);

   /**
    * Process key listeners.
    * 
    * @param    event
    *           Keyboard event.
    */
   public void processKeyListeners(KeyInput event);

   /**
    * Request widget repainting.
    */
   public void repaint();

   /**
    * Request repainting of the area occupied by widget.
    * 
    * Only the given update rectangle will be repainted (note that it is
    * clipped by all parent containers). No event is posted if the widget
    * itself or any of the parent containers are hidden.
    * 
    * @param    ur
    *           The update rectangle (relative to the origin of its parent
    *           Container) which will be repainted. Note that it is clipped by
    *           all parent containers.
    */
   public void repaint(Rectangle ur);

   /**
    * Request focus.
    */
   public void requestFocus();

   /**
    * Request screen synchronization.  
    */
   public void requestSync();

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   public Point screenLocation();

   /**
    * Get absolute location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  absolute widget location.
    */
   public NativePoint screenPhysicalLocation();
   
   /**
    * Get display location of widget in screen coordinates.
    * The result is in native units.
    * 
    * @return  display widget location.
    */
   public NativePoint displayPhysicalLocation();
   
   /**
    * Enable/disable widget.
    * 
    * @param   enabled
    */
   public void setEnabled(boolean enabled);

   /**
    * Tracks the hidden state of the widget.
    * 
    * @param    hidden
    *           New hidden state.
    */
   public void setHidden(boolean hidden);

   /**
    * Tracks the highlighted state of the widget.
    * 
    * @param    highlighted
    *           New highlighted state.
    */
   public void setHighlighted(boolean highlighted);

   /**
    * Set widget relative location.
    * 
    * @param   x
    *          Widget X.
    * @param   y
    *          Widget Y.
    */
   public void setLocation(double x, double y);

   /**
    * Set widget relative location.
    * 
    * @param   origin
    *          Location to set.
    */
   public void setLocation(Point origin);

   /**
    * Set widget name.
    * 
    * @param   name
    *          New widget name.
    */
   public void setName(String name);

   /**
    * Set widget parent.
    * 
    * @param   container
    *          Parent to set.
    */
   public void setParent(Container<O> container);

   /**
    * Hide/show widget.
    * 
    * @param   visible
    *          New visible state.
    */
   public void setVisible(boolean visible);

   /**
    * Show widget.
    */
   public void show();

   /**
    * Check if widget is totally obscured.
    * 
    * @return  <code>true</code> if widget is totally obscured.
    */
   public boolean isTotallyObscured();

   /**
    * Check if widget is visible.
    * 
    * @return  <code>true</code> if widget is visible.
    */
   public boolean isVisible();

   /**
    * Check if widget is part of a header.
    * 
    * @return  <code>true</code> if widget is part of header.
    */
   public boolean isHeader();
   
   /**
    * Check if widget needs in cursor to be on when having the input focus.
    * 
    * @return  <code>true</code> if cursor is required.
    */
   public boolean forceCursorOnFocus();

   /**
    * Get widget width in character units.
    * 
    * @return  widget width.
    */
   public double width();

   /**
    * Process event and dispatch it to specific method.
    * 
    * @param   evt
    *          Event to process.
    */
   public void processEvent(Event evt);
   
   /**
    * Get externalizable widget config which can be used to restore widget 
    * state after via network. 
    * Note that only very limited set of widgets actually have config.
    * 
    * @param   <W>
    *          Config type.
    *          
    * @return  Reference to widget config.
    */
   public <W extends WidgetConfig> W config();
   
   /**
    * Called by the framework to notify the widget that the associated config state 
    * may have changed.
    */
   public void syncWithConfig();

   /**
    * Checks whether this widget can receive focus generally.
    * 
    * @return  <code>true</code> if this widget can be focused under
    *          certain conditions. <code>false</code> if, and only if, this
    *          widget can never receive focus.    
    */
   boolean isFocusable();

   /**
    * Set widget relative physical location.
    * 
    * @param   x
    *          Widget physical X.
    * @param   y
    *          Widget physical Y.
    */
   void setPhysicalLocation(int x, int y);

   /**
    * Get widget physical location relative to container, expressed in units native to UI type.
    * 
    * @return  relative widget location.
    */
   public NativePoint physicalLocation();

   /**
    * Determine the physical bounds, relative to screen, expressed in units native to UI type.
    * 
    * @return  See above.
    */
   public NativeRectangle physicalBounds();
  
   /**
    * Find the widget positioned just bellow the specified mouse position (in physical units).
    * 
    * @param    p
    *           The mouse physical location.
    * 
    * @return   A widget over which the mouse is positioned or <code>null</code> if none found.
    */
   public Widget<O> findMouseSource(NativePoint p);
   
   /**
    * Destroy this widget.
    */
   public void destroy();

   /**
    * Move this widget on the X and Y axis, by the specified deltas.  Values are specified in
    * native units depending on the UI type.
    * 
    * @param   dx
    *          The X-axis delta.
    * @param   dy
    *          The Y-axis delta.
    */
   public void moveBy(int dx, int dy);

   /**
    * Resize the widget to the specified dimension.  Values are specified in native units 
    * depending on the UI type.
    * 
    * @param    dx
    *           The X axis adjustment.
    * @param    dy
    *           The Y axis adjustment.
    * @param    moveX
    *           Flag indicating that the top-left corner needs to be moved on the X axis.
    * @param    moveY
    *           Flag indicating that the top-left corner needs to be moved on the Y axis.
    */
   public void resize(int dx, int dy, boolean moveX, boolean moveY);
   
   /**
    * Return whether width is inherited from parent widget.
    * 
    * @return  see above.
    */
   public boolean isInheritWidth();

   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritWidth
    *          <code>true</code> if width is inherited from parent. 
    */
   public void setInheritWidth(boolean inheritWidth);

   /**
    * Return whether height is inherited from parent widget.
    * 
    * @return  see above.
    */
   public boolean isInheritHeight();

   /**
    * Set if width is inherited from parent widget.
    * 
    * @param   inheritHeight
    *          <code>true</code> if height is inherited from parent. 
    */
   public void setInheritHeight(boolean inheritHeight);
   
   /**
    * Gets the current state of this component's data.
    * 
    * @return   The state of this widget's data  (for example 
    *           <code>ScreenBuffer.UNINITIALIZED</code>).
    */
   public byte getState();

   /**
    * Stores the new state of the component's data. 
    * 
    * @param    state
    *           The new state (for example <code>ScreenBuffer.CHANGED</code>).
    */
   public void setState(byte state);

   /**
    * Check if the widget implementation can raise legacy, 4GL-style, mouse events.
    * 
    * @return   <code>true</code> if a mouse action on this widget will raise legacy, 4GL-style,
    *           mouse events.
    */
   public boolean raiseLegacyMouseEvents();
}
